#%% Load Packages & Define Functions
import os
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt

def Get_Locations_with_Data(a_data):
    m_data = (~np.isnan(a_data)).any(dim='time')
    m_coords = np.empty([0,2])
    for i_lat in m_data.lat:
        for i_lon in m_data.lon:
            if m_data.sel(lat=i_lat,lon=i_lon):
                m_coords = np.r_[m_coords,np.array([i_lat,i_lon],ndmin=2)]          
    return m_coords

def Array_to_Matrix_at_Coords(a_data,m_coords):
    n_coords = m_coords.shape[0]
    if hasattr(a_data,'time'):
        n_times = len(a_data.time)
    else:
        n_times = 1
    m_data = np.empty([n_times,n_coords])
    m_data[:] = np.nan
    for i_coord in range(n_coords):
        m_data[:,i_coord] = np.array(a_data.to_array().sel(lat=m_coords[i_coord,0],lon=m_coords[i_coord,1]))
    return m_data

#%% Basic Settings
        
figure_format = '.svg'

PLOT_TYPE = 1
# 0: two figures
# 1: one figure

PLOT_STYLE = 1
# 0: Target structure
# 1: Median Cross Only
# 2: Median Marker with Shaded Cross
# 3: Median Markers
# 4: Shaped Median Markers (all lead times)
# 5: Shaded Median Markers (all lead times)
# 6: Shaded Median Line (all lead times)

SUBSET = 5
# 0: All
# 1: GEOS-CF, GEOS-CF + TROPOMI, GEOS-CF + TROPOMI & Kriging
# 2: GEOS-CF + VIIRS, GEOS-CF + TROPOMI + VIIRS, GEOS-CF + TROPOMI + MET
# 3: GEOS-CF, GEOS-CF + TROPOMI, GEOS-CF + VIIRS, GEOS-CF + TROPOMI + VIIRS
# 4: GEOS-CF, GEOS-CF + TROPOMI, GEOS-CF + TROPOMI + MET, GEOS-CF + TROPOMI & Kriging
# 5: GEOS-CF + TROPOMI, -CF + VIIRS, GEOS-CF + TROPOMI + VIIRS, GEOS-CF + TROPOMI + MET


l_regions = ['New York City','Salt Lake City','New Orleans','Las Vegas','San Francisco']
# l_regions = ['New York City']

d_regional_settings = {'New York City':'2A7L',\
                       'New Orleans':'2A7L',\
                       'Salt Lake City':'2A4L',\
                       'Las Vegas':'1A4L',\
                       'San Francisco':'2A7L'}

# l_modeltypes = ['Persistance Model',\
#                 'Diurnal Model',\
#                 'Prediction Model (GEOS-CF)',\
#                 'Prediction Model (GEOS-CF + TROPOMI)',\
#                 'Prediction Model (GEOS-CF + VPN46A1)',\
#                 'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)',\
#                 'Prediction Model (GEOS-CF + TROPOMI + MET)',\
#                 'Prediction & Kriging Model (GEOS-CF + TROPOMI)']
# l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF','GEOS-CF + TROPOMI','GEOS-CF + VIIRS','GEOS-CF + TROPOMI + VIIRS','GEOS-CF + TROPOMI + MET','GEOS-CF + TROPOMI with Final Updating']
# l_colors = ['black','grey','blue','red','green','cyan','orange','magenta']

if SUBSET == 0:
    l_modeltypes = ['Persistance Model',\
                'Diurnal Model',\
                'Prediction Model (GEOS-CF)',\
                'Prediction Model (GEOS-CF + TROPOMI)',\
                'Prediction & Kriging Model (GEOS-CF + TROPOMI)',\
                'Prediction Model (GEOS-CF + VPN46A1)',\
                'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)',\
                'Prediction Model (GEOS-CF + TROPOMI + MET)']
    l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF','GEOS-CF + TROPOMI','GEOS-CF + TROPOMI & Kriging','GEOS-CF + VIIRS','GEOS-CF + TROPOMI + VIIRS','GEOS-CF + TROPOMI + MET']
    l_colors = ['black','grey','blue','red','magenta','green','cyan','orange']
elif SUBSET == 1:
    l_modeltypes = ['Persistance Model',\
                'Diurnal Model',\
                'Prediction Model (GEOS-CF)',\
                'Prediction Model (GEOS-CF + TROPOMI)',\
                'Prediction & Kriging Model (GEOS-CF + TROPOMI)']
    l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF','GEOS-CF + TROPOMI','GEOS-CF + TROPOMI & Kriging']
    l_colors = ['black','grey','blue','red','magenta']
elif SUBSET == 2:
    l_modeltypes = ['Persistance Model',\
                'Diurnal Model',\
                'Prediction Model (GEOS-CF + VPN46A1)',\
                'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)',\
                'Prediction Model (GEOS-CF + TROPOMI + MET)']
    l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF + VIIRS','GEOS-CF + TROPOMI + VIIRS','GEOS-CF + TROPOMI + MET']
    l_colors = ['black','grey','green','cyan','orange']
elif SUBSET == 3:
    l_modeltypes = ['Persistance Model',\
                'Diurnal Model',\
                'Prediction Model (GEOS-CF)',\
                'Prediction Model (GEOS-CF + TROPOMI)',\
                'Prediction Model (GEOS-CF + VPN46A1)',\
                'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)']
    l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF','GEOS-CF + TROPOMI','GEOS-CF + VIIRS','GEOS-CF + TROPOMI + VIIRS']
    l_colors = ['black','grey','blue','red','green','cyan']
elif SUBSET == 4:
    l_modeltypes = ['Persistance Model',\
                'Diurnal Model',\
                'Prediction Model (GEOS-CF)',\
                'Prediction Model (GEOS-CF + TROPOMI)',\
                'Prediction Model (GEOS-CF + TROPOMI + MET)',\
                'Prediction & Kriging Model (GEOS-CF + TROPOMI)']
    l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF','GEOS-CF + TROPOMI','GEOS-CF + TROPOMI + MET','GEOS-CF + TROPOMI & Kriging']
    l_colors = ['black','grey','blue','red','orange','magenta']
elif SUBSET == 5:
    l_modeltypes = ['Prediction Model (GEOS-CF + TROPOMI)',\
                    'Prediction Model (GEOS-CF + VPN46A1)',\
                    'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)',\
                    'Prediction Model (GEOS-CF + TROPOMI + MET)',\
                    ]
    l_labels = ['GEOS-CF + TROPOMI','GEOS-CF + VIIRS','GEOS-CF + TROPOMI + VIIRS','GEOS-CF + TROPOMI + MET']
    l_colors = ['red','blue','black','orange']
    
d_modeltype_colorpalettes = {'Diurnal Model':'grey',\
                             'Persistance Model':'grey',\
                             'Prediction Model (GEOS-CF)':'blue',\
                             'Prediction Model (GEOS-CF + TROPOMI)':'red',\
                             'Prediction Model (GEOS-CF + VPN46A1)':'green',\
                             'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'cyan',\
                             'Prediction Model (GEOS-CF + TROPOMI + MET)':'orange',\
                             'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'purple'}
    
d_colorpalettes = {'grey':['black','dimgrey','grey','darkgrey','silver','lightgrey'],\
                   'blue':['darkblue','mediumblue','blue','mediumslateblue','lightsteelblue','lavender'],\
                   'red':['darkred','firebrick','red','indianred','lightcoral','pink'],\
                   'orange':['sienna','darkorange','orange','gold','khaki','wheat'],\
                   'green':['darkolivegreen','darkgreen','green','limegreen','darkseagreen','mediumseagreen'],\
                   'purple':['purple','darkviolet','magenta','violet','plum','thistle'],\
                   'cyan':['teal','darkturquoise','cyan','turquoise','paleturquoise','lightcyan']}


l_lead_time = [1,3,6,12,18,24]
# l_lead_time = [3,6,12]     

if PLOT_STYLE >= 4:
    l_lead_time = [1,3,6,12,18,24] 
    l_lead_time_labels = [' 1 hr',' 3 hr',' 6 hr','12 hr','18 hr','24 hr']
    l_lead_time_symbols = ['o','s','*','+','x','.']
    l_lead_time_alphas = [1,0.7,0.55,0.4,0.25,0.1]


if (PLOT_TYPE == 0):
    l_metricsets = [('RMSE','r'),('MAE','AB')]
    
    d_range = {'RMSE':[0,20],\
               'MAE':[0,10],\
               'bias':[-10,10],\
               'AB':[0,10],\
               'r':[0,1]}
elif (PLOT_TYPE == 1):
    l_metricsets = [('MAE','AB'),('RMSE','r')]
    
    d_range = {'RMSE':[20,0],\
               'MAE':[0,10],\
               'bias':[-10,10],\
               'AB':[0,10],\
               'r':[0,1]}
            
    
d_metriclabels = {'RMSE':'RMSE [ppb]',\
                  'MAE':'MAE [ppb]',\
                  'bias':'bias [ppb]',\
                  'AB':'AB [ppb]',\
                  'r':'correlation'}
    
BASIC_FOLDER = os.path.join('C:\\','Data Archive')

OUTPUT_FIGURE_FOLDER = os.path.join(BASIC_FOLDER,'Generated Figures')


#%% Plotting Loop
for s_region in l_regions:
    s_regioncode = s_region.replace(' ','')
    s_spec = d_regional_settings[s_region]
    l_true_data_files = []
    l_prediction_data_files = []
    for i_model in range(len(l_modeltypes)):
        s_modeltype = l_modeltypes[i_model]
        basepath = os.path.join(BASIC_FOLDER,'Generated Data')
        if (s_modeltype == 'Diurnal Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        elif (s_modeltype == 'Persistance Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        else:
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (prediction).nc')]
    if (PLOT_TYPE == 0):
        for lead_time in l_lead_time:
            for metric_x,metric_y in l_metricsets:
                v_range_x = d_range[metric_x]
                v_range_y = d_range[metric_y]
                label_x = d_metriclabels[metric_x]
                label_y = d_metriclabels[metric_y]
                
                FIGURE_TITLE = s_region + ', ' + 'Lead time: '+str(lead_time)+' hr'
                
                OUTPUT_FIGURE_NAME = s_region + ' ' + s_spec + ' ' + str(lead_time) + ' hr' +  ' Comparison '+str(SUBSET)+' - ' + metric_x + ' & ' + metric_y + figure_format
                
                for i_dataset in range(0,len(l_true_data_files)):
                    # Load Data
                    a_data_true = xr.open_dataset(l_true_data_files[i_dataset])
                    a_data_pred = xr.open_dataset(l_prediction_data_files[i_dataset])
                    
                    # Compute Metrics
                    m_site_coords = Get_Locations_with_Data(a_data_true)
                    n_sites = m_site_coords.shape[0]
                    m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
                    
                    v_results_r = np.empty([n_sites,1])
                    v_results_MAE = np.empty([n_sites,1])
                    v_results_RMSE = np.empty([n_sites,1])
                    v_results_bias = np.empty([n_sites,1])
                    v_results_AB = np.empty([n_sites,1])
                    v_results_MAPE = np.empty([n_sites,1])
                    v_results_RMSPE = np.empty([n_sites,1])
                    v_results_Pbias = np.empty([n_sites,1])
                    
                    v_results_r[:] = np.nan
                    v_results_MAE[:] = np.nan
                    v_results_MAPE[:] = np.nan
                    v_results_RMSE[:] = np.nan
                    v_results_RMSPE[:] = np.nan
                    v_results_bias[:] = np.nan
                    v_results_AB[:] = np.nan
                    v_results_Pbias[:] = np.nan
                    
                    if isinstance(lead_time,list):
                        m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time[i_dataset]),m_site_coords)
                    else:
                        m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
                    for site_index in range(n_sites):
                        v_y = m_data_true[:,site_index]
                        v_y_hat = m_data_pred[:,site_index]
                       
                        v_error = v_y - v_y_hat
                        v_Perror = np.abs(v_error/v_y)
                                
                        id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
                    
                        if (sum(id_num) > 24):
                            v_results_r[site_index] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                            v_results_MAE[site_index] = np.nanmean(np.abs(v_error))
                            v_results_RMSE[site_index] = np.sqrt(np.nanmean(v_error**2))
                            v_results_bias[site_index] = np.nanmean(v_error)
                            v_results_AB[site_index] = np.abs(np.nanmean(v_error))
                            v_results_MAPE[site_index] = np.nanmean(np.abs(v_Perror))
                            v_results_RMSPE[site_index] = np.sqrt(np.nanmean(v_Perror**2))
                            v_results_Pbias[site_index] = np.nanmean(v_error)/np.nanmean(v_y)
                    
                    # Identify PoI
                    if (metric_x == 'r'):
                        v_metric_x = v_results_r
                    elif (metric_x == 'RMSE'):
                        v_metric_x = v_results_RMSE
                    elif (metric_x == 'MAE'):
                        v_metric_x = v_results_MAE
                    elif (metric_x == 'bias'):
                        v_metric_x = v_results_bias
                    elif (metric_x == 'AB'):
                        v_metric_x = v_results_AB
                    elif (metric_x == 'MAPE'):
                        v_metric_x = v_results_MAPE
                    elif (metric_x == 'RMSPE'):
                        v_metric_x = v_results_RMSPE
                    elif (metric_x == 'Pbias'):
                        v_metric_x = v_results_Pbias
                        
                    if (metric_y == 'r'):
                        v_metric_y = v_results_r
                    elif (metric_y == 'RMSE'):
                        v_metric_y = v_results_RMSE
                    elif (metric_y == 'MAE'):
                        v_metric_y = v_results_MAE
                    elif (metric_y == 'bias'):
                        v_metric_y = v_results_bias
                    elif (metric_y == 'AB'):
                        v_metric_y = v_results_AB
                    elif (metric_y == 'MAPE'):
                        v_metric_y = v_results_MAPE
                    elif (metric_y == 'RMSPE'):
                        v_metric_y = v_results_RMSPE
                    elif (metric_y == 'Pbias'):
                        v_metric_y = v_results_Pbias
                
                    
                    n_min_x = np.nanmin(v_metric_x)
                    n_25p_x = np.nanpercentile(v_metric_x,25)
                    n_med_x = np.nanmedian(v_metric_x)
                    n_75p_x = np.nanpercentile(v_metric_x,75)
                    n_max_x = np.nanmax(v_metric_x)
                    
                    n_min_y = np.nanmin(v_metric_y)
                    n_25p_y = np.nanpercentile(v_metric_y,25)
                    n_med_y = np.nanmedian(v_metric_y)
                    n_75p_y = np.nanpercentile(v_metric_y,75)
                    n_max_y = np.nanmax(v_metric_y)
                    
                    # sequence_x = [n_25p_x,n_25p_x,n_med_x,n_med_x,n_med_x,n_med_x,n_75p_x,n_75p_x,n_max_x,n_med_x,n_75p_x,n_75p_x,n_med_x,n_med_x,n_med_x,n_med_x,n_25p_x,n_25p_x,n_min_x,n_med_x,n_25p_x]
                    # sexuence_y = [n_med_y,n_75p_y,n_75p_y,n_max_y,n_med_y,n_75p_y,n_75p_y,n_med_y,n_med_y,n_med_y,n_med_y,n_25p_y,n_25p_y,n_min_y,n_med_y,n_25p_y,n_25p_y,n_med_y,n_med_y,n_med_y,n_med_y]
                    
                    # Plot
                    # plt.plot(sequence_x,sexuence_y,color=l_colors[i_dataset],label=l_labels[i_dataset])
                    plt.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    plt.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                    plt.plot([n_75p_x,n_75p_x,n_25p_x,n_25p_x,n_75p_x],[n_75p_y,n_25p_y,n_25p_y,n_75p_y,n_75p_y],color=l_colors[i_dataset])
                
                # Save
                plt.legend()
                plt.xlim(v_range_x)
                plt.ylim(v_range_y)
                plt.xlabel(label_x)
                plt.ylabel(label_y)
                plt.title(FIGURE_TITLE)
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,OUTPUT_FIGURE_NAME),dpi=300)
                plt.show()
                
    elif (PLOT_TYPE == 1) & (PLOT_STYLE < 4):
        for lead_time in l_lead_time:
            fig = plt.figure();
            
            fig.text(0, 0.9, s_region, fontsize=12)
            fig.text(0, 0.85, 'Performance Comparison', fontsize=10)
            fig.text(0, 0.8, 'Lead time: '+str(lead_time)+' hr', fontsize=10)
            
            axes1 = fig.add_axes([0.52, 0.55, 0.45, 0.4]); # upper right
            axes2 = fig.add_axes([0.1, 0.12, 0.4, 0.4]); # lower left
            
            metric_x_1 = l_metricsets[0][0]
            metric_y_1 = l_metricsets[0][1]
            metric_x_2 = l_metricsets[1][0]
            metric_y_2 = l_metricsets[1][1]
            
            v_range_x_1 = d_range[metric_x_1]
            v_range_y_1 = d_range[metric_y_1]
            label_x_1 = d_metriclabels[metric_x_1]
            label_y_1 = d_metriclabels[metric_y_1]
            
            v_range_x_2 = d_range[metric_x_2]
            v_range_y_2 = d_range[metric_y_2]
            label_x_2 = d_metriclabels[metric_x_2]
            label_y_2 = d_metriclabels[metric_y_2]
            
            OUTPUT_FIGURE_NAME = s_region + ' ' + s_spec + ' ' + str(lead_time) + ' hr' +  ' Comparison ' +str(SUBSET)+ '-' + str(PLOT_STYLE) + figure_format
            # OUTPUT_FIGURE_NAME = s_region + ' ' + s_spec + ' Comparison ' +str(SUBSET)+ '-' + str(PLOT_STYLE) + figure_format
        
            for i_dataset in range(0,len(l_true_data_files)):
                # Load Data
                a_data_true = xr.open_dataset(l_true_data_files[i_dataset])
                a_data_pred = xr.open_dataset(l_prediction_data_files[i_dataset])
                
                # Compute Metrics
                m_site_coords = Get_Locations_with_Data(a_data_true)
                n_sites = m_site_coords.shape[0]
                m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
                
                v_results_r = np.empty([n_sites,1])
                v_results_MAE = np.empty([n_sites,1])
                v_results_RMSE = np.empty([n_sites,1])
                v_results_bias = np.empty([n_sites,1])
                v_results_AB = np.empty([n_sites,1])
                v_results_MAPE = np.empty([n_sites,1])
                v_results_RMSPE = np.empty([n_sites,1])
                v_results_Pbias = np.empty([n_sites,1])
                
                v_results_r[:] = np.nan
                v_results_MAE[:] = np.nan
                v_results_MAPE[:] = np.nan
                v_results_RMSE[:] = np.nan
                v_results_RMSPE[:] = np.nan
                v_results_bias[:] = np.nan
                v_results_AB[:] = np.nan
                v_results_Pbias[:] = np.nan
                
                if isinstance(lead_time,list):
                    m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time[i_dataset]),m_site_coords)
                else:
                    m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
                for site_index in range(n_sites):
                    v_y = m_data_true[:,site_index]
                    v_y_hat = m_data_pred[:,site_index]
                   
                    v_error = v_y - v_y_hat
                    v_Perror = np.abs(v_error/v_y)
                            
                    id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
                
                    if (sum(id_num) > 24):
                        v_results_r[site_index] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                        v_results_MAE[site_index] = np.nanmean(np.abs(v_error))
                        v_results_RMSE[site_index] = np.sqrt(np.nanmean(v_error**2))
                        v_results_bias[site_index] = np.nanmean(v_error)
                        v_results_AB[site_index] = np.abs(np.nanmean(v_error))
                        v_results_MAPE[site_index] = np.nanmean(np.abs(v_Perror))
                        v_results_RMSPE[site_index] = np.sqrt(np.nanmean(v_Perror**2))
                        v_results_Pbias[site_index] = np.nanmean(v_error)/np.nanmean(v_y)
                
                # Identify PoI
                metric_x = metric_x_2
                if (metric_x == 'r'):
                    v_metric_x = v_results_r
                elif (metric_x == 'RMSE'):
                    v_metric_x = v_results_RMSE
                elif (metric_x == 'MAE'):
                    v_metric_x = v_results_MAE
                elif (metric_x == 'bias'):
                    v_metric_x = v_results_bias
                elif (metric_x == 'AB'):
                    v_metric_x = v_results_AB
                elif (metric_x == 'MAPE'):
                    v_metric_x = v_results_MAPE
                elif (metric_x == 'RMSPE'):
                    v_metric_x = v_results_RMSPE
                elif (metric_x == 'Pbias'):
                    v_metric_x = v_results_Pbias
                    
                metric_y = metric_y_2    
                if (metric_y == 'r'):
                    v_metric_y = v_results_r
                elif (metric_y == 'RMSE'):
                    v_metric_y = v_results_RMSE
                elif (metric_y == 'MAE'):
                    v_metric_y = v_results_MAE
                elif (metric_y == 'bias'):
                    v_metric_y = v_results_bias
                elif (metric_y == 'AB'):
                    v_metric_y = v_results_AB
                elif (metric_y == 'MAPE'):
                    v_metric_y = v_results_MAPE
                elif (metric_y == 'RMSPE'):
                    v_metric_y = v_results_RMSPE
                elif (metric_y == 'Pbias'):
                    v_metric_y = v_results_Pbias
            
                
                n_min_x = np.nanmin(v_metric_x)
                n_25p_x = np.nanpercentile(v_metric_x,25)
                n_med_x = np.nanmedian(v_metric_x)
                n_75p_x = np.nanpercentile(v_metric_x,75)
                n_max_x = np.nanmax(v_metric_x)
                
                n_min_y = np.nanmin(v_metric_y)
                n_25p_y = np.nanpercentile(v_metric_y,25)
                n_med_y = np.nanmedian(v_metric_y)
                n_75p_y = np.nanpercentile(v_metric_y,75)
                n_max_y = np.nanmax(v_metric_y)
                
                # Plot
                if PLOT_STYLE == 0:
                    axes2.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes2.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                    axes2.plot([n_75p_x,n_75p_x,n_25p_x,n_25p_x,n_75p_x],[n_75p_y,n_25p_y,n_25p_y,n_75p_y,n_75p_y],color=l_colors[i_dataset])
                elif PLOT_STYLE == 1.5:
                    axes2.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes2.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                elif PLOT_STYLE == 1:
                    axes2.plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes2.plot([n_25p_x,n_75p_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                elif PLOT_STYLE == 2:
                    axes2.plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes2.plot([n_25p_x,n_75p_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                    axes2.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],alpha=0.5)
                    axes2.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset],alpha=0.5)
                elif PLOT_STYLE == 3:
                    axes2.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],label=l_labels[i_dataset],marker='o')
                    
                # Identify PoI
                metric_x = metric_x_1
                if (metric_x == 'r'):
                    v_metric_x = v_results_r
                elif (metric_x == 'RMSE'):
                    v_metric_x = v_results_RMSE
                elif (metric_x == 'MAE'):
                    v_metric_x = v_results_MAE
                elif (metric_x == 'bias'):
                    v_metric_x = v_results_bias
                elif (metric_x == 'AB'):
                    v_metric_x = v_results_AB
                elif (metric_x == 'MAPE'):
                    v_metric_x = v_results_MAPE
                elif (metric_x == 'RMSPE'):
                    v_metric_x = v_results_RMSPE
                elif (metric_x == 'Pbias'):
                    v_metric_x = v_results_Pbias
                    
                metric_y = metric_y_1    
                if (metric_y == 'r'):
                    v_metric_y = v_results_r
                elif (metric_y == 'RMSE'):
                    v_metric_y = v_results_RMSE
                elif (metric_y == 'MAE'):
                    v_metric_y = v_results_MAE
                elif (metric_y == 'bias'):
                    v_metric_y = v_results_bias
                elif (metric_y == 'AB'):
                    v_metric_y = v_results_AB
                elif (metric_y == 'MAPE'):
                    v_metric_y = v_results_MAPE
                elif (metric_y == 'RMSPE'):
                    v_metric_y = v_results_RMSPE
                elif (metric_y == 'Pbias'):
                    v_metric_y = v_results_Pbias
            
                
                n_min_x = np.nanmin(v_metric_x)
                n_25p_x = np.nanpercentile(v_metric_x,25)
                n_med_x = np.nanmedian(v_metric_x)
                n_75p_x = np.nanpercentile(v_metric_x,75)
                n_max_x = np.nanmax(v_metric_x)      
                
                n_min_y = np.nanmin(v_metric_y)
                n_25p_y = np.nanpercentile(v_metric_y,25)
                n_med_y = np.nanmedian(v_metric_y)
                n_75p_y = np.nanpercentile(v_metric_y,75)
                n_max_y = np.nanmax(v_metric_y)
                
                # Plot
                if PLOT_STYLE == 0:
                    axes1.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes1.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                    axes1.plot([n_75p_x,n_75p_x,n_25p_x,n_25p_x,n_75p_x],[n_75p_y,n_25p_y,n_25p_y,n_75p_y,n_75p_y],color=l_colors[i_dataset])
                elif PLOT_STYLE == 1.5:
                    axes1.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes1.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                elif PLOT_STYLE == 1:
                    axes1.plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes1.plot([n_25p_x,n_75p_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                elif PLOT_STYLE == 2:
                    axes1.plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=l_colors[i_dataset],label=l_labels[i_dataset])
                    axes1.plot([n_25p_x,n_75p_x],[n_med_y,n_med_y],color=l_colors[i_dataset])
                    axes1.plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=l_colors[i_dataset],alpha=0.5)
                    axes1.plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=l_colors[i_dataset],alpha=0.5)
                elif PLOT_STYLE == 3:
                    axes1.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],label=l_labels[i_dataset],marker='o')
                
            # Upper Right
            axes1.text(0.1,0.2,u"\u2b0b best",fontsize=10);
                
            axes1.set_xlabel(label_x_1);
            axes1.set_ylabel(label_y_1);
            axes1.legend(title = 'Method', bbox_to_anchor=(1.0, -0.25), prop={'size': 8});
            axes1.set_ylim(v_range_y_1);
            axes1.set_xlim(v_range_x_1);
            
            # Lower Left
            axes2.text(4,0.92,u"best \u2b08",fontsize=10);
            
            axes2.set_xlabel(label_x_2);
            axes2.set_ylabel(label_y_2);
            axes2.set_ylim(v_range_y_2);
            axes2.set_xlim(v_range_x_2);
        
            
            # Save
            fig.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,OUTPUT_FIGURE_NAME),dpi=300)
            fig.show()
    elif (PLOT_TYPE == 1) & (PLOT_STYLE >= 4):
        fig = plt.figure();
        
        fig.text(0, 0.9, s_region, fontsize=12)
        # fig.text(0, 0.85, 'Performance Comparison', fontsize=10)
        
        axes1 = fig.add_axes([0.52, 0.55, 0.45, 0.4]); # upper right
        axes2 = fig.add_axes([0.1, 0.12, 0.4, 0.4]); # lower left
        
        metric_x_1 = l_metricsets[0][0]
        metric_y_1 = l_metricsets[0][1]
        metric_x_2 = l_metricsets[1][0]
        metric_y_2 = l_metricsets[1][1]
        
        v_range_x_1 = d_range[metric_x_1]
        v_range_y_1 = d_range[metric_y_1]
        label_x_1 = d_metriclabels[metric_x_1]
        label_y_1 = d_metriclabels[metric_y_1]
        
        v_range_x_2 = d_range[metric_x_2]
        v_range_y_2 = d_range[metric_y_2]
        label_x_2 = d_metriclabels[metric_x_2]
        label_y_2 = d_metriclabels[metric_y_2]
        
        OUTPUT_FIGURE_NAME = s_region + ' ' + s_spec + ' Comparison ' +str(SUBSET)+ '-' + str(PLOT_STYLE) + figure_format
    
        for i_dataset in range(0,len(l_true_data_files)):
            # Load Data
            a_data_true = xr.open_dataset(l_true_data_files[i_dataset])
            a_data_pred = xr.open_dataset(l_prediction_data_files[i_dataset])
            
            
            m_site_coords = Get_Locations_with_Data(a_data_true)
            n_sites = m_site_coords.shape[0]
            m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
            
            x1_save = np.nan
            y1_save = np.nan
            x2_save = np.nan
            y2_save = np.nan
            
            for id_lead in range(len(l_lead_time)):
                lead_time = l_lead_time[id_lead]
                
                # Compute Metrics
                v_results_r = np.empty([n_sites,1])
                v_results_MAE = np.empty([n_sites,1])
                v_results_RMSE = np.empty([n_sites,1])
                v_results_bias = np.empty([n_sites,1])
                v_results_AB = np.empty([n_sites,1])
                v_results_MAPE = np.empty([n_sites,1])
                v_results_RMSPE = np.empty([n_sites,1])
                v_results_Pbias = np.empty([n_sites,1])
                
                v_results_r[:] = np.nan
                v_results_MAE[:] = np.nan
                v_results_MAPE[:] = np.nan
                v_results_RMSE[:] = np.nan
                v_results_RMSPE[:] = np.nan
                v_results_bias[:] = np.nan
                v_results_AB[:] = np.nan
                v_results_Pbias[:] = np.nan
                
                m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
                for site_index in range(n_sites):
                    v_y = m_data_true[:,site_index]
                    v_y_hat = m_data_pred[:,site_index]
                   
                    v_error = v_y - v_y_hat
                    v_Perror = np.abs(v_error/v_y)
                            
                    id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
                
                    if (sum(id_num) > 24):
                        v_results_r[site_index] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                        v_results_MAE[site_index] = np.nanmean(np.abs(v_error))
                        v_results_RMSE[site_index] = np.sqrt(np.nanmean(v_error**2))
                        v_results_bias[site_index] = np.nanmean(v_error)
                        v_results_AB[site_index] = np.abs(np.nanmean(v_error))
                        v_results_MAPE[site_index] = np.nanmean(np.abs(v_Perror))
                        v_results_RMSPE[site_index] = np.sqrt(np.nanmean(v_Perror**2))
                        v_results_Pbias[site_index] = np.nanmean(v_error)/np.nanmean(v_y)
            
                # Identify PoI
                metric_x = metric_x_2
                if (metric_x == 'r'):
                    v_metric_x = v_results_r
                elif (metric_x == 'RMSE'):
                    v_metric_x = v_results_RMSE
                elif (metric_x == 'MAE'):
                    v_metric_x = v_results_MAE
                elif (metric_x == 'bias'):
                    v_metric_x = v_results_bias
                elif (metric_x == 'AB'):
                    v_metric_x = v_results_AB
                elif (metric_x == 'MAPE'):
                    v_metric_x = v_results_MAPE
                elif (metric_x == 'RMSPE'):
                    v_metric_x = v_results_RMSPE
                elif (metric_x == 'Pbias'):
                    v_metric_x = v_results_Pbias
                    
                metric_y = metric_y_2    
                if (metric_y == 'r'):
                    v_metric_y = v_results_r
                elif (metric_y == 'RMSE'):
                    v_metric_y = v_results_RMSE
                elif (metric_y == 'MAE'):
                    v_metric_y = v_results_MAE
                elif (metric_y == 'bias'):
                    v_metric_y = v_results_bias
                elif (metric_y == 'AB'):
                    v_metric_y = v_results_AB
                elif (metric_y == 'MAPE'):
                    v_metric_y = v_results_MAPE
                elif (metric_y == 'RMSPE'):
                    v_metric_y = v_results_RMSPE
                elif (metric_y == 'Pbias'):
                    v_metric_y = v_results_Pbias
            
                
                n_min_x = np.nanmin(v_metric_x)
                n_25p_x = np.nanpercentile(v_metric_x,25)
                n_med_x = np.nanmedian(v_metric_x)
                n_75p_x = np.nanpercentile(v_metric_x,75)
                n_max_x = np.nanmax(v_metric_x)
                
                n_min_y = np.nanmin(v_metric_y)
                n_25p_y = np.nanpercentile(v_metric_y,25)
                n_med_y = np.nanmedian(v_metric_y)
                n_75p_y = np.nanpercentile(v_metric_y,75)
                n_max_y = np.nanmax(v_metric_y)
                
                # Plot
                if PLOT_STYLE == 4:
                    if i_dataset == 0:
                        axes2.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],label=l_lead_time_labels[id_lead],marker=l_lead_time_symbols[id_lead])
                    else:
                        axes2.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],marker=l_lead_time_symbols[id_lead])
                elif PLOT_STYLE == 5:
                    if i_dataset == 0:
                        axes2.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],label=l_lead_time_labels[id_lead],alpha=l_lead_time_alphas[id_lead],marker='o')
                    else:
                        axes2.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead],marker='o')
                elif PLOT_STYLE == 6:
                    if ~np.isnan(x2_save):
                        x_half = (x2_save+n_med_x)/2
                        y_half = (y2_save+n_med_y)/2
                        if i_dataset == 0:
                            if id_lead == 1:
                                axes2.plot([x2_save,x_half],[y2_save,y_half],color=l_colors[i_dataset],label=l_lead_time_labels[0],alpha=l_lead_time_alphas[id_lead-1])
                            else:
                                axes2.plot([x2_save,x_half],[y2_save,y_half],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead-1])
                            axes2.plot([n_med_x,x_half],[n_med_y,y_half],color=l_colors[i_dataset],label=l_lead_time_labels[id_lead],alpha=l_lead_time_alphas[id_lead])
                        else:
                            axes2.plot([x2_save,x_half],[y2_save,y_half],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead-1])
                            axes2.plot([n_med_x,x_half],[n_med_y,y_half],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead]) 
                    x2_save = n_med_x
                    y2_save = n_med_y
                
                
                # Identify PoI
                metric_x = metric_x_1
                if (metric_x == 'r'):
                    v_metric_x = v_results_r
                elif (metric_x == 'RMSE'):
                    v_metric_x = v_results_RMSE
                elif (metric_x == 'MAE'):
                    v_metric_x = v_results_MAE
                elif (metric_x == 'bias'):
                    v_metric_x = v_results_bias
                elif (metric_x == 'AB'):
                    v_metric_x = v_results_AB
                elif (metric_x == 'MAPE'):
                    v_metric_x = v_results_MAPE
                elif (metric_x == 'RMSPE'):
                    v_metric_x = v_results_RMSPE
                elif (metric_x == 'Pbias'):
                    v_metric_x = v_results_Pbias
                    
                metric_y = metric_y_1    
                if (metric_y == 'r'):
                    v_metric_y = v_results_r
                elif (metric_y == 'RMSE'):
                    v_metric_y = v_results_RMSE
                elif (metric_y == 'MAE'):
                    v_metric_y = v_results_MAE
                elif (metric_y == 'bias'):
                    v_metric_y = v_results_bias
                elif (metric_y == 'AB'):
                    v_metric_y = v_results_AB
                elif (metric_y == 'MAPE'):
                    v_metric_y = v_results_MAPE
                elif (metric_y == 'RMSPE'):
                    v_metric_y = v_results_RMSPE
                elif (metric_y == 'Pbias'):
                    v_metric_y = v_results_Pbias
            
                
                n_min_x = np.nanmin(v_metric_x)
                n_25p_x = np.nanpercentile(v_metric_x,25)
                n_med_x = np.nanmedian(v_metric_x)
                n_75p_x = np.nanpercentile(v_metric_x,75)
                n_max_x = np.nanmax(v_metric_x)      
                
                n_min_y = np.nanmin(v_metric_y)
                n_25p_y = np.nanpercentile(v_metric_y,25)
                n_med_y = np.nanmedian(v_metric_y)
                n_75p_y = np.nanpercentile(v_metric_y,75)
                n_max_y = np.nanmax(v_metric_y)
                
                # Plot
                if PLOT_STYLE == 4:
                    if id_lead == 0:
                        axes1.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],label=l_labels[i_dataset],marker=l_lead_time_symbols[id_lead])
                    else:
                        axes1.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],marker=l_lead_time_symbols[id_lead])
                elif PLOT_STYLE == 5:
                    if id_lead == 0:
                        axes1.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],label=l_labels[i_dataset],alpha=l_lead_time_alphas[id_lead],marker='o')
                    else:
                        axes1.scatter([n_med_x],[n_med_y],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead],marker='o')
                elif PLOT_STYLE == 6:
                    if ~np.isnan(x1_save):
                        x_half = (x1_save+n_med_x)/2
                        y_half = (y1_save+n_med_y)/2
                        if id_lead == 1:
                            axes1.plot([x1_save,x_half],[y1_save,y_half],color=l_colors[i_dataset],label=l_labels[i_dataset],alpha=l_lead_time_alphas[id_lead-1])
                        else:
                            axes1.plot([x1_save,x_half],[y1_save,y_half],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead-1])
                        axes1.plot([n_med_x,x_half],[n_med_y,y_half],color=l_colors[i_dataset],alpha=l_lead_time_alphas[id_lead])
                    x1_save = n_med_x
                    y1_save = n_med_y
                
                
        # Upper Right
        axes1.text(0.1,0.2,u"\u2b0b best",fontsize=10);
            
        axes1.set_xlabel(label_x_1);
        axes1.set_ylabel(label_y_1);
        axes1.legend(title = 'Method', bbox_to_anchor=(1.0, -0.25), prop={'size': 8});
        axes1.set_ylim(v_range_y_1);
        axes1.set_xlim(v_range_x_1);
        
        # Lower Left
        axes2.text(4,0.92,u"best \u2b08",fontsize=10);
        
        axes2.set_xlabel(label_x_2);
        axes2.set_ylabel(label_y_2);
        axes2.legend(title = 'Lead Time', bbox_to_anchor=(0.5, 1.9), prop={'size': 8});
        axes2.set_ylim(v_range_y_2);
        axes2.set_xlim(v_range_x_2);
    
        
        # Save
        fig.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,OUTPUT_FIGURE_NAME),dpi=300)
        fig.show()



    
    
    
    