#%% Load Packages & Define Functions
import os
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt

def Get_Locations_with_Data(a_data):
    m_data = (~np.isnan(a_data)).any(dim='time')
    m_coords = np.empty([0,2])
    for i_lat in m_data.lat:
        for i_lon in m_data.lon:
            if m_data.sel(lat=i_lat,lon=i_lon):
                m_coords = np.r_[m_coords,np.array([i_lat,i_lon],ndmin=2)]          
    return m_coords

def Array_to_Matrix_at_Coords(a_data,m_coords):
    n_coords = m_coords.shape[0]
    if hasattr(a_data,'time'):
        n_times = len(a_data.time)
    else:
        n_times = 1
    m_data = np.empty([n_times,n_coords])
    m_data[:] = np.nan
    for i_coord in range(n_coords):
        m_data[:,i_coord] = np.array(a_data.to_array().sel(lat=m_coords[i_coord,0],lon=m_coords[i_coord,1]))
    return m_data

#%% Basic Settings
        
figure_format = '.svg'

BASIC_FOLDER = os.path.join('C:\\','Data Archive')

OUTPUT_FIGURE_FOLDER = os.path.join(BASIC_FOLDER,'Generated Figures')

PLOT_STYLE = 1
# 0: Target structure
# 1: Median Cross Only
# 2: Median Marker with Shaded Cross
# 3: Median Markers


SUBSET = 1
# 1: GEOS-CF + TROPOMI v. Climatology Baseline
# 2: GEOS-CF + TROPOMI v. GEOS-CF
    
    
    

l_regions = ['Las Vegas','New Orleans','New York City','Salt Lake City','San Francisco']


d_regional_colors = {'New York City':'blue',\
                       'New Orleans':'cyan',\
                       'Salt Lake City':'orange',\
                       'Las Vegas':'black',\
                       'San Francisco':'red'}
    
d_regional_settings = {'New York City':'2A7L',\
                       'New Orleans':'2A7L',\
                       'Salt Lake City':'2A4L',\
                       'Las Vegas':'1A4L',\
                       'San Francisco':'2A7L'}
    
if SUBSET == 1:
    l_modeltypes = ['Diurnal Model','Prediction Model (GEOS-CF + TROPOMI)']
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                  'Diurnal Model':'Climatology Baseline',\
                  'Prediction Model (GEOS-CF)':'GEOS-CF',\
                  'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                  'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                  'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                  'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                  'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
elif SUBSET == 2:
    l_modeltypes = ['Prediction Model (GEOS-CF)','Prediction Model (GEOS-CF + TROPOMI)']
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                  'Diurnal Model':'Climatology Baseline',\
                  'Prediction Model (GEOS-CF)':'GEOS-CF',\
                  'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                  'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                  'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                  'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                  'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
    
# l_modeltypes = ['Persistance Model',\
#                 'Diurnal Model',\
#                 'Prediction Model (GEOS-CF)',\
#                 'Prediction Model (GEOS-CF + TROPOMI)',\
#                 'Prediction Model (GEOS-CF + VPN46A1)',\
#                 'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)',\
#                 'Prediction Model (GEOS-CF + TROPOMI + MET)',\
#                 'Prediction & Kriging Model (GEOS-CF + TROPOMI)']
# l_labels = ['Persistence Baseline','Climatology Baseline','GEOS-CF','GEOS-CF + TROPOMI','GEOS-CF + VIIRS','GEOS-CF + TROPOMI + VIIRS','GEOS-CF + TROPOMI + MET','GEOS-CF + TROPOMI with Final Updating']

    
l_lead_time = [1,3,6,12,18,24]
l_lead_time_labels = [' 1 hr',' 3 hr',' 6 hr','12 hr','18 hr','24 hr']

l_metricsets = [('MAE','AB'),('RMSE','r')]

d_range = {'RMSE':[20,0],\
           'MAE':[0,15],\
           'bias':[-10,10],\
           'AB':[0,10],\
           'r':[0,1]}
            
d_metriclabels = {'RMSE':'RMSE [ppb]',\
                  'MAE':'MAE [ppb]',\
                  'bias':'bias [ppb]',\
                  'AB':'AB [ppb]',\
                  'r':'correlation'}
    
#%% Calculate Metrics:
a_metrics = np.zeros([5,2,len(l_metricsets),len(l_modeltypes),len(l_regions),len(l_lead_time)])
a_metrics[:] = np.nan 

for i_region in range(len(l_regions)):
    s_region = l_regions[i_region]
    s_regioncode = s_region.replace(' ','')
    s_spec = d_regional_settings[s_region]
    l_true_data_files = []
    l_prediction_data_files = []
    for i_model in range(len(l_modeltypes)):
        s_modeltype = l_modeltypes[i_model]
        basepath = os.path.join(BASIC_FOLDER,'Generated Data')
        if (s_modeltype == 'Diurnal Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        elif (s_modeltype == 'Persistance Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        else:
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (prediction).nc')]
       
    for i_dataset in range(0,len(l_true_data_files)):
        # Load Data
        a_data_true = xr.open_dataset(l_true_data_files[i_dataset])
        a_data_pred = xr.open_dataset(l_prediction_data_files[i_dataset])
        
        # Compute Metrics
        m_site_coords = Get_Locations_with_Data(a_data_true)
        n_sites = m_site_coords.shape[0]
        m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
        
        v_results_r = np.empty([n_sites,1])
        v_results_MAE = np.empty([n_sites,1])
        v_results_RMSE = np.empty([n_sites,1])
        v_results_bias = np.empty([n_sites,1])
        v_results_AB = np.empty([n_sites,1])
        v_results_MAPE = np.empty([n_sites,1])
        v_results_RMSPE = np.empty([n_sites,1])
        v_results_Pbias = np.empty([n_sites,1])
        
        v_results_r[:] = np.nan
        v_results_MAE[:] = np.nan
        v_results_MAPE[:] = np.nan
        v_results_RMSE[:] = np.nan
        v_results_RMSPE[:] = np.nan
        v_results_bias[:] = np.nan
        v_results_AB[:] = np.nan
        v_results_Pbias[:] = np.nan
        
        for i_lead_time in range(len(l_lead_time)):
            lead_time = l_lead_time[i_lead_time]
            m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
            
            for site_index in range(n_sites):
                v_y = m_data_true[:,site_index]
                v_y_hat = m_data_pred[:,site_index]
               
                v_error = v_y - v_y_hat
                v_Perror = np.abs(v_error/v_y)
                        
                id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
            
                if (sum(id_num) > 24):
                    v_results_r[site_index] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                    v_results_MAE[site_index] = np.nanmean(np.abs(v_error))
                    v_results_RMSE[site_index] = np.sqrt(np.nanmean(v_error**2))
                    v_results_bias[site_index] = np.nanmean(v_error)
                    v_results_AB[site_index] = np.abs(np.nanmean(v_error))
                    v_results_MAPE[site_index] = np.nanmean(np.abs(v_Perror))
                    v_results_RMSPE[site_index] = np.sqrt(np.nanmean(v_Perror**2))
                    v_results_Pbias[site_index] = np.nanmean(v_error)/np.nanmean(v_y)
                    
            for i_metricset in range(len(l_metricsets)):
                for i_metric in range(len(l_metricsets[i_metricset])):
                    metric = l_metricsets[i_metricset][i_metric]
                    if (metric == 'r'):
                        v_metric = v_results_r
                    elif (metric == 'RMSE'):
                        v_metric = v_results_RMSE
                    elif (metric == 'MAE'):
                        v_metric = v_results_MAE
                    elif (metric == 'bias'):
                        v_metric = v_results_bias
                    elif (metric == 'AB'):
                        v_metric = v_results_AB
                    elif (metric == 'MAPE'):
                        v_metric = v_results_MAPE
                    elif (metric == 'RMSPE'):
                        v_metric = v_results_RMSPE
                    elif (metric == 'Pbias'):
                        v_metric = v_results_Pbias

                    n_min = np.nanmin(v_metric)
                    n_25p = np.nanpercentile(v_metric,25)
                    n_med = np.nanmedian(v_metric)
                    n_75p = np.nanpercentile(v_metric,75)
                    n_max = np.nanmax(v_metric)
                    
                    a_metrics[0,i_metric,i_metricset,i_dataset,i_region,i_lead_time] = n_med
                    a_metrics[1,i_metric,i_metricset,i_dataset,i_region,i_lead_time] = n_min
                    a_metrics[2,i_metric,i_metricset,i_dataset,i_region,i_lead_time] = n_max
                    a_metrics[3,i_metric,i_metricset,i_dataset,i_region,i_lead_time] = n_25p 
                    a_metrics[4,i_metric,i_metricset,i_dataset,i_region,i_lead_time] = n_75p

#%% Plot Results
for i_lead_time in range(len(l_lead_time)):
    
    lead_time = l_lead_time[i_lead_time]
    
    fig = plt.figure(figsize=(6,5));
            
    # fig.text(0, 0.9, s_region, fontsize=12)
    # fig.text(0, 0.85, 'Performance Comparison', fontsize=10)
    fig.text(0.1, 0.9, 'Lead time: '+str(lead_time)+' hr', fontsize=10)
    
    fig.text(  0, 0.52, 'a)', fontsize=12)
    fig.text(0.4, 0.95, 'b)', fontsize=12)
    
    axes = [\
            # upper right
            fig.add_axes([0.52, 0.55, 0.45, 0.4]),\
            # lower left
            fig.add_axes([0.1, 0.12, 0.4, 0.4]),\
            ]
    
    for i_metricset in range(len(l_metricsets)):
        metric_x = l_metricsets[i_metricset][0]
        metric_y = l_metricsets[i_metricset][1]
        v_range_x = d_range[metric_x]
        v_range_y = d_range[metric_y]
        label_x = d_metriclabels[metric_x]
        label_y = d_metriclabels[metric_y]
        
        
        
        for i_region in range(len(l_regions)):
            s_region = l_regions[i_region]
            for i_model in range(len(l_modeltypes)):
                method = l_modeltypes[i_model]
                
                n_med_x = a_metrics[0,0,i_metricset,i_model,i_region,i_lead_time]
                n_min_x = a_metrics[1,0,i_metricset,i_model,i_region,i_lead_time]
                n_max_x = a_metrics[2,0,i_metricset,i_model,i_region,i_lead_time]
                n_25p_x = a_metrics[3,0,i_metricset,i_model,i_region,i_lead_time]
                n_75p_x = a_metrics[4,0,i_metricset,i_model,i_region,i_lead_time]
                
                n_med_y = a_metrics[0,1,i_metricset,i_model,i_region,i_lead_time]
                n_min_y = a_metrics[1,1,i_metricset,i_model,i_region,i_lead_time]
                n_max_y = a_metrics[2,1,i_metricset,i_model,i_region,i_lead_time]
                n_25p_y = a_metrics[3,1,i_metricset,i_model,i_region,i_lead_time]
                n_75p_y = a_metrics[4,1,i_metricset,i_model,i_region,i_lead_time]
                
                # Plot
                if PLOT_STYLE == 0:
                    if i_metricset == 0:
                        if (i_model == 1):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=s_region)
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    elif i_metricset == 1:
                        if (i_region == 0):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=d_modeltype_labels[method])
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    axes[i_metricset].plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    axes[i_metricset].plot([n_75p_x,n_75p_x,n_25p_x,n_25p_x,n_75p_x],[n_75p_y,n_25p_y,n_25p_y,n_75p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                elif PLOT_STYLE == 1.5:
                    if i_metricset == 0:
                        if (i_model == 1):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=s_region)
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    elif i_metricset == 1:
                        if (i_region == 0):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=d_modeltype_labels[method])
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    axes[i_metricset].plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                elif PLOT_STYLE == 1:
                    if i_metricset == 0:
                        if (i_model == 1):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=s_region)
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    elif i_metricset == 1:
                        if (i_region == 0):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=d_modeltype_labels[method])
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    axes[i_metricset].plot([n_25p_x,n_75p_x],[n_med_y,n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                elif PLOT_STYLE == 2:
                    if i_metricset == 0:
                        if (i_model == 1):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=s_region)
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    elif i_metricset == 1:
                        if (i_region == 0):
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=d_modeltype_labels[method])
                        else:
                            axes[i_metricset].plot([n_med_x,n_med_x],[n_25p_y,n_75p_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    axes[i_metricset].plot([n_25p_x,n_75p_x],[n_med_y,n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method])
                    axes[i_metricset].plot([n_med_x,n_med_x],[n_min_y,n_max_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],alpha=0.5)
                    axes[i_metricset].plot([n_min_x,n_max_x],[n_med_y,n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],alpha=0.5)
                elif PLOT_STYLE == 3:
                    if i_metricset == 0:
                        if (i_model == 1):
                            axes[i_metricset].scatter([n_med_x],[n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=s_region,marker='o')
                        else:
                            axes[i_metricset].scatter([n_med_x],[n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],marker='o')
                    elif i_metricset == 1:
                        if (i_region == 0):
                            axes[i_metricset].scatter([n_med_x],[n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],label=d_modeltype_labels[method],marker='o')
                        else:
                            axes[i_metricset].scatter([n_med_x],[n_med_y],color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[method],marker='o')
                 
        axes[i_metricset].set_xlabel(label_x);
        axes[i_metricset].set_ylabel(label_y);
        axes[i_metricset].set_xlim(v_range_x);
        axes[i_metricset].set_ylim(v_range_y);
        
    # Upper Right
    axes[0].text(0.1,0.2,u"\u2b0b best",fontsize=10);  
    axes[0].legend(title = 'Area', bbox_to_anchor=(0.5, -0.3), prop={'size': 8});

    # Lower Left
    axes[1].text(4,0.92,u"best \u2b08",fontsize=10);
    axes[1].legend(title = 'Method', bbox_to_anchor=(0.7, 1.5), prop={'size': 8});

    # Save
    OUTPUT_FIGURE_NAME = 'Multiregion ' + str(lead_time) + ' hr' +  ' Comparison ' + str(SUBSET) + figure_format
    fig.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,OUTPUT_FIGURE_NAME),dpi=300)
    fig.show()
    
       
