#%% Load Packages & Define Functions
import os
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt

def Get_Locations_with_Data(a_data):
    m_data = (~np.isnan(a_data)).any(dim='time')
    m_coords = np.empty([0,2])
    for i_lat in m_data.lat:
        for i_lon in m_data.lon:
            if m_data.sel(lat=i_lat,lon=i_lon):
                m_coords = np.r_[m_coords,np.array([i_lat,i_lon],ndmin=2)]          
    return m_coords

def Array_to_Matrix_at_Coords(a_data,m_coords):
    n_coords = m_coords.shape[0]
    if hasattr(a_data,'time'):
        n_times = len(a_data.time)
    else:
        n_times = 1
    m_data = np.empty([n_times,n_coords])
    m_data[:] = np.nan
    for i_coord in range(n_coords):
        m_data[:,i_coord] = np.array(a_data.to_array().sel(lat=m_coords[i_coord,0],lon=m_coords[i_coord,1]))
    return m_data

#%% Basic Settings
        
figure_format = '.svg'

BASIC_FOLDER = os.path.join('C:\\','Data Archive')

OUTPUT_FIGURE_FOLDER = os.path.join(BASIC_FOLDER,'Generated Figures')

# l_regions = ['New York City','Salt Lake City','New Orleans','Las Vegas','San Francisco']
l_regions = ['Las Vegas','New Orleans','New York City','Salt Lake City','San Francisco']
# l_modeltypes = ['Persistance Model',\
#                 'Diurnal Model',\
#                 'Prediction Model (GEOS-CF)',\
#                 'Prediction Model (GEOS-CF + TROPOMI)',\
#                 'Prediction & Kriging Model (GEOS-CF + TROPOMI)',\
#                 'Prediction Model (GEOS-CF + VPN46A1)',\
#                 'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)',\
#                 'Prediction Model (GEOS-CF + TROPOMI + MET)']

SUBSET = 2

if SUBSET == 1:
    l_modeltypes = ['Persistance Model','Prediction Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
elif SUBSET == 2:
    l_modeltypes = ['Prediction Model (GEOS-CF + TROPOMI)','Prediction & Kriging Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'without updating',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'with updating',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':':',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
elif SUBSET == 3:
    l_modeltypes = ['Prediction Model (GEOS-CF)','Prediction Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'} 
    
# l_lead_time = [1,3,6,12,18,24]
# l_lead_time_labels = [' 1 hr',' 3 hr',' 6 hr','12 hr','18 hr','24 hr']

l_lead_time = [0,1,3,6,9,12,15,18,21,24]
l_lead_time_labels = [' 0 hr',' 1 hr',' 3 hr',' 6 hr',' 9 hr','12 hr','15 hr','18 hr','21 hr','24 hr']

# d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
#                       'Diurnal Model':'Climatology Baseline',\
#                       'Prediction Model (GEOS-CF)':'GEOS-CF',\
#                       'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
#                       'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
#                       'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
#                       'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
#                       'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}

# d_modeltype_styles = {'Persistance Model':':',\
#                       'Diurnal Model':':',\
#                       'Prediction Model (GEOS-CF)':'-',\
#                       'Prediction Model (GEOS-CF + TROPOMI)':'-',\
#                       'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
#                       'Prediction Model (GEOS-CF + VPN46A1)':'-',\
#                       'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
#                       'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}

d_regional_colors = {'New York City':'blue',\
                       'New Orleans':'cyan',\
                       'Salt Lake City':'orange',\
                       'Las Vegas':'black',\
                       'San Francisco':'red'}
    
d_regional_settings = {'New York City':'2A7L',\
                       'New Orleans':'2A7L',\
                       'Salt Lake City':'2A4L',\
                       'Las Vegas':'1A4L',\
                       'San Francisco':'2A7L'}

# l_metricsets = ['r','AB','RMSE','MAE']
l_metricsets = ['r','RMSE']

if SUBSET == 1:
    d_metricrange = {'RMSE':[0,20],\
               'MAE':[0,20],\
               'bias':[-10,10],\
               'AB':[0,10],\
               'r':[-0.2,1]}
elif SUBSET == 2:
    d_metricrange = {'RMSE':[0,12],\
               'MAE':[0,10],\
               'bias':[-10,10],\
               'AB':[0,6],\
               'r':[0,1]}
elif SUBSET == 3:
    d_metricrange = {'RMSE':[0,12],\
               'MAE':[0,10],\
               'bias':[-10,10],\
               'AB':[0,4],\
               'r':[-1,1]}
        
d_metriclabels = {'RMSE':'RMSE [ppb]',\
                  'MAE':'MAE [ppb]',\
                  'bias':'bias [ppb]',\
                  'AB':'AB [ppb]',\
                  'r':'correlation'}
    
if len(l_metricsets) == 4:
    if SUBSET == 1:
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.15,-0.4]
    elif SUBSET == 2:
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.0,-0.4]
    elif SUBSET == 3:
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.14,-0.4]

if len(l_metricsets) == 2:
    if SUBSET == 1:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 2:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 3:
        legend_1_pos = [3.4,1]
        legend_2_pos = [2.14,0.5]
    
#%% Compute Metrics
a_metrics = np.zeros([len(l_lead_time),len(l_modeltypes),len(l_regions),len(l_metricsets)])
a_metrics[:] = np.nan

for i_region in range(len(l_regions)):
    s_region = l_regions[i_region]
    s_regioncode = s_region.replace(' ','')
    s_spec = d_regional_settings[s_region]
    l_true_data_files = []
    l_prediction_data_files = []
    for i_model in range(len(l_modeltypes)):
        s_modeltype = l_modeltypes[i_model]
        basepath = os.path.join(BASIC_FOLDER,'Generated Data')
        if (s_modeltype == 'Diurnal Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        elif (s_modeltype == 'Persistance Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        else:
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (prediction).nc')]

    for i_dataset in range(len(l_true_data_files)):
        # Load Data
        a_data_true = xr.open_dataset(l_true_data_files[i_dataset])
        a_data_pred = xr.open_dataset(l_prediction_data_files[i_dataset])
        
        # Compute Metrics
        m_site_coords = Get_Locations_with_Data(a_data_true)
        n_sites = m_site_coords.shape[0]
        m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
        
        for i_leadtime in range(len(l_lead_time)):
            lead_time = l_lead_time[i_leadtime]
            
            v_results_r = np.empty([n_sites,1])
            v_results_MAE = np.empty([n_sites,1])
            v_results_RMSE = np.empty([n_sites,1])
            v_results_bias = np.empty([n_sites,1])
            v_results_AB = np.empty([n_sites,1])
            v_results_MAPE = np.empty([n_sites,1])
            v_results_RMSPE = np.empty([n_sites,1])
            v_results_Pbias = np.empty([n_sites,1])
            
            v_results_r[:] = np.nan
            v_results_MAE[:] = np.nan
            v_results_MAPE[:] = np.nan
            v_results_RMSE[:] = np.nan
            v_results_RMSPE[:] = np.nan
            v_results_bias[:] = np.nan
            v_results_AB[:] = np.nan
            v_results_Pbias[:] = np.nan
            
            m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
        
            for site_index in range(n_sites):
                v_y = m_data_true[:,site_index]
                v_y_hat = m_data_pred[:,site_index]
               
                v_error = v_y - v_y_hat
                v_Perror = np.abs(v_error/v_y)
                        
                id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
            
                if (sum(id_num) > 24):
                    v_results_r[site_index] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                    v_results_MAE[site_index] = np.nanmean(np.abs(v_error))
                    v_results_RMSE[site_index] = np.sqrt(np.nanmean(v_error**2))
                    v_results_bias[site_index] = np.nanmean(v_error)
                    v_results_AB[site_index] = np.abs(np.nanmean(v_error))
                    v_results_MAPE[site_index] = np.nanmean(np.abs(v_Perror))
                    v_results_RMSPE[site_index] = np.sqrt(np.nanmean(v_Perror**2))
                    v_results_Pbias[site_index] = np.nanmean(v_error)/np.nanmean(v_y)
                
            for i_metric in range(len(l_metricsets)):
                metric = l_metricsets[i_metric]
                # Identify PoI
                if (metric == 'r'):
                    v_metric = v_results_r
                elif (metric == 'RMSE'):
                    v_metric = v_results_RMSE
                elif (metric == 'MAE'):
                    v_metric = v_results_MAE
                elif (metric == 'bias'):
                    v_metric = v_results_bias
                elif (metric == 'AB'):
                    v_metric = v_results_AB
                elif (metric == 'MAPE'):
                    v_metric = v_results_MAPE
                elif (metric == 'RMSPE'):
                    v_metric = v_results_RMSPE
                elif (metric == 'Pbias'):
                    v_metric = v_results_Pbias
                    
                a_metrics[i_leadtime,i_dataset,i_region,i_metric] = np.nanmedian(v_metric)
        
#%% Plotting    
fig = plt.figure(figsize=(6,3));

scalefactor = 0.9;

# axes = [\
#         fig.add_axes([scalefactor*0.1, scalefactor*0.66, scalefactor*0.28, scalefactor*0.4]),\
#         fig.add_axes([scalefactor*0.5, scalefactor*0.66, scalefactor*0.28, scalefactor*0.4]),\
#         fig.add_axes([scalefactor*0.1, scalefactor*0.12, scalefactor*0.28, scalefactor*0.4]),\
#         fig.add_axes([scalefactor*0.5, scalefactor*0.12, scalefactor*0.28, scalefactor*0.4]),\
#         ]
    
axes = [\
        fig.add_axes([scalefactor*0.15, scalefactor*0.15, scalefactor*0.25, scalefactor*0.92]),\
        fig.add_axes([scalefactor*0.55, scalefactor*0.15, scalefactor*0.25, scalefactor*0.92]),\
        ]
    
fig.text(0.0, 0.95,'a)',fontsize=12)
fig.text(0.37, 0.95,'b)',fontsize=12)

for i_metric in range(len(l_metricsets)):
    metric = l_metricsets[i_metric]
    
    axes[i_metric].set_xlabel('lead time [hrs]');
    axes[i_metric].set_ylabel(d_metriclabels[metric]);
    if (SUBSET == 2) | (SUBSET == 3):
        if metric == 'RMSE':
            axes[i_metric].set_yticks([0,1,2,3,4,5,6,7,8,9,10,11,12]);
        elif metric == 'AB':
            axes[i_metric].set_yticks([0,1,2,3,4,5,6]);
        elif metric == 'r':
            axes[i_metric].set_yticks([0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0]);
    
    axes[i_metric].set_ylim(d_metricrange[metric]);
    axes[i_metric].set_xlim([0,24]);
    # axes[i_metric].set_xticks([0,3,6,9,12,15,18,21,24]);
    axes[i_metric].set_xticks([0,6,12,18,24]);
    
    if (SUBSET == 2):
        if metric == 'r':
            # axes[i_metric].text(0.5,0.95,u"\u2b06 best",fontsize=10);
            axes[i_metric].text(-9,0.85,u"best \u27A1",fontsize=10,rotation=90);
        elif metric == 'RMSE':
            # axes[i_metric].text(0.5,0.25,u"\u1f814 best",fontsize=10);
            axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
        elif metric == 'MAE': 
            # axes[i_metric].text(0.5,0.5,u"\u1f814 best",fontsize=10);
            axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
        elif metric == 'AB': 
            # axes[i_metric].text(0.5,0.25,u"\u1f814 best",fontsize=10);
            axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
    elif (SUBSET == 3):
        if metric == 'r':
            # axes[i_metric].text(0.5,0.9,u"\u2b06 best",fontsize=10);
            axes[i_metric].text(-10,0.8,u"best \u27A1",fontsize=10,rotation=90);
        elif metric == 'RMSE':
            # axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
            axes[i_metric].text(-10,0,u"\u2B05 best",fontsize=10,rotation=90);
        elif metric == 'MAE': 
            # axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
            axes[i_metric].text(-10,0,u"\u2B05 best",fontsize=10,rotation=90);
        elif metric == 'AB': 
            # axes[i_metric].text(0.3,0.25,u"\u2b07 best",fontsize=10);
            axes[i_metric].text(-10,0,u"\u2B05 best",fontsize=10,rotation=90);
    else:
        if metric == 'r':
            # axes[i_metric].text(0.5,0.93,u"\u2b06 best",fontsize=10);
            axes[i_metric].text(-10,0.8,u"best \u27A1",fontsize=10,rotation=90);
        elif metric == 'AB':
            # axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
            axes[i_metric].text(-10,0,u"\u2B05 best",fontsize=10,rotation=90);
        else:
            # axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
            axes[i_metric].text(-10,0,u"\u2B05 best",fontsize=10,rotation=90);
        
    
    # if i_metric == 0:
    #     axes[i_metric] = fig.add_axes([0.52, 0.55, 0.4, 0.4]); # upper right
    # elif i_metric == 1:
    #     axes[i_metric] = fig.add_axes([0.1, 0.12, 0.4, 0.4]); # lower left
    # elif i_metric == 2:
    #     axes[i_metric] = fig.add_axes([0.1, 0.55, 0.4, 0.4]); # lower right
    # elif i_metric == 3:
    #     axes[i_metric] = fig.add_axes([0.52, 0.12, 0.4, 0.4]); # upper left
    
    for i_region in range(len(l_regions)):
        s_region = l_regions[i_region]
        
        for i_model in range(len(l_modeltypes)):
            s_modeltype = l_modeltypes[i_model]
            v_plot = a_metrics[:,i_model,i_region,i_metric]
            if ((i_metric == 0) & (i_model == 1)):
                axes[i_metric].plot(np.array(l_lead_time),v_plot,color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[s_modeltype],label=l_regions[i_region])
            elif ((i_metric == 1) & (i_region == 0)):
                axes[i_metric].plot(np.array(l_lead_time),v_plot,color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[s_modeltype],label=d_modeltype_labels[s_modeltype])
            else:
                axes[i_metric].plot(np.array(l_lead_time),v_plot,color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[s_modeltype])

    if i_metric == 0:
        axes[i_metric].legend(title = 'Area', bbox_to_anchor=(legend_1_pos[0],legend_1_pos[1]), prop={'size': 8});
    elif i_metric == 1:
        axes[i_metric].legend(title = 'Method', bbox_to_anchor=(legend_2_pos[0],legend_2_pos[1]), prop={'size': 8}); 


# Saving
OUTPUT_FIGURE_NAME = 'Comparison versus Lead Time - ' +str(SUBSET)+ figure_format
fig.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,OUTPUT_FIGURE_NAME),dpi=300)
fig.show()        
