#%% Load Packages & Define Functions
import os
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
import datetime as dt
import pytz

def Get_Locations_with_Data(a_data):
    m_data = (~np.isnan(a_data)).any(dim='time')
    m_coords = np.empty([0,2])
    for i_lat in m_data.lat:
        for i_lon in m_data.lon:
            if m_data.sel(lat=i_lat,lon=i_lon):
                m_coords = np.r_[m_coords,np.array([i_lat,i_lon],ndmin=2)]          
    return m_coords

def Array_to_Matrix_at_Coords(a_data,m_coords):
    n_coords = m_coords.shape[0]
    if hasattr(a_data,'time'):
        n_times = len(a_data.time)
    else:
        n_times = 1
    m_data = np.empty([n_times,n_coords])
    m_data[:] = np.nan
    for i_coord in range(n_coords):
        m_data[:,i_coord] = np.array(a_data.to_array().sel(lat=m_coords[i_coord,0],lon=m_coords[i_coord,1]))
    return m_data

#%% Basic Settings
        
figure_format = '.svg'

BASIC_FOLDER = os.path.join('C:\\','Data Archive')

OUTPUT_FIGURE_FOLDER = os.path.join(BASIC_FOLDER,'Generated Figures')

l_regions = ['Las Vegas','New Orleans','New York City','Salt Lake City','San Francisco']

SUBSET = 2

if SUBSET == 1:
    l_modeltypes = ['Persistance Model','Prediction Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
elif SUBSET == '1A':
    l_modeltypes = ['Diurnal Model','Prediction Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
elif SUBSET == 2:
    l_modeltypes = ['Prediction Model (GEOS-CF + TROPOMI)','Prediction & Kriging Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'without updating',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'with updating',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':':',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'}
elif SUBSET == 3:
    l_modeltypes = ['Prediction Model (GEOS-CF)','Prediction Model (GEOS-CF + TROPOMI)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'} 
elif SUBSET == 4:
    l_modeltypes = ['Prediction Model (GEOS-CF + TROPOMI)','Prediction Model (GEOS-CF + TROPOMI + VPN46A1)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'without VIIRS',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'with VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'GEOS-CF + TROPOMI + MET'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':':',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'} 
elif SUBSET == 5:
    l_modeltypes = ['Prediction Model (GEOS-CF + TROPOMI)','Prediction Model (GEOS-CF + TROPOMI + MET)']    
    d_modeltype_labels = {'Persistance Model':'Persistence Baseline',\
                      'Diurnal Model':'Climatology Baseline',\
                      'Prediction Model (GEOS-CF)':'GEOS-CF',\
                      'Prediction Model (GEOS-CF + TROPOMI)':'without Met.',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'GEOS-CF + TROPOMI & Kriging',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'GEOS-CF + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'GEOS-CF + TROPOMI + VIIRS',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'with Met.'}
    d_modeltype_styles = {'Persistance Model':':',\
                      'Diurnal Model':':',\
                      'Prediction Model (GEOS-CF)':':',\
                      'Prediction Model (GEOS-CF + TROPOMI)':':',\
                      'Prediction & Kriging Model (GEOS-CF + TROPOMI)':'-',\
                      'Prediction Model (GEOS-CF + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + VPN46A1)':'-',\
                      'Prediction Model (GEOS-CF + TROPOMI + MET)':'-'} 
    
l_lead_time = [1,3,6,12,18,24]
l_lead_time_labels = [' 1 hr',' 3 hr',' 6 hr','12 hr','18 hr','24 hr']

d_regional_timezones = {'New York City':'America/New_York',\
                       'New Orleans':'America/Chicago',\
                       'Salt Lake City':'America/Denver',\
                       'Las Vegas':'America/Los_Angeles',\
                       'San Francisco':'America/Los_Angeles'}

d_regional_colors = {'New York City':'blue',\
                       'New Orleans':'cyan',\
                       'Salt Lake City':'orange',\
                       'Las Vegas':'black',\
                       'San Francisco':'red'}
    
d_regional_settings = {'New York City':'2A7L',\
                       'New Orleans':'2A7L',\
                       'Salt Lake City':'2A4L',\
                       'Las Vegas':'1A4L',\
                       'San Francisco':'2A7L'}

# l_metricsets = ['r','AB','RMSE','MAE']
# l_metricsets = ['r','bias','RMSE','MAE']
if SUBSET == 2:
    l_metricsets = ['r','AB']
else:
    l_metricsets = ['r','RMSE']
    
# l_metricsets = ['bias']    

if SUBSET == 1:
    d_metricrange = {'RMSE':[0,20],\
               'MAE':[0,20],\
               'bias':[-5,5],\
               'AB':[0,10],\
               'r':[0,1]}
elif SUBSET == '1A':
    d_metricrange = {'RMSE':[0,20],\
               'MAE':[0,20],\
               'bias':[-5,5],\
               'AB':[0,10],\
               'r':[0,1]}
elif SUBSET == 2:
    d_metricrange = {'RMSE':[0,14],\
               'MAE':[0,10],\
               'bias':[-5,5],\
               'AB':[0,4],\
               'r':[-0.4,1]}
elif SUBSET == 3:
    d_metricrange = {'RMSE':[0,14],\
               'MAE':[0,10],\
               'bias':[-5,5],\
               'AB':[0,4],\
               'r':[-0.4,1]}
elif SUBSET == 4:
    d_metricrange = {'RMSE':[0,14],\
               'MAE':[0,10],\
               'bias':[-5,5],\
               'AB':[0,4],\
               'r':[0,1]}
elif SUBSET == 5:
    d_metricrange = {'RMSE':[0,14],\
               'MAE':[0,10],\
               'bias':[-5,5],\
               'AB':[0,4],\
               'r':[0,1]}
        
d_metriclabels = {'RMSE':'RMSE [ppb]',\
                  'MAE':'MAE [ppb]',\
                  'bias':'bias [ppb]',\
                  'AB':'AB [ppb]',\
                  'r':'correlation'}
    
if len(l_metricsets) == 4:
    if SUBSET == 1:
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.15,-0.4]
    elif SUBSET == '1A':
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.15,-0.4]
    elif SUBSET == 2:
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.0,-0.4]
    elif SUBSET == 3:
        legend_1_pos = [3.3,1]
        legend_2_pos = [2.14,-0.4]
    elif SUBSET == 4:
        legend_1_pos = [3.3,1]
        legend_2_pos = [1.85,-0.4]
    elif SUBSET == 5:
        legend_1_pos = [3.3,1]
        legend_2_pos = [1.8,-0.4]
        
if len(l_metricsets) == 2:
    if SUBSET == 1:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == '1A':
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 2:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 3:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 4:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 5:
        legend_1_pos = [3.55,1]
        legend_2_pos = [1.0,0.5]
        
if len(l_metricsets) == 1:
    if SUBSET == 1:
        legend_1_pos = [1.0,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == '1A':
        legend_1_pos = [1.0,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 2:
        legend_1_pos = [1.0,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 3:
        legend_1_pos = [1.0,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 4:
        legend_1_pos = [1.0,1]
        legend_2_pos = [1.0,0.5]
    elif SUBSET == 5:
        legend_1_pos = [1.0,1]
        legend_2_pos = [1.0,0.5]
        
#%% Compute Metrics
a_metrics = np.zeros([24,len(l_modeltypes),len(l_regions),len(l_metricsets),len(l_lead_time)])
a_metrics[:] = np.nan

for i_region in range(len(l_regions)):
    s_region = l_regions[i_region]
    s_regioncode = s_region.replace(' ','')
    s_spec = d_regional_settings[s_region]
    l_true_data_files = []
    l_prediction_data_files = []
    for i_model in range(len(l_modeltypes)):
        s_modeltype = l_modeltypes[i_model]
        basepath = os.path.join(BASIC_FOLDER,'Generated Data')
        if (s_modeltype == 'Diurnal Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        elif (s_modeltype == 'Persistance Model'):
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' (prediction).nc')]
        else:
            l_true_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (truth).nc')]
            l_prediction_data_files += [os.path.join(basepath,s_regioncode+' '+s_modeltype +' '+s_spec+' (prediction).nc')]

    for i_dataset in range(len(l_true_data_files)):
        # Load Data
        a_data_true = xr.open_dataset(l_true_data_files[i_dataset])
        a_data_pred = xr.open_dataset(l_prediction_data_files[i_dataset])
  
        m_site_coords = Get_Locations_with_Data(a_data_true)
        n_sites = m_site_coords.shape[0]  
        
        m_y = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
        
        
  
        for i_leadtime in range(len(l_lead_time)):
            lead_time = l_lead_time[i_leadtime]
            
            a_data_pred_leadtime = a_data_pred.sel(leadtime=lead_time).drop('leadtime')
            
            m_y_hat = Array_to_Matrix_at_Coords(a_data_pred_leadtime,m_site_coords)
            
            v_results_r = np.empty([24,1])
            v_results_MAE = np.empty([24,1])
            v_results_RMSE = np.empty([24,1])
            v_results_bias = np.empty([24,1])
            v_results_AB = np.empty([24,1])
            v_results_MAPE = np.empty([24,1])
            v_results_RMSPE = np.empty([24,1])
            v_results_Pbias = np.empty([24,1])
            
            v_results_r[:] = np.nan
            v_results_MAE[:] = np.nan
            v_results_MAPE[:] = np.nan
            v_results_RMSE[:] = np.nan
            v_results_RMSPE[:] = np.nan
            v_results_bias[:] = np.nan
            v_results_AB[:] = np.nan
            v_results_Pbias[:] = np.nan
            
            v_time = a_data_pred_leadtime['time'].to_index()
            v_time_local = v_time.tz_localize(pytz.UTC).tz_convert(pytz.timezone(d_regional_timezones[s_region]))
            v_hour_local = v_time_local.hour
            
            for local_hour in range(24):
                id_hour = (v_hour_local == local_hour)
                # a_data_pred_temp = a_data_pred_leadtime.sel(time=id_hour)
                # a_data_true_temp = a_data_true.sel(time=id_hour)
                
                # m_y = Array_to_Matrix_at_Coords(a_data_true_temp,m_site_coords)
                # m_y_hat = Array_to_Matrix_at_Coords(a_data_pred_temp,m_site_coords)
                
                v_y = m_y[id_hour,:].reshape(-1)
                v_y_hat = m_y_hat[id_hour,:].reshape(-1)

                # v_y = a_data_true_temp.to_array().variable.values.reshape(-1)
                # v_y_hat = a_data_pred_temp.to_array().variable.values.reshape(-1)
               
                v_error = v_y - v_y_hat
                v_Perror = np.abs(v_error/v_y)
                        
                id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
            
                if (sum(id_num) > 24):
                    v_results_r[local_hour] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                    v_results_MAE[local_hour] = np.nanmean(np.abs(v_error))
                    v_results_RMSE[local_hour] = np.sqrt(np.nanmean(v_error**2))
                    v_results_bias[local_hour] = np.nanmean(v_error)
                    v_results_AB[local_hour] = np.abs(np.nanmean(v_error))
                    v_results_MAPE[local_hour] = np.nanmean(np.abs(v_Perror))
                    v_results_RMSPE[local_hour] = np.sqrt(np.nanmean(v_Perror**2))
                    v_results_Pbias[local_hour] = np.nanmean(v_error)/np.nanmean(v_y)
                
            for i_metric in range(len(l_metricsets)):
                metric = l_metricsets[i_metric]
                # Identify PoI
                if (metric == 'r'):
                    v_metric = v_results_r
                elif (metric == 'RMSE'):
                    v_metric = v_results_RMSE
                elif (metric == 'MAE'):
                    v_metric = v_results_MAE
                elif (metric == 'bias'):
                    v_metric = v_results_bias
                elif (metric == 'AB'):
                    v_metric = v_results_AB
                elif (metric == 'MAPE'):
                    v_metric = v_results_MAPE
                elif (metric == 'RMSPE'):
                    v_metric = v_results_RMSPE
                elif (metric == 'Pbias'):
                    v_metric = v_results_Pbias
                    
                a_metrics[:,i_dataset,i_region,i_metric,i_leadtime] = v_metric.reshape(-1)
        
#%% Plotting    

SPECIAL_SECOND_PLOT = 1


for i_lead in range(len(l_lead_time)):
    lead_time = l_lead_time[i_lead]
    fig = plt.figure(figsize=(6,3));
    
    scalefactor = 0.9;
    
    if len(l_metricsets) == 4:
        axes = [\
                fig.add_axes([scalefactor*0.1, scalefactor*0.66, scalefactor*0.28, scalefactor*0.4]),\
                fig.add_axes([scalefactor*0.5, scalefactor*0.66, scalefactor*0.28, scalefactor*0.4]),\
                fig.add_axes([scalefactor*0.1, scalefactor*0.12, scalefactor*0.28, scalefactor*0.4]),\
                fig.add_axes([scalefactor*0.5, scalefactor*0.12, scalefactor*0.28, scalefactor*0.4]),\
                ]
    if len(l_metricsets) == 2:
        axes = [\
                fig.add_axes([scalefactor*0.15, scalefactor*0.15, scalefactor*0.25, scalefactor*0.92]),\
                fig.add_axes([scalefactor*0.55, scalefactor*0.15, scalefactor*0.25, scalefactor*0.92]),\
                ]

    
        fig.text(0.0, 0.95,'a)',fontsize=12)
        fig.text(0.37, 0.95,'b)',fontsize=12)
    
    if len(l_metricsets) == 1:
        axes = [\
                fig.add_axes([scalefactor*0.15, scalefactor*0.12, scalefactor*0.65, scalefactor*0.95]),\
                ]
            
    fig.text(0.74, 0.20,'Lead Time: '+str(lead_time)+' hrs',fontsize=10)
    

    for i_metric in range(len(l_metricsets)):
        metric = l_metricsets[i_metric]
        
        axes[i_metric].set_xlabel('local time');
        axes[i_metric].set_ylabel(d_metriclabels[metric]);
        if (SUBSET == 2) | (SUBSET == 3) | (SUBSET >= 4):
            if metric == 'RMSE':
                axes[i_metric].set_yticks([0,2,4,6,8,10,12,14]);
            elif metric == 'AB':
                axes[i_metric].set_yticks([0,0.5,1,1.5,2,2.5,3,3.5,4,4.5,5,5.5,6]);
        
        axes[i_metric].set_ylim(d_metricrange[metric]);
        axes[i_metric].set_xlim([0,24]);
        axes[i_metric].set_xticks([0,6,12,18,24]);
        
        if (SUBSET == 2):
            if metric == 'r':
                # axes[i_metric].text(0.5,0.93,u"\u2b06 best",fontsize=10);
                axes[i_metric].text(-10,0.8,u"best \u27A1",fontsize=10,rotation=90);
            elif metric == 'RMSE':
                # axes[i_metric].text(0.5,0.75,u"\u2b07 best",fontsize=10);
                axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
            elif metric == 'MAE': 
                # axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
                axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
            elif metric == 'AB': 
                # axes[i_metric].text(0.5,0.1,u"\u2b07 best",fontsize=10);
                axes[i_metric].text(-9,0,u"\u2B05 best",fontsize=10,rotation=90);
            elif metric == 'bias':
                axes[i_metric].text(0.5,-0.12,u"\u2b05 best",fontsize=10);
        elif (SUBSET == 3):
            if metric == 'r':
                # axes[i_metric].text(0.5,0.93,u"\u2b06 best",fontsize=10);
                axes[i_metric].text(-10,0.8,u"best \u27A1",fontsize=10,rotation=90);
            elif metric == 'RMSE':
                # axes[i_metric].text(0.5,0.3,u"\u2b07 best",fontsize=10);
                axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
            elif metric == 'MAE': 
                # axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
                axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
            elif metric == 'AB': 
                # axes[i_metric].text(0.3,0.25,u"\u2b07 best",fontsize=10);
                axes[i_metric].text(-8,0,u"\u2B05 best",fontsize=10,rotation=90);
            elif metric == 'bias':
                axes[i_metric].text(0.5,-0.25,u"\u2b05 best",fontsize=10);
        elif (SUBSET >= 4):
            if metric == 'r':
                axes[i_metric].text(0.5,0.9,u"\u2b06 best",fontsize=10);
            elif metric == 'RMSE':
                axes[i_metric].text(0.5,0.75,u"\u2b07 best",fontsize=10);
            elif metric == 'MAE': 
                axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
            elif metric == 'AB': 
                axes[i_metric].text(0.3,0.25,u"\u2b07 best",fontsize=10);
            elif metric == 'bias':
                axes[i_metric].text(0.5,-0.25,u"\u2b05 best",fontsize=10);
        else:
            if metric == 'r':
                axes[i_metric].text(0.5,0.9,u"\u2b06 best",fontsize=10);
            elif metric == 'AB':
                axes[i_metric].text(0.5,0.5,u"\u2b07 best",fontsize=10);
            elif metric == 'bias':
                axes[i_metric].text(0.5,-0.25,u"\u2b05 best",fontsize=10);
            else:
                axes[i_metric].text(0.5,1,u"\u2b07 best",fontsize=10);
    
        lines = [[]]*len(l_regions)
        
        for i_region in range(len(l_regions)):
            s_region = l_regions[i_region]
            lines[i_region] = [[]]*len(l_modeltypes)
            for i_model in range(len(l_modeltypes)):
                s_modeltype = l_modeltypes[i_model]
                v_plot = a_metrics[:,i_model,i_region,i_metric,i_lead]
                
                if ((i_metric == 0) & (i_model == 1)):
                    lines[i_region][i_model] = axes[i_metric].plot(np.array(range(24)),v_plot,color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[s_modeltype],label=l_regions[i_region])
                elif ((i_metric == 1) & (i_region == 0)):
                    lines[i_region][i_model] = axes[i_metric].plot(np.array(range(24)),v_plot,color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[s_modeltype],label=d_modeltype_labels[s_modeltype])
                else:
                    lines[i_region][i_model] = axes[i_metric].plot(np.array(range(24)),v_plot,color=d_regional_colors[s_region],linewidth=1,linestyle=d_modeltype_styles[s_modeltype])
        if len(l_metricsets) == 1:
            legend_1 = axes[i_metric].legend(title = 'Area', bbox_to_anchor=(legend_1_pos[0],legend_1_pos[1]), prop={'size': 8});
            for i_region in range(len(l_regions)):
                for i_model in range(len(l_modeltypes)):
                    if (i_region == 0): 
                        lines[i_region][i_model][0].set_label(d_modeltype_labels[l_modeltypes[i_model]])
                    else:
                        lines[i_region][i_model][0].set_label('_')
            legend_2 = axes[i_metric].legend(title = 'Method', bbox_to_anchor=(legend_2_pos[0],legend_2_pos[1]), prop={'size': 8}); 
            axes[i_metric].add_artist(legend_1)
        else:            
            if i_metric == 0:
                axes[i_metric].legend(title = 'Area', bbox_to_anchor=(legend_1_pos[0],legend_1_pos[1]), prop={'size': 8});
            elif i_metric == 1:
                axes[i_metric].legend(title = 'Method', bbox_to_anchor=(legend_2_pos[0],legend_2_pos[1]), prop={'size': 8}); 
        
        if SUBSET == 3:
            axes[i_metric].fill_between(np.array(range(24)), d_metricrange[metric][0], d_metricrange[metric][1], where=(np.array(range(24)) >= 13)&(np.array(range(24)) <= 14), color='grey', alpha=0.5)
            # axes[i_metric].text(14.75, 0.7*(d_metricrange[metric][1] - d_metricrange[metric][0]) + d_metricrange[metric][0],'satellite pass',fontsize=10,color='grey',rotation='vertical')
            axes[i_metric].text(10.75, 0.75*(d_metricrange[metric][1] - d_metricrange[metric][0]) + d_metricrange[metric][0],'satellite',fontsize=10,color='grey',rotation='vertical')
            axes[i_metric].text(14.25, 0.75*(d_metricrange[metric][1] - d_metricrange[metric][0]) + d_metricrange[metric][0],'   pass   ',fontsize=10,color='grey',rotation='vertical')
    
        if SPECIAL_SECOND_PLOT == 1:
            if (i_metric == 1):
                for i_region in range(len(l_regions)):
                    if (((l_regions[i_region] == 'Salt Lake City') | (l_regions[i_region] == 'New Orleans') | (l_regions[i_region] == 'Las Vegas'))):
                        for i_model in range(len(l_modeltypes)):
                            lines[i_region][i_model][0].set_visible(False)
        
    # Saving
    OUTPUT_FIGURE_NAME = 'Performance versus Local Time, '+str(lead_time)+' hr lead - ' +str(SUBSET)+ figure_format
    fig.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,OUTPUT_FIGURE_NAME),dpi=300)
    fig.show()        
