# Import Packages
import os
import numpy as np
import pandas as pd
import datetime as dt
os.environ['PROJ_LIB'] = r'D:\Programs\Anaconda\pkgs\proj4-5.2.0-ha925a31_1\Library\share'
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt

# Basic Settings
INPUT_FILE = 'EPA DATA\\hourly_42602_2019.csv'
OUTPUT_FOLDER = 'EPA DATA'
SUBSET_REGION = 'LasVegas'
OUTPUT_FILE = 'EPA_NO2_Sept_2019_'+SUBSET_REGION+'.csv'
GRIDRES_DEG = np.nan
PLOT = True
time_start = dt.datetime(2019,9,1,0,0,0)
time_stop = dt.datetime(2019,9,30,23,59,59)

# Function Definitions
def Regional_LatLon_Bounds(region_name):
    maxlat = np.inf
    maxlon = np.inf
    minlat = -np.inf
    minlon = -np.inf
    if region_name == 'CONUS':
        maxlat = 50 # dlat = 25
        maxlon = -65 # dlon = 60
        minlat = 25
        minlon = -125
    if region_name == 'SoCal':
        maxlat = 39 # dlat = 7
        maxlon = -115 # dlon = 8
        minlat = 32
        minlon = -123
    if region_name == 'UtCoWy':
        maxlat = 46 # dlat = 8
        maxlon = -104 # dlon = 9
        minlat = 38
        minlon = -113
    if region_name == 'NorthEast':
        maxlat = 43 # dlat = 5
        maxlon = -70 # dlon = 11
        minlat = 38
        minlon = -81
    if region_name == 'GulfCoast':
        maxlat = 35 # dlat = 7
        maxlon = -88 # dlon = 11
        minlat = 28
        minlon = -99
    if region_name == 'NewYorkCity':
        maxlat = 42 # dlat = 2
        maxlon = -73 # dlon = 2
        minlat = 40
        minlon = -75
    if region_name == 'Boston':
        maxlat = 43 # dlat = 2
        maxlon = -70 # dlon = 2
        minlat = 41
        minlon = -72
    if region_name == 'NewOrleans':
        maxlat = 31 # dlat = 2
        maxlon = -90 # dlon = 2
        minlat = 29
        minlon = -92
    if region_name == 'Houston':
        maxlat = 31 # dlat = 2
        maxlon = -97 # dlon = 2
        minlat = 29
        minlon = -99
    if region_name == 'SaltLakeCity':
        maxlat = 42 # dlat = 2
        maxlon = -111 # dlon = 2
        minlat = 40
        minlon = -113
    if region_name == 'Denver':
        maxlat = 41 # dlat = 2
        maxlon = -104 # dlon = 2
        minlat = 39
        minlon = -106
    if region_name == 'SanFrancisco':
        maxlat = 39 # dlat = 2
        maxlon = -121 # dlon = 2
        minlat = 37
        minlon = -123
    if region_name == 'LosAngeles':
        maxlat = 35 # dlat = 2
        maxlon = -116 # dlon = 4
        minlat = 33
        minlon = -120
    if region_name == 'LasVegas':
        maxlat = 37 # dlat = 2
        maxlon = -114 # dlon = 2
        minlat = 35
        minlon = -116
    return maxlat,maxlon,minlat,minlon

def Significant_Figure_Rounding(x,sigdec):
    exponent = np.floor(np.log10(np.abs(x)))
    value = x/np.power(10,exponent)
    new_value = np.round(value,decimals=sigdec)
    new_x = new_value*np.power(10,exponent)
    return new_x

# Read File
f_data = pd.read_csv(INPUT_FILE)

v_lat = np.array(f_data['Latitude'])
v_lon = np.array(f_data['Longitude'])
v_date = f_data['Date GMT']
v_time = f_data['Time GMT']
v_data = np.array(f_data['Sample Measurement'])
v_mdl = np.array(f_data['MDL'])

# Handle MDL
v_data[v_data < v_mdl] = np.nan

# Handle Timestamps
v_time = pd.to_datetime(v_time)
v_date = pd.to_datetime(v_date)

v_year = v_date.dt.year
v_month = v_date.dt.month
v_day = v_date.dt.day
v_hour = v_time.dt.hour
v_minute = (v_time + pd.to_timedelta(30,unit='minute')).dt.minute

v_datetime = np.array([dt.datetime(v_year[i_time],v_month[i_time],v_day[i_time],v_hour[i_time],v_minute[i_time]) for i_time in range(len(v_year))])

# Subsetting (Optional)
n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)
v_data[v_lat < n_min_lat]=np.nan
v_data[v_lat > n_max_lat]=np.nan
v_data[v_lon < n_min_lon]=np.nan
v_data[v_lon > n_max_lon]=np.nan

v_data[v_datetime < time_start]=np.nan
v_data[v_datetime > time_stop]=np.nan

# Trim Missing Data
v_year = v_year[~np.isnan(v_data)]
v_month = v_month[~np.isnan(v_data)]
v_day = v_day[~np.isnan(v_data)]
v_hour = v_hour[~np.isnan(v_data)]
v_minute = v_minute[~np.isnan(v_data)]
v_lat = v_lat[~np.isnan(v_data)]
v_lon = v_lon[~np.isnan(v_data)]
v_data = v_data[~np.isnan(v_data)]

f_data = pd.DataFrame()
f_data['Year'] = v_year
f_data['Month'] = v_month
f_data['Day'] = v_day
f_data['Hour'] = v_hour
f_data['Minute'] = v_minute
f_data['Lat'] = v_lat
f_data['Lon'] = v_lon
f_data['NO2'] = v_data

# Plotting
if PLOT:
    # Location-based Averages
    m_latlon = np.c_[(v_lat,v_lon)]
    m_unique_latlon = np.unique(m_latlon,axis=0)
    v_location_averages = np.empty(m_unique_latlon.shape[0])
    v_location_averages[:] = np.nan
    for location_latlon in m_unique_latlon:
        index = (m_unique_latlon == location_latlon).all(axis=1)
        value = np.nanmean(v_data[(m_latlon == location_latlon).all(axis=1)])
        v_location_averages[index] = value
        # m_latlon[(m_latlon == location_latlon).all(axis=1),]
        
    # Create Plot
    mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([n_min_lat,-90]),urcrnrlat = np.min([n_max_lat,90]),llcrnrlon=np.max([n_min_lon,-180]),urcrnrlon = np.min([n_max_lon,180]))
    mapobj.drawcoastlines(linewidth=0.5)
    mapobj.drawparallels(np.arange(-90., 90., 1.), labels=[1, 0, 0, 0])
    mapobj.drawmeridians(np.arange(-180, 180., 1.), labels=[0, 0, 0, 1])
    my_cmap = plt.cm.get_cmap('jet')
    my_cmap.set_under('w')
    mapobj.scatter(m_unique_latlon[:,1],m_unique_latlon[:,0],c=v_location_averages,vmin=0*np.nanmin(v_location_averages),vmax=1*np.nanmax(v_location_averages),cmap=my_cmap)
    colorbarobj = mapobj.colorbar()
    colorbarobj.set_label('Average NO2 [ppb]')
    # plt.autoscale()
    fig = plt.gcf()
    plt.show()


# Save Output File
f_data.to_csv(os.path.join(OUTPUT_FOLDER,OUTPUT_FILE), index = False)
print('File saved. Processing complete.')








