# Import Packages
import os
import datetime
import pandas as pd
import urllib.request
import urllib.parse

# Basic Settings
URL_BASE = 'https://portal.nccs.nasa.gov/datashare/gmao/geos-cf/v1/das'
# FILENAME_BASE = 'GEOS-CF.v01.rpl.aqc_tavg_1hr_g1440x721_v1.'
FILENAME_BASE = 'GEOS-CF.v01.rpl.met_tavg_1hr_g1440x721_x1.'
FILENAME_EXTENSION = '.nc4'
# OUTPUT_FOLDER = 'GEOS-CF DATA'
OUTPUT_FOLDER = 'GEOS-CF MET DATA'

START_YEAR = 2019
START_MONTH = 9
START_DAY = 1
END_YEAR = 2019
END_MONTH = 9
END_DAY = 30


# Initial Setup: Date Range
d_start = datetime.datetime(START_YEAR,START_MONTH,START_DAY)
d_end = datetime.datetime(END_YEAR,END_MONTH,END_DAY)

l_datetimes = pd.date_range(start=d_start,end=d_end,freq='D').tolist()
v_dates = []
v_daily_urls = []
for d_day in l_datetimes:
    s_date = str(d_day.year)
    s_url = URL_BASE + '/Y' + str(d_day.year)
    if (d_day.month < 10):
        s_date += ('0' + str(d_day.month))
        s_url += ('/M' + '0' + str(d_day.month))
    else:
        s_date += str(d_day.month)
        s_url += ('/M' + str(d_day.month))
    if (d_day.day < 10):
        s_date += ('0' + str(d_day.day))
        s_url += ('/D' + '0' + str(d_day.day))
    else:
        s_date += str(d_day.day)
        s_url += ('/D' + str(d_day.day))
    v_dates += [s_date]
    v_daily_urls += [s_url]
        
# Initial Setup: Hours Range
v_hours = []
for i_hour in range(0,24):
    if (i_hour < 10):
        s_hour = '0'+str(i_hour)+'30z'
    else:
        s_hour = str(i_hour)+'30z' 
    v_hours += [s_hour]

# Download 
for s_date in v_dates:
    for s_hour in v_hours:
        s_filename = FILENAME_BASE + s_date + '_' + s_hour + FILENAME_EXTENSION
        s_url = v_daily_urls[v_dates.index(s_date)] + '/' + s_filename
        s_savepath = os.path.join(OUTPUT_FOLDER,s_filename)
        urllib.request.urlretrieve(s_url,s_savepath)
        print("""\nDownloaded: %s to: %s""" %(s_url,s_savepath))
        
print('Downloads Complete')
        
        
        
        
