# Import Packages
import os
import datetime 
import numpy as np
from netCDF4 import Dataset
os.environ['PROJ_LIB'] = r'D:\Programs\Anaconda\pkgs\proj4-5.2.0-ha925a31_1\Library\share'
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
import pandas as pd

# Basic Settings
INPUT_FOLDER = 'GEOS-CF MET DATA'
OUTPUT_FOLDER = 'GEOS-CF MET DATA'
GRIDRES_DEG = np.nan
PLOT = False

v_DATATYPES = ['T2M','Q2M','V2M','U2M','ZPBL']

v_SUBSET_REGIONS = ['NewYorkCity','Boston','NewOrleans','Houston','SaltLakeCity','Denver','SanFrancisco','LosAngeles','LasVegas']


# Function Definitions
def Regional_LatLon_Bounds(region_name):
    maxlat = np.inf
    maxlon = np.inf
    minlat = -np.inf
    minlon = -np.inf
    if region_name == 'CONUS':
        maxlat = 50 # dlat = 25
        maxlon = -65 # dlon = 60
        minlat = 25
        minlon = -125
    if region_name == 'SoCal':
        maxlat = 39 # dlat = 7
        maxlon = -115 # dlon = 8
        minlat = 32
        minlon = -123
    if region_name == 'UtCoWy':
        maxlat = 46 # dlat = 8
        maxlon = -104 # dlon = 9
        minlat = 38
        minlon = -113
    if region_name == 'NorthEast':
        maxlat = 43 # dlat = 5
        maxlon = -70 # dlon = 11
        minlat = 38
        minlon = -81
    if region_name == 'GulfCoast':
        maxlat = 35 # dlat = 7
        maxlon = -88 # dlon = 11
        minlat = 28
        minlon = -99
    if region_name == 'NewYorkCity':
        maxlat = 42 # dlat = 2
        maxlon = -73 # dlon = 2
        minlat = 40
        minlon = -75
    if region_name == 'Boston':
        maxlat = 43 # dlat = 2
        maxlon = -70 # dlon = 2
        minlat = 41
        minlon = -72
    if region_name == 'NewOrleans':
        maxlat = 31 # dlat = 2
        maxlon = -90 # dlon = 2
        minlat = 29
        minlon = -92
    if region_name == 'Houston':
        maxlat = 31 # dlat = 2
        maxlon = -97 # dlon = 2
        minlat = 29
        minlon = -99
    if region_name == 'SaltLakeCity':
        maxlat = 42 # dlat = 2
        maxlon = -111 # dlon = 2
        minlat = 40
        minlon = -113
    if region_name == 'Denver':
        maxlat = 41 # dlat = 2
        maxlon = -104 # dlon = 2
        minlat = 39
        minlon = -106
    if region_name == 'SanFrancisco':
        maxlat = 39 # dlat = 2
        maxlon = -121 # dlon = 2
        minlat = 37
        minlon = -123
    if region_name == 'LosAngeles':
        maxlat = 35 # dlat = 2
        maxlon = -116 # dlon = 4
        minlat = 33
        minlon = -120
    if region_name == 'LasVegas':
        maxlat = 37 # dlat = 2
        maxlon = -114 # dlon = 2
        minlat = 35
        minlon = -116
    return maxlat,maxlon,minlat,minlon

def Significant_Figure_Rounding(x,sigdec):
    exponent = np.floor(np.log10(np.abs(x)))
    value = x/np.power(10,exponent)
    new_value = np.round(value,decimals=sigdec)
    new_x = new_value*np.power(10,exponent)
    return new_x

for i_subset_region in range(len(v_SUBSET_REGIONS)):
    # Determine Input Files
    good_extension = '.nc4'
    fileList = os.listdir(INPUT_FOLDER)
    fileList = [fileName for fileName in fileList if good_extension in fileName]
    
    # Initialize Storage Arrays
    v_year_ALL = []
    v_month_ALL = []
    v_day_ALL = []
    v_hour_ALL = []
    v_minute_ALL = []
    # v_second_ALL = []
    
    v_lat_ALL = []
    v_lon_ALL = []
    l_data_ALL = {DATATYPE:[] for DATATYPE in v_DATATYPES}
    
    SUBSET_REGION = v_SUBSET_REGIONS[i_subset_region]
    OUTPUT_FILE = 'GEOS-CF_MET_Sept_2019_'+SUBSET_REGION+'.csv'
    
    
    # Cycle Through Files
    for fileName in fileList:
        # Load Input File
        file = Dataset(os.path.join(INPUT_FOLDER,fileName), 'r')
    
        # Determine Input Data Types
        lat = file.variables['lat']
        lon = file.variables['lon']
        timestamp = file.variables['time']
        s_begin_date = str(file.variables['time'].begin_date)
        s_begin_time = str(file.variables['time'].begin_time)
        
        v_lat = np.array(lat)
        v_lon = np.array(lon)
        l_data = {DATATYPE:[] for DATATYPE in v_DATATYPES}
        
        for DATATYPE in v_DATATYPES:
            data = file.variables[DATATYPE]
            n_fill_data = data._FillValue
            m_data = np.array(data[0][:][:])
            m_data[m_data == n_fill_data]=np.nan
            l_data[DATATYPE] = m_data
            
        n_minutes_from_start = int(timestamp[0])
    
        file.close()
        
        
    
    
        # Handle Times
        n_year = int(s_begin_date[0:4])
        n_month = int(s_begin_date[4:6])
        n_day = int(s_begin_date[6:8])
        
        if (len(s_begin_time) == 6):
            n_begin_hour = int(s_begin_time[0:2])
            n_begin_minute = int(s_begin_time[2:4])
        elif (len(s_begin_time) == 5):
            n_begin_hour = int(s_begin_time[0:1])
            n_begin_minute = int(s_begin_time[1:3])
        elif (len(s_begin_time) == 4):
            n_begin_hour = 0
            n_begin_minute = int(s_begin_time[0:2])
        elif (len(s_begin_time) == 3):
            n_begin_hour = 0
            n_begin_minute = int(s_begin_time[0:1])
        elif (len(s_begin_time) == 2):
            n_begin_hour = 0
            n_begin_minute = 0
        elif (len(s_begin_time) == 1):
            n_begin_hour = 0
            n_begin_minute = 0
        else:
            n_begin_hour = 0
            n_begin_minute = 0
    
        if (n_minutes_from_start == 0):
            n_hour = n_begin_hour
            n_minute = n_begin_minute
        else:
            n_time = datetime.datetime(n_year,n_month,n_day,n_begin_hour,n_begin_minute) + datetime.timedelta(minutes=n_minutes_from_start)
            n_year = n_time.year
            n_month = n_time.month
            n_day = n_time.day
            n_hour = n_time.hour
            n_minute = n_time.minute
    
        # Handle Latitude and Longitude
        n_lat = len(v_lat)   
        n_lon = len(v_lon)
        n_dim_1,n_dim_2 = m_data.shape
        
        if (n_dim_1 == n_lat):
            m_lat = np.transpose(np.repeat(np.array(v_lat,ndmin=2),n_dim_2,0))
            m_lon = np.repeat(np.array(v_lon,ndmin=2),n_dim_1,0)
        else:
            m_lat = np.transpose(np.repeat(np.array(v_lat,ndmin=2),n_dim_1,0))
            m_lon = np.repeat(np.array(v_lon,ndmin=2),n_dim_2,0)
            
    
        # Subsetting (Optional)
        n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)
        for DATATYPE in v_DATATYPES:
            m_data = l_data[DATATYPE]
            m_data[m_lat < n_min_lat]=np.nan
            m_data[m_lat > n_max_lat]=np.nan
            m_data[m_lon < n_min_lon]=np.nan
            m_data[m_lon > n_max_lon]=np.nan
            l_data[DATATYPE] = m_data
            
        # Regridding (Optional)
        # if ~np.isnan(GRIDRES_DEG):
        
        # Verification Plot (Optional)
        if PLOT:
            for DATATYPE in v_DATATYPES:
                m_data = l_data[DATATYPE]
                m_data_masked = np.ma.masked_array(m_data, np.isnan(m_data))
                m_lat_masked = np.ma.masked_array(m_lat, np.isnan(m_data))
                m_lon_masked = np.ma.masked_array(m_lon, np.isnan(m_data))
                mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([n_min_lat,-90]),urcrnrlat = np.min([n_max_lat,90]),llcrnrlon=np.max([n_min_lon,-180]),urcrnrlon = np.min([n_max_lon,180]))
                mapobj.drawcoastlines(linewidth=0.5)
                mapobj.drawparallels(np.arange(-90., 120., 30.), labels=[1, 0, 0, 0])
                mapobj.drawmeridians(np.arange(-180, 180., 45.), labels=[0, 0, 0, 1])
                my_cmap = plt.cm.get_cmap('jet')
                my_cmap.set_under('w')
                mapobj.pcolormesh(m_lon,m_lat,m_data_masked,latlon=True,vmin=1*np.nanmin(m_data_masked),vmax=1*np.nanmax(m_data_masked),cmap=my_cmap)
                # mapobj.pcolormesh(m_lon,m_lat,m_lat_masked,latlon=True,vmin=np.nanmin(m_lat[~np.isnan(m_data)]),vmax=np.nanmax(m_lat[~np.isnan(m_data)]),cmap=my_cmap)
                # mapobj.pcolormesh(m_lon,m_lat,m_lon_masked,latlon=True,vmin=np.nanmin(m_lon[~np.isnan(m_data)]),vmax=np.nanmax(m_lon[~np.isnan(m_data)]),cmap=my_cmap)
        
                colorbarobj = mapobj.colorbar()
                # colorbarobj.set_label('Colorbar Label')
                # plt.autoscale()
                fig = plt.gcf()
                plt.show()
    
        # Rearrange matrices to Vectors
        v_lat = m_lat.flatten()
        v_lon = m_lon.flatten()
        for DATATYPE in v_DATATYPES:
            m_data = l_data[DATATYPE]
            v_data = m_data.flatten()
            l_data[DATATYPE] = v_data
        
        # Trim Missing Data
        ID_NUM = np.zeros(len(v_data)) > 1
        for DATATYPE in v_DATATYPES:
            ID_NUM = np.logical_or(ID_NUM,~np.isnan(l_data[DATATYPE]))
            
        v_lat = v_lat[ID_NUM]
        v_lon = v_lon[ID_NUM]
        
        for DATATYPE in v_DATATYPES:
            v_data = l_data[DATATYPE]
            v_data = v_data[ID_NUM]
            l_data[DATATYPE] = v_data
                    
        # Concatenate to Outputs
        v_year_ALL = v_year_ALL + np.repeat(n_year,len(v_data)).tolist()
        v_month_ALL = v_month_ALL + np.repeat(n_month,len(v_data)).tolist()
        v_day_ALL = v_day_ALL + np.repeat(n_day,len(v_data)).tolist()
        v_hour_ALL = v_hour_ALL + np.repeat(n_hour,len(v_data)).tolist()
        v_minute_ALL = v_minute_ALL + np.repeat(n_minute,len(v_data)).tolist()
        # v_second_ALL = v_second_ALL + np.repeat(n_second,len(v_data)).tolist()
        
        v_lat_ALL = v_lat_ALL + np.round(v_lat,decimals=3).tolist()
        v_lon_ALL = v_lon_ALL + np.round(v_lon,decimals=3).tolist()
        
        for DATATYPE in v_DATATYPES:
            v_data_ALL = l_data_ALL[DATATYPE]
            v_data = l_data[DATATYPE]
            v_data_ALL = v_data_ALL + Significant_Figure_Rounding(v_data,3).tolist()
            l_data_ALL[DATATYPE] = v_data_ALL
            
        # Print Progress
        print('Processed file ',fileList.index(fileName)+1,' of ',len(fileList))
        
    # Compile Output Data
    f_data = pd.DataFrame()
    f_data['Year'] = v_year_ALL
    f_data['Month'] = v_month_ALL
    f_data['Day'] = v_day_ALL
    f_data['Hour'] = v_hour_ALL
    f_data['Minute'] = v_minute_ALL
    # f_data['Second'] = v_second_ALL
    f_data['Lat'] = v_lat_ALL
    f_data['Lon'] = v_lon_ALL
    for DATATYPE in v_DATATYPES:
        v_data_ALL = l_data_ALL[DATATYPE]
        f_data[DATATYPE] = v_data_ALL
    
    # Save Output File
    f_data.to_csv(os.path.join(OUTPUT_FOLDER,OUTPUT_FILE), index = False)
    print(SUBSET_REGION + ' file saved. Processing complete.')

