# Import Packages
import os
import numpy as np
import pandas as pd
os.environ['PROJ_LIB'] = r'D:\Programs\Anaconda\pkgs\proj4-5.2.0-ha925a31_1\Library\share'
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
from scipy import interpolate

# Basic Settings
s_base = 'Roads_GulfCoast_Grid_001'
INPUT_FOLDER = 'ROADWAY DATA\\' + s_base
OUTPUT_FOLDER = 'ROADWAY DATA'
OUTPUT_FILENAME = s_base + '.csv'
PLOT = True

# Define Functions

# def Significant_Figure_Rounding(x_list,sigdec):
#     new_x = []
#     for x in x_list:
#         if (x == 0):
#             new_x = new_x + [0]
#         else:
#             exponent = np.floor(np.log10(np.abs(x)))
#             value = x/np.power(10,exponent)
#             new_value = np.round(value,decimals=sigdec)
#             new_x = new_x + [new_value*np.power(10,exponent)]
#     return new_x

def Significant_Figure_Rounding(x,sigdec):
    exponent = np.floor(np.log10(np.abs(x)))
    value = x/np.power(10,exponent)
    new_value = np.round(value,decimals=sigdec)
    new_x = new_value*np.power(10,exponent)
    new_x[np.isnan(new_x)] = 0
    return new_x


def Distance_Haversine(lat_1,lon_1,lat_2,lon_2):
    R = 6371 # radius of the earth in km
    lat_1 = np.radians(np.array(lat_1))
    lat_2 = np.radians(np.array(lat_2))
    lon_1 = np.radians(np.array(lon_1))
    lon_2 = np.radians(np.array(lon_2))
    delta_lat = lat_2 - lat_1
    delta_lon = lon_2 - lon_1
    a = (np.sin(delta_lat/2))*(np.sin(delta_lat/2)) + (np.cos(lat_1))*(np.cos(lat_2))*(np.sin(delta_lon/2))*(np.sin(delta_lon/2))
    # c = np.arctan2(np.sqrt(a),np.sqrt(1-a))
    c = np.arcsin(np.sqrt(a))
    d = 2*R*c
    return d
    # x,y = np.unravel_index(d.argmin(),d.shape)

# List Files
good_extension = '.csv'
fileList = os.listdir(INPUT_FOLDER)
fileList = [fileName for fileName in fileList if good_extension in fileName]



# Read Files
for fileName in fileList:
    f_csv = pd.read_csv(os.path.join(INPUT_FOLDER,fileName))
    
    s_CLASS = 'Class_' + fileName.split('_Class_')[1].split('.')[0]
    
    v_left = np.array(f_csv['left'])
    v_right = np.array(f_csv['right'])
    v_top = np.array(f_csv['top'])
    v_bottom = np.array(f_csv['bottom'])
    v_LENGTH = np.array(f_csv['LENGTH'])
    
    # Find Centers
    v_lat = np.add(v_top,v_bottom)/2
    v_lon = np.add(v_left,v_right)/2
    
    # Find Areas
    v_l_1 = Distance_Haversine(v_top,v_left,v_top,v_right)
    v_l_2 = Distance_Haversine(v_bottom,v_left,v_bottom,v_right)
    v_h = Distance_Haversine(v_bottom,v_lon,v_top,v_lon)
    v_area = v_h * (v_l_1 + v_l_2)/2
    
    v_density = v_LENGTH/v_area
    
    # Create data frame
    f_new = pd.DataFrame()
    f_new['Lat'] = np.round(v_lat,decimals=3)
    f_new['Lon'] = np.round(v_lon,decimals=3)
    f_new[s_CLASS] = Significant_Figure_Rounding(v_density,3)
    
    # Plotting Test
    if PLOT:
        # Regridding
        v_lat_grid = np.sort(np.unique(v_lat))
        v_lon_grid = np.sort(np.unique(v_lon))
        m_lon,m_lat = np.meshgrid(v_lon_grid,v_lat_grid)
        m_data = interpolate.griddata(np.c_[(v_lat,v_lon)], v_density, np.c_[m_lat.ravel(),m_lon.ravel()]).reshape(m_lon.shape)
        
        # Plotting
        m_data_masked = np.ma.masked_array(m_data, np.isnan(m_data))
        mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([np.nanmin(v_lat),-90]),urcrnrlat = np.min([np.nanmax(v_lat),90]),llcrnrlon=np.max([np.nanmin(v_lon),-180]),urcrnrlon = np.min([np.nanmax(v_lon),180]))
        mapobj.drawcoastlines(linewidth=0.5)
        mapobj.drawparallels(np.arange(-90., 120., 30.), labels=[1, 0, 0, 0])
        mapobj.drawmeridians(np.arange(-180, 180., 45.), labels=[0, 0, 0, 1])
        my_cmap = plt.cm.get_cmap('jet')
        my_cmap.set_under('w')
        mapobj.pcolormesh(m_lon,m_lat,m_data_masked,latlon=True,vmin=0*np.nanmin(m_data_masked),vmax=1*np.nanmax(m_data_masked),cmap=my_cmap)
        colorbarobj = mapobj.colorbar()
        colorbarobj.set_label('Density for ' + s_CLASS)
        # plt.autoscale()
        fig = plt.gcf()
        plt.show()
        
        
    # Combine
    if (fileList.index(fileName) == 0):
        f_combined = f_new
    else:
        f_combined = pd.merge(f_combined,f_new,how='outer',on=['Lat','Lon'])
    
    
    print('Processed',fileName)    
    
# Save
f_combined.to_csv(os.path.join(OUTPUT_FOLDER,OUTPUT_FILENAME), index = False)
print('File saved. Processing complete.')
