# Import Packages
import os
import datetime 
import numpy as np
from netCDF4 import Dataset
os.environ['PROJ_LIB'] = r'D:\Programs\Anaconda\pkgs\proj4-5.2.0-ha925a31_1\Library\share'
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
import pandas as pd

# Basic Settings
INPUT_FOLDER = 'TROPOMI DATA'
OUTPUT_FOLDER = 'TROPOMI DATA'
GRIDRES_DEG = np.nan
PLOT = False

v_SUBSET_REGIONS = ['NewYorkCity','Boston','NewOrleans','Houston','SaltLakeCity','Denver','SanFrancisco','LosAngeles','LasVegas']


# Function Definitions
def Regional_LatLon_Bounds(region_name):
    maxlat = np.inf
    maxlon = np.inf
    minlat = -np.inf
    minlon = -np.inf
    if region_name == 'CONUS':
        maxlat = 50 # dlat = 25
        maxlon = -65 # dlon = 60
        minlat = 25
        minlon = -125
    if region_name == 'SoCal':
        maxlat = 39 # dlat = 7
        maxlon = -115 # dlon = 8
        minlat = 32
        minlon = -123
    if region_name == 'UtCoWy':
        maxlat = 46 # dlat = 8
        maxlon = -104 # dlon = 9
        minlat = 38
        minlon = -113
    if region_name == 'NorthEast':
        maxlat = 43 # dlat = 5
        maxlon = -70 # dlon = 11
        minlat = 38
        minlon = -81
    if region_name == 'GulfCoast':
        maxlat = 35 # dlat = 7
        maxlon = -88 # dlon = 11
        minlat = 28
        minlon = -99
    if region_name == 'NewYorkCity':
        maxlat = 42 # dlat = 2
        maxlon = -73 # dlon = 2
        minlat = 40
        minlon = -75
    if region_name == 'Boston':
        maxlat = 43 # dlat = 2
        maxlon = -70 # dlon = 2
        minlat = 41
        minlon = -72
    if region_name == 'NewOrleans':
        maxlat = 31 # dlat = 2
        maxlon = -90 # dlon = 2
        minlat = 29
        minlon = -92
    if region_name == 'Houston':
        maxlat = 31 # dlat = 2
        maxlon = -97 # dlon = 2
        minlat = 29
        minlon = -99
    if region_name == 'SaltLakeCity':
        maxlat = 42 # dlat = 2
        maxlon = -111 # dlon = 2
        minlat = 40
        minlon = -113
    if region_name == 'Denver':
        maxlat = 41 # dlat = 2
        maxlon = -104 # dlon = 2
        minlat = 39
        minlon = -106
    if region_name == 'SanFrancisco':
        maxlat = 39 # dlat = 2
        maxlon = -121 # dlon = 2
        minlat = 37
        minlon = -123
    if region_name == 'LosAngeles':
        maxlat = 35 # dlat = 2
        maxlon = -116 # dlon = 4
        minlat = 33
        minlon = -120
    if region_name == 'LasVegas':
        maxlat = 37 # dlat = 2
        maxlon = -114 # dlon = 2
        minlat = 35
        minlon = -116
    return maxlat,maxlon,minlat,minlon

def Significant_Figure_Rounding(x,sigdec):
    exponent = np.floor(np.log10(np.abs(x)))
    value = x/np.power(10,exponent)
    new_value = np.round(value,decimals=sigdec)
    new_x = new_value*np.power(10,exponent)
    return new_x

for i_subset_region in range(len(v_SUBSET_REGIONS)):
    SUBSET_REGION = v_SUBSET_REGIONS[i_subset_region]
    OUTPUT_FILE = 'TROPOMI_NO2_Sept_2019_'+SUBSET_REGION+'.csv'
    
    
    # Determine Input Files
    good_extension = '.nc'
    fileList = os.listdir(INPUT_FOLDER)
    fileList = [fileName for fileName in fileList if good_extension in fileName]
    
    # Initialize Storage Arrays
    v_year_ALL = []
    v_month_ALL = []
    v_day_ALL = []
    v_hour_ALL = []
    v_minute_ALL = []
    # v_second_ALL = []
    
    v_lat_ALL = []
    v_lon_ALL = []
    v_data_ALL = []
    
    
    # Cycle Through Files
    for fileName in fileList:
        # Load Input File
        file = Dataset(os.path.join(INPUT_FOLDER,fileName), 'r')
    
        # Determine Input Data Types
        grp='PRODUCT'  
        lat = file.groups[grp].variables['latitude']
        n_fill_lat = lat._FillValue
        lon = file.groups[grp].variables['longitude']
        n_fill_lon = lon._FillValue
        data =file.groups[grp].variables['nitrogendioxide_tropospheric_column']
        n_fill_data = data._FillValue
        qa = file.groups[grp].variables['qa_value']
        n_fill_qa = qa._FillValue
        timestamps = file.groups[grp].variables['time_utc']
        n_fill_timestamps = timestamps._FillValue
        
        
          
        m_lat = np.array(lat[0][:][:])
        m_lon = np.array(lon[0][:][:])
        m_data = np.array(data[0][:][:])
        m_qa = np.array(qa[0][:][:])
        v_timestamps_1 = np.array(timestamps)
        
        file.close()
        
        m_lat[m_lat == n_fill_lat]=np.nan
        m_lon[m_lon == n_fill_lon]=np.nan
        m_data[m_data == n_fill_data]=np.nan
        m_qa[m_qa == n_fill_qa]=np.nan
        v_timestamps_1[v_timestamps_1 == n_fill_timestamps]=''
    
        # Handle Times
        # v_timestamps_2 = [datetime.datetime.strptime(s_timestamps, "%Y-%m-%dT%H:%M:%S.%fZ") for s_timestamps in v_timestamps_1[0]]
        v_timestamps_3 = [datetime.datetime.strptime(s_timestamps, "%Y-%m-%dT%H:%M:%S.%fZ").timestamp() for s_timestamps in v_timestamps_1[0]]
        # m_timestamps = np.transpose(np.repeat(np.array(v_timestamps_2,ndmin=2),m_lat.shape[1],0))
        m_timestamps = np.transpose(np.repeat(np.array(v_timestamps_3,ndmin=2),m_lat.shape[1],0))
    
    
        # datetime.datetime.fromtimestamp(m_timestamps[1,1])
        
    
    
    
        # Apply QA Checks
        m_data[m_qa < 0.5] = np.nan
    
        # Subsetting (Optional)
        n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)
        m_data[m_lat < n_min_lat]=np.nan
        m_data[m_lat > n_max_lat]=np.nan
        m_data[m_lon < n_min_lon]=np.nan
        m_data[m_lon > n_max_lon]=np.nan
        
    
        # Regridding (Optional)
        # if ~np.isnan(GRIDRES_DEG):
        
        # Verification Plot (Optional)
        if PLOT:
            m_data_masked = np.ma.masked_array(m_data, np.isnan(m_data))
            m_lat_masked = np.ma.masked_array(m_lat, np.isnan(m_data))
            m_lon_masked = np.ma.masked_array(m_lon, np.isnan(m_data))
            m_timestamps_masked = np.ma.masked_array(m_timestamps, np.isnan(m_data))
            mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([n_min_lat,-90]),urcrnrlat = np.min([n_max_lat,90]),llcrnrlon=np.max([n_min_lon,-180]),urcrnrlon = np.min([n_max_lon,180]))
            mapobj.drawcoastlines(linewidth=0.5)
            mapobj.drawparallels(np.arange(-90., 120., 30.), labels=[1, 0, 0, 0])
            mapobj.drawmeridians(np.arange(-180, 180., 45.), labels=[0, 0, 0, 1])
            my_cmap = plt.cm.get_cmap('jet')
            my_cmap.set_under('w')
            mapobj.pcolormesh(m_lon,m_lat,m_data_masked,latlon=True,vmin=0*np.nanmin(m_data_masked),vmax=0.1*np.nanmax(m_data_masked),cmap=my_cmap)
            # mapobj.pcolormesh(m_lon,m_lat,m_lat_masked,latlon=True,vmin=np.nanmin(m_lat[~np.isnan(m_data)]),vmax=np.nanmax(m_lat[~np.isnan(m_data)]),cmap=my_cmap)
            # mapobj.pcolormesh(m_lon,m_lat,m_lon_masked,latlon=True,vmin=np.nanmin(m_lon[~np.isnan(m_data)]),vmax=np.nanmax(m_lon[~np.isnan(m_data)]),cmap=my_cmap)
            # mapobj.pcolormesh(m_lon,m_lat,m_timestamps_masked,latlon=True,vmin=np.nanmin(m_timestamps[~np.isnan(m_data)]),vmax=np.nanmax(m_timestamps[~np.isnan(m_data)]),cmap=my_cmap)
            
            colorbarobj = mapobj.colorbar()
            # colorbarobj.set_label('Colorbar Label')
            # plt.autoscale()
            fig = plt.gcf()
            plt.show()
    
        # Rearrange matrices to Vectors
        v_timestamps = m_timestamps.flatten()
        v_lat = m_lat.flatten()
        v_lon = m_lon.flatten()
        v_data = m_data.flatten()
        v_qa = m_qa.flatten()
        
        # Trim Missing Data
        v_timestamps = v_timestamps[~np.isnan(v_data)]
        v_lat = v_lat[~np.isnan(v_data)]
        v_lon = v_lon[~np.isnan(v_data)]
        v_qa = v_qa[~np.isnan(v_data)]
        v_data = v_data[~np.isnan(v_data)]
                    
            
        # Concatenate to Outputs
        v_year_ALL = v_year_ALL + [datetime.datetime.fromtimestamp(n_timestamp).year for n_timestamp in v_timestamps.tolist()]
        v_month_ALL = v_month_ALL + [datetime.datetime.fromtimestamp(n_timestamp).month for n_timestamp in v_timestamps.tolist()]
        v_day_ALL = v_day_ALL + [datetime.datetime.fromtimestamp(n_timestamp).day for n_timestamp in v_timestamps.tolist()]
        v_hour_ALL = v_hour_ALL + [datetime.datetime.fromtimestamp(n_timestamp).hour for n_timestamp in v_timestamps.tolist()]
        v_minute_ALL = v_minute_ALL + [datetime.datetime.fromtimestamp(n_timestamp).minute for n_timestamp in v_timestamps.tolist()]
        # v_second_ALL = v_second_ALL + [datetime.datetime.fromtimestamp(n_timestamp).second for n_timestamp in v_timestamps.tolist()]
        
        v_lat_ALL = v_lat_ALL + np.round(v_lat,decimals=3).tolist()
        v_lon_ALL = v_lon_ALL + np.round(v_lon,decimals=3).tolist()
        v_data_ALL = v_data_ALL + Significant_Figure_Rounding(v_data,3).tolist()
            
        # Print Progress
        print('Processed file ',fileList.index(fileName)+1,' of ',len(fileList))
        
    # Compile Output Data
    f_data = pd.DataFrame()
    f_data['Year'] = v_year_ALL
    f_data['Month'] = v_month_ALL
    f_data['Day'] = v_day_ALL
    f_data['Hour'] = v_hour_ALL
    f_data['Minute'] = v_minute_ALL
    # f_data['Second'] = v_second_ALL
    f_data['Lat'] = v_lat_ALL
    f_data['Lon'] = v_lon_ALL
    f_data['NO2'] = v_data_ALL 
    
    # Save Output File
    f_data.to_csv(os.path.join(OUTPUT_FOLDER,OUTPUT_FILE), index = False)
    print(SUBSET_REGION+' file saved. Processing complete.')