# Import Packages
import os
import datetime
import numpy as np
import pandas as pd
os.environ['PROJ_LIB'] = r'D:\Programs\Anaconda\pkgs\proj4-5.2.0-ha925a31_1\Library\share'
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
from scipy import interpolate
import sys

from scipy.interpolate.interpnd import _ndim_coords_from_arrays
from scipy.spatial import cKDTree



# Settings
    
METHOD = 2

INPUT_FOLDER = 'TROPOMI DATA'
OUTPUT_FOLDER = 'TROPOMI DATA'
    
    
REGIONS = ['NewYorkCity','Boston','NewOrleans','Houston','SaltLakeCity','Denver','SanFrancisco','LosAngeles','LasVegas']


n_res_space = 0.05
s_res_space = '005'
s_res_time = 'hourly'


# Other Functions
def Regional_LatLon_Bounds(region_name):
    maxlat = np.inf
    maxlon = np.inf
    minlat = -np.inf
    minlon = -np.inf
    if region_name == 'CONUS':
        maxlat = 50 # dlat = 25
        maxlon = -65 # dlon = 60
        minlat = 25
        minlon = -125
    if region_name == 'SoCal':
        maxlat = 39 # dlat = 7
        maxlon = -115 # dlon = 8
        minlat = 32
        minlon = -123
    if region_name == 'UtCoWy':
        maxlat = 46 # dlat = 8
        maxlon = -104 # dlon = 9
        minlat = 38
        minlon = -113
    if region_name == 'NorthEast':
        maxlat = 43 # dlat = 5
        maxlon = -70 # dlon = 11
        minlat = 38
        minlon = -81
    if region_name == 'GulfCoast':
        maxlat = 35 # dlat = 7
        maxlon = -88 # dlon = 11
        minlat = 28
        minlon = -99
    if region_name == 'NewYorkCity':
        maxlat = 42 # dlat = 2
        maxlon = -73 # dlon = 2
        minlat = 40
        minlon = -75
    if region_name == 'Boston':
        maxlat = 43 # dlat = 2
        maxlon = -70 # dlon = 2
        minlat = 41
        minlon = -72
    if region_name == 'NewOrleans':
        maxlat = 31 # dlat = 2
        maxlon = -90 # dlon = 2
        minlat = 29
        minlon = -92
    if region_name == 'Houston':
        maxlat = 31 # dlat = 2
        maxlon = -97 # dlon = 2
        minlat = 29
        minlon = -99
    if region_name == 'SaltLakeCity':
        maxlat = 42 # dlat = 2
        maxlon = -111 # dlon = 2
        minlat = 40
        minlon = -113
    if region_name == 'Denver':
        maxlat = 41 # dlat = 2
        maxlon = -104 # dlon = 2
        minlat = 39
        minlon = -106
    if region_name == 'SanFrancisco':
        maxlat = 39 # dlat = 2
        maxlon = -121 # dlon = 2
        minlat = 37
        minlon = -123
    if region_name == 'LosAngeles':
        maxlat = 35 # dlat = 2
        maxlon = -116 # dlon = 4
        minlat = 33
        minlon = -120
    if region_name == 'LasVegas':
        maxlat = 37 # dlat = 2
        maxlon = -114 # dlon = 2
        minlat = 35
        minlon = -116
    return maxlat,maxlon,minlat,minlon

def Progressbar(it,tot,barsize):
    if it > 1:
        string = '\b'*(barsize+2)
        sys.stdout.write(string)
    n_full = int(np.floor(barsize*it/tot))
    n_empty = int(barsize-n_full)
    string = '['+ ('='*n_full) + (' '*n_empty) +']'
    sys.stdout.write(string)

for SUBSET_REGION in REGIONS:
    INPUT_FILENAME = 'TROPOMI_NO2_Sept_2019_' + SUBSET_REGION + '.csv'
    OUTPUT_FILENAME = 'TROPOMI_NO2_Sept_2019_' + SUBSET_REGION + '_Register_' + s_res_space + '_' + s_res_time + '.csv'
    
    # Load Files
    
    f_data = pd.read_csv(os.path.join(INPUT_FOLDER,INPUT_FILENAME))
    
    v_year = np.array(f_data['Year'])
    v_month = np.array(f_data['Month'])
    v_day = np.array(f_data['Day'])
    v_hour = np.array(f_data['Hour'])
    v_minute = np.array(f_data['Minute'])
    
    v_time = [datetime.datetime(v_year[line],v_month[line],v_day[line],v_hour[line],v_minute[line]) for line in range(len(v_year))]
    
    v_lat = np.array(f_data['Lat'])
    v_lon = np.array(f_data['Lon']) 
    
    m_coord = np.c_[v_lat,v_lon,v_time]
    
    v_data = np.array(f_data['NO2']) 
    
    if (METHOD == 1):
        # Registration 
        v_time_register = [datetime.datetime(v_year[line],v_month[line],v_day[line],v_hour[line],30) for line in range(len(v_year))]
        v_timestamp_register = np.array([n_timestamp.timestamp() for n_timestamp in v_time_register],dtype='i')
        
        v_lat_register = np.round(np.round(v_lat/n_res_space,decimals=0)*n_res_space,-int(np.floor(np.log10(n_res_space))))
        v_lon_register = np.round(np.round(v_lon/n_res_space,decimals=0)*n_res_space,-int(np.floor(np.log10(n_res_space))))
        m_coord_register = np.c_[v_lat_register,v_lon_register,v_timestamp_register]
        
        # Eliminate Duplicated Coordinates
        m_coord_unique = np.unique(m_coord_register,axis=0)
        v_data_unique = np.empty(m_coord_unique.shape[0])
        v_data_unique[:] = np.nan
        # for coord in m_coord_unique:
        #     index = (m_coord_unique == coord).all(axis=1)
        #     id_in_orig = (m_coord_register == coord).all(axis=1)
        #     v_data_unique[index] = np.nanmean(v_data[id_in_orig])
        #     Progressbar(int(np.where(index)[0]),int(m_coord_unique.shape[0]),50)
        for index in range(m_coord_unique.shape[0]):
            id_in_orig = (m_coord_register == m_coord_unique[index,:]).all(axis=1)
            v_data_unique[index] = np.nanmean(v_data[id_in_orig])
            Progressbar(int(index),int(m_coord_unique.shape[0]),50)
    elif (METHOD == 2):
        v_time_register = [datetime.datetime(v_year[line],v_month[line],v_day[line],v_hour[line],30) for line in range(len(v_year))]
        v_timestamp_register = np.array([n_timestamp.timestamp() for n_timestamp in v_time_register],dtype='i')
        
        # Define target grid
        n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)

        v_grid_mins = np.array([n_min_lat,n_min_lon])
        v_grid_maxs = np.array([n_max_lat,n_max_lon])
        v_grid_steps = np.array([n_res_space,n_res_space])
        THRESHOLD = n_res_space*np.sqrt(2)

        v_lat_grid = np.mgrid[v_grid_mins[0]:v_grid_maxs[0]:v_grid_steps[0]] + 0.5*v_grid_steps[0]
        v_lon_grid = np.mgrid[v_grid_mins[1]:v_grid_maxs[1]:v_grid_steps[1]] + 0.5*v_grid_steps[1]
        
        m_lon_target,m_lat_target = np.meshgrid(v_lon_grid,v_lat_grid) 
        
        m_coords_target = np.c_[m_lat_target.ravel(),m_lon_target.ravel()]
        
        m_coord_unique = np.empty([m_coords_target.shape[0]*len(np.unique(v_timestamp_register)),3])
        v_data_unique = np.empty(m_coords_target.shape[0]*len(np.unique(v_timestamp_register)))
        m_coord_unique[:] = np.nan
        m_coord_unique = m_coord_unique.astype('O')
        v_data_unique[:] = np.nan
        
        id_range = 0;
        for timestamp in np.unique(v_timestamp_register):
            id_timestamp = range(id_range*m_coords_target.shape[0],(id_range+1)*m_coords_target.shape[0])
            # m_coord_unique[id_timestamp,2] = datetime.datetime.fromtimestamp(timestamp)
            m_coord_unique[id_timestamp,2] = timestamp
            m_coord_unique[id_timestamp,0:2] = m_coords_target
            
            v_data_temp = interpolate.griddata(m_coord[timestamp == v_timestamp_register,0:2],v_data[timestamp == v_timestamp_register],m_coords_target,method='linear')
            tree = cKDTree(m_coord[timestamp == v_timestamp_register,0:2])
            xi = _ndim_coords_from_arrays(m_coords_target,ndim=2)
            dists,indexes = tree.query(xi)
            v_data_temp[dists > THRESHOLD] = np.nan
            
            # v_data_unique[id_timestamp] = interpolate.griddata(m_coord[timestamp == v_timestamp_register,0:2],v_data[timestamp == v_timestamp_register],m_coords_target,method='nearest')
            v_data_unique[id_timestamp] = v_data_temp
            
            id_range += 1
            Progressbar(int(id_range),int(len(np.unique(v_timestamp_register))),50)
            
        m_coord_unique = m_coord_unique[~np.isnan(v_data_unique),:]
        v_data_unique = v_data_unique[~np.isnan(v_data_unique)]
        
    # Save
    f_data_new = pd.DataFrame()
    f_data_new['Year'] = [datetime.datetime.fromtimestamp(n_timestamp).year for n_timestamp in m_coord_unique[:,2]]
    f_data_new['Month'] = [datetime.datetime.fromtimestamp(n_timestamp).month for n_timestamp in m_coord_unique[:,2]]
    f_data_new['Day'] = [datetime.datetime.fromtimestamp(n_timestamp).day for n_timestamp in m_coord_unique[:,2]]
    f_data_new['Hour'] = [datetime.datetime.fromtimestamp(n_timestamp).hour for n_timestamp in m_coord_unique[:,2]]
    f_data_new['Minute'] = [datetime.datetime.fromtimestamp(n_timestamp).minute for n_timestamp in m_coord_unique[:,2]]
    f_data_new['Lat'] = m_coord_unique[:,0].tolist()
    f_data_new['Lon'] = m_coord_unique[:,1].tolist()
    f_data_new['NO2'] = v_data_unique.tolist()
    
    # Save Output File
    f_data_new.to_csv(os.path.join(OUTPUT_FOLDER,OUTPUT_FILENAME),index = False)
    print('File saved.')
    
    # # Plotting Test
    # v_lat = np.unique(m_coord_unique[:,0])
    # v_lon = np.unique(m_coord_unique[:,1])
    # m_lat,m_lon = np.meshgrid(v_lat,v_lon)
    # m_latlon = np.c_[m_lat.ravel(),m_lon.ravel()]
    # v_data_plot = np.empty(m_latlon.shape[0])
    # v_data_plot[:] = np.nan
    # # for n_latlon in m_latlon:
    # #     index = (m_latlon == n_latlon).all(axis=1)
    # #     id_in_orig = (n_latlon == m_coord_unique[:,0:1]).all(axis=1)
    # #     v_data_plot[index] = np.nanmean(v_data_unique[id_in_orig])
    # #     Progressbar(int(np.where(index)[0]),int(m_latlon.shape[0]),50)
    # for line in range(m_latlon.shape[0]):
    #     id_in_orig = (m_latlon[line,:] == m_coord_unique[:,0:2]).all(axis=1)
    #     v_data_plot[line] = np.nanmean(v_data_unique[id_in_orig])
    #     Progressbar(int(line),int(m_latlon.shape[0]),50)
        
    # m_data_plot = v_data_plot.reshape(m_lat.shape)
    
    # n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)
        
    # m_data_masked = np.ma.masked_array(m_data_plot, np.isnan(m_data_plot))
    # mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([n_min_lat,-90]),urcrnrlat = np.min([n_max_lat,90]),llcrnrlon=np.max([n_min_lon,-180]),urcrnrlon = np.min([n_max_lon,180]))
    # mapobj.drawcoastlines(linewidth=0.5)
    # mapobj.drawparallels(np.arange(-90., 120., 30.), labels=[1, 0, 0, 0])
    # mapobj.drawmeridians(np.arange(-180, 180., 45.), labels=[0, 0, 0, 1])
    # my_cmap = plt.cm.get_cmap('jet')
    # my_cmap.set_under('w')
    # mapobj.pcolormesh(m_lon,m_lat,m_data_masked,latlon=True,vmin=0*np.nanmin(m_data_masked),vmax=1*np.nanmax(m_data_masked),cmap=my_cmap)
    # colorbarobj = mapobj.colorbar()
    # colorbarobj.set_label('NO2')
    # fig = plt.gcf()
    # plt.show()
    
    # Plotting Test
    
    m_coord_unique[:,2] = [datetime.datetime.fromtimestamp(n_timestamp) for n_timestamp in m_coord_unique[:,2]]
    
    mindex = pd.MultiIndex.from_tuples(tuple(m_coord_unique),names=['lat','lon','time'])
    series_data = pd.Series(v_data_unique,index = mindex)
    a_data = series_data.to_xarray()
    a_data.lat.attrs['units'] = 'deg N'
    a_data.lon.attrs['units'] = 'deg E'
    a_data.attrs['long_name'] = 'TROPOMI NO2'
    a_data.attrs['units'] = 'mol m-2'
    
    n_day_start = 7
    n_day_end = n_day_start+1
    mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([n_min_lat,-90]),urcrnrlat = np.min([n_max_lat,90]),llcrnrlon=np.max([n_min_lon,-180]),urcrnrlon = np.min([n_max_lon,180]))
    my_cmap = plt.cm.get_cmap('jet')
    mapobj.drawcoastlines(linewidth=0.5)
    a_data.loc[dict(time=slice(datetime.datetime(2019,9,n_day_start),datetime.datetime(2019,9,n_day_end)))].mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=0.0002)
    fig = plt.gcf()
    plt.show()

print('Processing Complete.')
