# Import Packages
import os
import pandas as pd
import requests

# Basic Settings
URL_BASE = 'https://ladsweb.modaps.eosdis.nasa.gov'
URL_CSV = 'NIGHT LIGHT DATA\\CONUS_Sept_2019.txt'
OUTPUT_FOLDER = 'NIGHT LIGHT DATA\\CONUS_Sept_2019'

f_URLs = pd.read_csv(URL_CSV)

v_URLs = f_URLs.iloc[:,1]

# overriding requests.Session.rebuild_auth to mantain headers when redirected

class SessionWithHeaderRedirection(requests.Session):
    AUTH_HOST = 'urs.earthdata.nasa.gov'
    def __init__(self, username, password):
        super().__init__()
        self.auth = (username, password)
    # Overrides from the library to keep headers when redirected to or from
    # the NASA auth host.
    def rebuild_auth(self, prepared_request, response):
        headers = prepared_request.headers
        url = prepared_request.url
        if 'Authorization' in headers:
            original_parsed = requests.utils.urlparse(response.request.url)
            redirect_parsed = requests.utils.urlparse(url)
            if (original_parsed.hostname != redirect_parsed.hostname) and \
                redirect_parsed.hostname != self.AUTH_HOST and \
                    original_parsed.hostname != self.AUTH_HOST:
                del headers['Authorization']
        return
    
# create session with the user credentials that will be used to authenticate access to the data
username = "USERNAME"
password= "PASSWORD"
session = SessionWithHeaderRedirection(username, password)
 

# Download 
for s_url in v_URLs:
    s_filename = s_url.split('/')[-1]
    s_savepath = os.path.join(OUTPUT_FOLDER,s_filename)
    #urllib.request.urlretrieve((URL_BASE + s_url),s_savepath)
    
    try:
        # submit the request using the session
        response = session.get((URL_BASE + s_url),stream=True)
        # print(response.status_code)
        # raise an exception in case of http errors
        response.raise_for_status()  
        # save the file
        with open(s_savepath,'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024*1024):
                fd.write(chunk)
    except requests.exceptions.HTTPError as e:
        # handle any errors here
        print(e)
        
    print("""\nDownloaded: %s to: %s""" %(s_url,s_savepath))
        
print('Downloads Complete')


