# Import Packages
import os
import datetime 
import numpy as np
# from netCDF4 import Dataset
import h5py
os.environ['PROJ_LIB'] = r'D:\Programs\Anaconda\pkgs\proj4-5.2.0-ha925a31_1\Library\share'
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
import pandas as pd
import re
from scipy import interpolate
from scipy.interpolate.interpnd import _ndim_coords_from_arrays
from scipy.spatial import cKDTree

# Basic Settings
INPUT_FOLDER = 'NIGHT LIGHT DATA\\CONUS_Sept_2019'
OUTPUT_FOLDER = 'NIGHT LIGHT DATA'
GRIDRES_TIME = 'Hourly'
GRIDRES_DEG = 0.005
s_res_space = '0005'
DOWNSCALE_METHOD = 'ADHOC'

v_SUBSET_REGIONS = ['NewYorkCity','Boston','NewOrleans','Houston','SaltLakeCity','Denver','SanFrancisco','LosAngeles','LasVegas']

v_SUBSET_REGIONS = ['NewYorkCity']


PLOT = False

# Function Definitions
def Regional_LatLon_Bounds(region_name):
    maxlat = np.inf
    maxlon = np.inf
    minlat = -np.inf
    minlon = -np.inf
    if region_name == 'CONUS':
        maxlat = 50 # dlat = 25
        maxlon = -65 # dlon = 60
        minlat = 25
        minlon = -125
    if region_name == 'SoCal':
        maxlat = 39 # dlat = 7
        maxlon = -115 # dlon = 8
        minlat = 32
        minlon = -123
    if region_name == 'UtCoWy':
        maxlat = 46 # dlat = 8
        maxlon = -104 # dlon = 9
        minlat = 38
        minlon = -113
    if region_name == 'NorthEast':
        maxlat = 43 # dlat = 5
        maxlon = -70 # dlon = 11
        minlat = 38
        minlon = -81
    if region_name == 'GulfCoast':
        maxlat = 35 # dlat = 7
        maxlon = -88 # dlon = 11
        minlat = 28
        minlon = -99
    if region_name == 'NewYorkCity':
        maxlat = 42 # dlat = 2
        maxlon = -73 # dlon = 2
        minlat = 40
        minlon = -75
    if region_name == 'Boston':
        maxlat = 43 # dlat = 2
        maxlon = -70 # dlon = 2
        minlat = 41
        minlon = -72
    if region_name == 'NewOrleans':
        maxlat = 31 # dlat = 2
        maxlon = -90 # dlon = 2
        minlat = 29
        minlon = -92
    if region_name == 'Houston':
        maxlat = 31 # dlat = 2
        maxlon = -97 # dlon = 2
        minlat = 29
        minlon = -99
    if region_name == 'SaltLakeCity':
        maxlat = 42 # dlat = 2
        maxlon = -111 # dlon = 2
        minlat = 40
        minlon = -113
    if region_name == 'Denver':
        maxlat = 41 # dlat = 2
        maxlon = -104 # dlon = 2
        minlat = 39
        minlon = -106
    if region_name == 'SanFrancisco':
        maxlat = 39 # dlat = 2
        maxlon = -121 # dlon = 2
        minlat = 37
        minlon = -123
    if region_name == 'LosAngeles':
        maxlat = 35 # dlat = 2
        maxlon = -116 # dlon = 4
        minlat = 33
        minlon = -120
    if region_name == 'LasVegas':
        maxlat = 37 # dlat = 2
        maxlon = -114 # dlon = 2
        minlat = 35
        minlon = -116
    return maxlat,maxlon,minlat,minlon

def Significant_Figure_Rounding(x,sigdec):
    exponent = np.floor(np.log10(np.abs(x)))
    value = x/np.power(10,exponent)
    new_value = np.round(value,decimals=sigdec)
    new_x = new_value*np.power(10,exponent)
    return new_x

def Matrix_Subset_by_index(matrix,rows,columns):
    row_index = np.array(rows)
    col_index = np.array(columns)
    return matrix[row_index[:,None],col_index]

for i_subset_region in range(len(v_SUBSET_REGIONS)):
    SUBSET_REGION = v_SUBSET_REGIONS[i_subset_region]
    
    if (GRIDRES_TIME == 'Hourly') & (~np.isnan(GRIDRES_DEG)):
        OUTPUT_FILE = 'VNP46A1_Sept_2019_'+SUBSET_REGION+'_Register_' + s_res_space + '_' + GRIDRES_TIME + '.csv'
    else:
        OUTPUT_FILE = 'VNP46A1_Sept_2019_'+SUBSET_REGION+'.csv'
    
    # Determine Input Files
    good_extension = '.h5'
    fileList = os.listdir(INPUT_FOLDER)
    fileList = [fileName for fileName in fileList if good_extension in fileName]
    
    # Initialize Storage Arrays
    v_year_ALL = []
    v_month_ALL = []
    v_day_ALL = []
    v_hour_ALL = []
    v_minute_ALL = []
    
    v_lat_ALL = []
    v_lon_ALL = []
    v_data_ALL = []
    
    for fileName in fileList:
        # Load Input File
        file = h5py.File(os.path.join(INPUT_FOLDER,fileName),'r')
    
        s_date = fileName.split('.')[1]
        
        year = np.int(s_date[1:5])
        daynumber = np.int(s_date[5:9])
        
        day = datetime.datetime(year,1,1) + datetime.timedelta(days=daynumber) - datetime.timedelta(days=1)
        # file.keys()
        
        # dset = file['HDFEOS/GRIDS']
        # dset.name
        
        # file['HDFEOS/GRIDS'].keys()
        
        # file['HDFEOS/GRIDS/VNP_Grid_DNB'].keys()
        
        # file['HDFEOS/GRIDS/VNP_Grid_DNB/Data Fields'].keys()
        
        m_hour = np.array(file['HDFEOS/GRIDS/VNP_Grid_DNB/Data Fields'].get('UTC_Time'))
        
        m_hour[m_hour < 0] = np.nan
        
        m_data = np.array(file['HDFEOS/GRIDS/VNP_Grid_DNB/Data Fields'].get('DNB_At_Sensor_Radiance_500m'))
        
        m_data =m_data.astype(np.float32)
        m_data[m_data >= 65534] = np.nan
        m_data = 0.1*m_data
        
        fileMetadata = file['HDFEOS INFORMATION/StructMetadata.0'][()].split()
        fileMetadata = [m.decode('utf-8') for m in fileMetadata]
        
        ul_regex = re.compile(r'''UpperLeftPointMtrs=\(
                                  (?P<upper_left_x>[+-]?\d+\.\d+)
                                  ,
                                  (?P<upper_left_y>[+-]?\d+\.\d+)
                                  \)''', re.VERBOSE)
        
        match = ul_regex.search(fileMetadata[7])
        x0 = np.float(match.group('upper_left_x'))
        y0 = np.float(match.group('upper_left_y'))
        
        lr_regex = re.compile(r'''LowerRightMtrs=\(
                                  (?P<lower_right_x>[+-]?\d+\.\d+)
                                  ,
                                  (?P<lower_right_y>[+-]?\d+\.\d+)
                                  \)''', re.VERBOSE)
        match = lr_regex.search(fileMetadata[8])
        x1 = np.float(match.group('lower_right_x'))
        y1 = np.float(match.group('lower_right_y'))
        
        # nx, ny = m_data.shape
        # x = np.linspace(x0, x1, nx)
        # y = np.linspace(y0, y1, ny)
        # xv, yv = np.meshgrid(x, y)
        
        # sinu = pyproj.Proj("+proj=sinu +R=6371007.181 +nadgrids=@null +wktext")
        # wgs84 = pyproj.Proj("+init=EPSG:4326") 
        # m_lon,m_lat= pyproj.transform(sinu, wgs84, xv, yv)
        
        nx, ny = m_data.shape
        x = np.linspace(x0/1000000, x1/1000000, nx)
        y = np.linspace(y0/1000000, y1/1000000, ny)
        
        m_lon, m_lat = np.meshgrid(x, y)
        
        # m_lon[m_lon < 0] += 360
        
        # Subsetting (Optional)
        n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)
        m_data[m_lat < n_min_lat]=np.nan
        m_data[m_lat > n_max_lat]=np.nan
        m_data[m_lon < n_min_lon]=np.nan
        m_data[m_lon > n_max_lon]=np.nan
        
        # if PLOT:
        #     mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max(m_lat),urcrnrlat = np.min(m_lat),llcrnrlon=np.max(m_lon),urcrnrlon = np.min(m_lon))
        #     mapobj.drawcoastlines(linewidth=0.5)
        #     mapobj.drawparallels(np.arange(-90., 120., 30.), labels=[1, 0, 0, 0])
        #     mapobj.drawmeridians(np.arange(-180, 180., 45.), labels=[0, 0, 0, 1])
        #     my_cmap = plt.cm.get_cmap('jet')
        #     my_cmap.set_under('w')
        #     mapobj.pcolormesh(m_lon,m_lat,m_data,latlon=True,vmin=1*np.nanmin(m_data),vmax=1*np.nanmax(m_data),cmap=my_cmap)
        #     colorbarobj = mapobj.colorbar()
        #     fig = plt.gcf()
        #     plt.show()
            
        if PLOT:
            m_data_masked = np.ma.masked_array(m_data, np.isnan(m_data))
            m_lat_masked = np.ma.masked_array(m_lat, np.isnan(m_data))
            m_lon_masked = np.ma.masked_array(m_lon, np.isnan(m_data))
            mapobj = Basemap(projection='cyl',resolution='l',llcrnrlat=np.max([n_min_lat,-90]),urcrnrlat = np.min([n_max_lat,90]),llcrnrlon=np.max([n_min_lon,-180]),urcrnrlon = np.min([n_max_lon,180]))
            mapobj.drawcoastlines(linewidth=0.5)
            mapobj.drawparallels(np.arange(-90., 120., 30.), labels=[1, 0, 0, 0])
            mapobj.drawmeridians(np.arange(-180, 180., 45.), labels=[0, 0, 0, 1])
            my_cmap = plt.cm.get_cmap('jet')
            my_cmap.set_under('w')
            mapobj.pcolormesh(m_lon,m_lat,m_data_masked,latlon=True,vmin=0*np.nanmin(m_data_masked),vmax=0.1*np.nanmax(m_data_masked),cmap=my_cmap)
            # mapobj.pcolormesh(m_lon,m_lat,m_lat_masked,latlon=True,vmin=np.nanmin(m_lat[~np.isnan(m_data)]),vmax=np.nanmax(m_lat[~np.isnan(m_data)]),cmap=my_cmap)
            # mapobj.pcolormesh(m_lon,m_lat,m_lon_masked,latlon=True,vmin=np.nanmin(m_lon[~np.isnan(m_data)]),vmax=np.nanmax(m_lon[~np.isnan(m_data)]),cmap=my_cmap)
            # mapobj.pcolormesh(m_lon,m_lat,m_timestamps_masked,latlon=True,vmin=np.nanmin(m_timestamps[~np.isnan(m_data)]),vmax=np.nanmax(m_timestamps[~np.isnan(m_data)]),cmap=my_cmap)
            
            colorbarobj = mapobj.colorbar()
            # colorbarobj.set_label('Colorbar Label')
            # plt.autoscale()
            fig = plt.gcf()
            plt.show()
        
        v_year  = day.year*np.ones(nx*ny)
        v_month = day.month*np.ones(nx*ny)
        v_day = day.day*np.ones(nx*ny)
        
        v_hour = np.floor(m_hour.flatten())
        v_minute = np.floor(60*(m_hour.flatten() - v_hour))
        
        
        v_lat = m_lat.flatten()
        v_lon = m_lon.flatten()
        v_data = m_data.flatten()
        
        # strip missing data
        id_keep = ~(np.isnan(np.c_[v_year,v_month,v_day,v_hour,v_minute,v_lat,v_lon,v_data]).any(axis=1))
        v_year = v_year[id_keep]
        v_month = v_month[id_keep]
        v_day = v_day[id_keep]
        v_hour = v_hour[id_keep]
        v_minute = v_minute[id_keep]
        v_lat = v_lat[id_keep]
        v_lon = v_lon[id_keep]
        v_data = v_data[id_keep]
        
        if (len(v_data) > 0):
        
            # Register coordinates to regular grid
            if (~np.isnan(GRIDRES_DEG)):
                if (DOWNSCALE_METHOD == 'ADHOC'):
                    n_cell_interval = int(np.round(GRIDRES_DEG/0.005,decimals=0))
                    # n_cell_interval = 2
                    
                    m_lat_temp = m_lat.reshape((m_lat.shape[0]//n_cell_interval,n_cell_interval,m_lat.shape[1]//n_cell_interval,n_cell_interval))
                    m_lat_new = np.nanmean(m_lat_temp,axis=(1,3))
                    
                    m_lon_temp = m_lon.reshape((m_lon.shape[0]//n_cell_interval,n_cell_interval,m_lon.shape[1]//n_cell_interval,n_cell_interval))
                    m_lon_new = np.nanmean(m_lon_temp,axis=(1,3))
                    
                    m_hour_temp = m_data.reshape((m_hour.shape[0]//n_cell_interval,n_cell_interval,m_hour.shape[1]//n_cell_interval,n_cell_interval))
                    m_hour_new = np.nanmean(m_hour_temp,axis=(1,3))
    
                    m_data_temp = m_data.reshape((m_data.shape[0]//n_cell_interval,n_cell_interval,m_data.shape[1]//n_cell_interval,n_cell_interval))
                    m_data_new = np.nanmean(m_data_temp,axis=(1,3))
                    
                    m_lat = m_lat_new
                    m_lon = m_lon_new
                    m_data = m_data_new
                    m_hour = m_hour_new
                    
                    v_lat = m_lat.flatten()
                    v_lon = m_lon.flatten()
                    v_data = m_data.flatten()
                    
                    v_lon = np.round(np.round(v_lon/GRIDRES_DEG,decimals=0)*GRIDRES_DEG,-int(np.floor(np.log10(GRIDRES_DEG))))
                    v_lat = np.round(np.round(v_lat/GRIDRES_DEG,decimals=0)*GRIDRES_DEG,-int(np.floor(np.log10(GRIDRES_DEG))))
                    
                    v_year  = day.year*np.ones(len(v_data))
                    v_month = day.month*np.ones(len(v_data))
                    v_day = day.day*np.ones(len(v_data))
                    
                    v_hour = np.floor(m_hour.flatten())
                    v_hour[v_hour < 0] = np.nan
                    v_hour[v_hour > 23] = np.nan
                    v_minute = np.floor(60*(m_hour.flatten() - v_hour))
                    
                    # strip missing data
                    id_keep = ~(np.isnan(np.c_[v_year,v_month,v_day,v_hour,v_minute,v_lat,v_lon,v_data]).any(axis=1))
                    v_year = v_year[id_keep]
                    v_month = v_month[id_keep]
                    v_day = v_day[id_keep]
                    v_hour = v_hour[id_keep]
                    v_minute = v_minute[id_keep]
                    v_lat = v_lat[id_keep]
                    v_lon = v_lon[id_keep]
                    v_data = v_data[id_keep]
                    
                    
                    
                else:
                    # v_lat_register = np.round(np.round(v_lat/GRIDRES_DEG,decimals=0)*GRIDRES_DEG,-int(np.floor(np.log10(GRIDRES_DEG))))
                    # v_lon_register = np.round(np.round(v_lon/GRIDRES_DEG,decimals=0)*GRIDRES_DEG,-int(np.floor(np.log10(GRIDRES_DEG))))
                    # m_coord_register = np.c_[v_lat_register,v_lon_register]
                    
                    # # Eliminate Duplicated Coordinates
                    # m_coord_unique = np.unique(m_coord_register,axis=0)
                    # v_data_unique = np.empty(m_coord_unique.shape[0])
                    # v_data_unique[:] = np.nan
                    # # for coord in m_coord_unique:
                    # #     index = (m_coord_unique == coord).all(axis=1)
                    # #     id_in_orig = (m_coord_register == coord).all(axis=1)
                    # #     v_data_unique[index] = np.nanmean(v_data[id_in_orig])
                    # #     Progressbar(int(np.where(index)[0]),int(m_coord_unique.shape[0]),50)
                    # for index in range(m_coord_unique.shape[0]):
                    #     id_in_orig = (m_coord_register == m_coord_unique[index,:]).all(axis=1)
                    #     v_data_unique[index] = np.nanmean(v_data[id_in_orig])
                
                    # v_lat = v_lat_register
                    # v_lon = v_lon_register
                    # v_data = v_data_unique
                    
                    # Define target grid
                    n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)
            
                    v_grid_mins = np.array([n_min_lat,n_min_lon])
                    v_grid_maxs = np.array([n_max_lat,n_max_lon])
                    v_grid_steps = np.array([GRIDRES_DEG,GRIDRES_DEG])
                    THRESHOLD = GRIDRES_DEG*np.sqrt(2)
            
                    v_lat_grid = np.mgrid[v_grid_mins[0]:v_grid_maxs[0]:v_grid_steps[0]] + 0.5*v_grid_steps[0]
                    v_lon_grid = np.mgrid[v_grid_mins[1]:v_grid_maxs[1]:v_grid_steps[1]] + 0.5*v_grid_steps[1]
                    
                    m_lon_target,m_lat_target = np.meshgrid(v_lon_grid,v_lat_grid) 
                    
                    m_coords_target = np.c_[m_lat_target.ravel(),m_lon_target.ravel()]
                    
                    v_data_unique = np.empty(m_coords_target.shape[0])
                    v_data_unique[:] = np.nan
                    
                    m_coord_unique = m_coords_target
                    m_coord = np.c_[v_lat,v_lon]
                    
                    v_data_temp = interpolate.griddata(m_coord,v_data,m_coords_target,method='linear')
                    tree = cKDTree(m_coord)
                    xi = _ndim_coords_from_arrays(m_coords_target,ndim=2)
                    dists,indexes = tree.query(xi)
                    v_data_temp[dists > THRESHOLD] = np.nan
                    
                    v_hour_temp = interpolate.griddata(m_coord,v_hour+(v_minute/60),m_coords_target,method='linear')
                    v_hour_temp[dists > THRESHOLD] = np.nan
                    
                    v_data_unique = v_data_temp
                    v_lat_unique = m_lat_target.ravel()
                    v_lon_unique = m_lon_target.ravel()
                    
                    v_lat_unique = v_lat_unique[~np.isnan(v_data_unique)]
                    v_lon_unique = v_lon_unique[~np.isnan(v_data_unique)]
                    v_hour_new = v_hour_temp[~np.isnan(v_data_unique)]
                    v_data_unique = v_data_unique[~np.isnan(v_data_unique)]
                    
                    
                    v_lat = v_lat_unique
                    v_lon = v_lon_unique
                    v_data = v_data_unique
                    
                    v_year  = day.year*np.ones(len(v_data))
                    v_month = day.month*np.ones(len(v_data))
                    v_day = day.day*np.ones(len(v_data))
                    
                    v_hour = np.floor(v_hour_new)
                    v_minute = np.floor(60*(v_hour_new - v_hour))
              
        if (len(v_data) > 0):
            # Register times to nearest hour (at half-hour intervals)
            if (GRIDRES_TIME == 'Hourly'):
                v_minute = 30*np.ones(len(v_data))
         
        if (len(v_data) > 0):
            # Concatenate to Outputs
            v_year_ALL = v_year_ALL + v_year.astype(np.int).tolist()
            v_month_ALL = v_month_ALL + v_month.astype(np.int).tolist()
            v_day_ALL = v_day_ALL + v_day.astype(np.int).tolist()
            v_hour_ALL = v_hour_ALL + v_hour.astype(np.int).tolist()
            v_minute_ALL = v_minute_ALL + v_minute.astype(np.int).tolist()
            
            v_lat_ALL = v_lat_ALL + np.round(v_lat,decimals=3).tolist()
            v_lon_ALL = v_lon_ALL + np.round(v_lon,decimals=3).tolist()
            v_data_ALL = v_data_ALL + Significant_Figure_Rounding(v_data,3).tolist()
            
        # Print Progress
        print('Processed file ',fileList.index(fileName)+1,' of ',len(fileList))
        
       
    # Compile Output Data
    f_data = pd.DataFrame()
    f_data['Year'] = v_year_ALL
    f_data['Month'] = v_month_ALL
    f_data['Day'] = v_day_ALL
    f_data['Hour'] = v_hour_ALL
    f_data['Minute'] = v_minute_ALL
    f_data['Lat'] = v_lat_ALL
    f_data['Lon'] = v_lon_ALL
    f_data['DNB'] = v_data_ALL 
    
    # Save Output File
    f_data.to_csv(os.path.join(OUTPUT_FOLDER,OUTPUT_FILE), index = False)
    print(SUBSET_REGION+' file saved. Processing complete.')  