#%% Basic Settings
SUBSET_REGION = 'SanFrancisco'
S_TIMEFRAME_1 = 'Sept_2019'
n_days_calibration = 7
n_time_resolution_hrs = 1
# v_prediction_lead_times_hrs = [1,3,6,12,18,24]
v_prediction_lead_times_hrs = [0,1,3,6,9,12,15,18,21,24]
n_lead_time_hrs = v_prediction_lead_times_hrs[0]
MAIN_PATH = 'Data Archive'
OUTPUT_FIGURE_FOLDER = MAIN_PATH + '\\Generated Figures'
OUTPUT_DATA_FOLDER = MAIN_PATH + '\\Generated Data'

SHOW_PLOT = False
UN_NORMED = True
NORMALIZE = True
RESIDUALS = False

#%% Load Packages and Define Functions

import os
import datetime as dt
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import interpolate
from scipy import stats
import sys
import xarray as xr
from sklearn import linear_model 
from sklearn import tree
import cartopy

# Define Functions
def Progressbar(it,tot,barsize):
    if it > 1:
        string = '\b'*(barsize+2)
        sys.stdout.write(string)
    n_full = int(np.floor(barsize*it/tot))
    n_empty = int(barsize-n_full)
    string = '['+ ('='*n_full) + (' '*n_empty) +']'
    sys.stdout.write(string)

def Regional_LatLon_Bounds(region_name):
    maxlat = np.inf
    maxlon = np.inf
    minlat = -np.inf
    minlon = -np.inf
    if region_name == 'CONUS':
        maxlat = 50 # dlat = 25
        maxlon = -65 # dlon = 60
        minlat = 25
        minlon = -125
    if region_name == 'SoCal':
        maxlat = 39 # dlat = 7
        maxlon = -115 # dlon = 8
        minlat = 32
        minlon = -123
    if region_name == 'UtCoWy':
        maxlat = 46 # dlat = 8
        maxlon = -104 # dlon = 9
        minlat = 38
        minlon = -113
    if region_name == 'NorthEast':
        maxlat = 43 # dlat = 5
        maxlon = -70 # dlon = 11
        minlat = 38
        minlon = -81
    if region_name == 'GulfCoast':
        maxlat = 35 # dlat = 7
        maxlon = -88 # dlon = 11
        minlat = 28
        minlon = -99
    if region_name == 'NewYorkCity':
        maxlat = 42 # dlat = 2
        maxlon = -73 # dlon = 2
        minlat = 40
        minlon = -75
    if region_name == 'Boston':
        maxlat = 43 # dlat = 2
        maxlon = -70 # dlon = 2
        minlat = 41
        minlon = -72
    if region_name == 'NewOrleans':
        maxlat = 31 # dlat = 2
        maxlon = -90 # dlon = 2
        minlat = 29
        minlon = -92
    if region_name == 'Houston':
        maxlat = 31 # dlat = 2
        maxlon = -97 # dlon = 2
        minlat = 29
        minlon = -99
    if region_name == 'SaltLakeCity':
        maxlat = 42 # dlat = 2
        maxlon = -111 # dlon = 2
        minlat = 40
        minlon = -113
    if region_name == 'Denver':
        maxlat = 41 # dlat = 2
        maxlon = -104 # dlon = 2
        minlat = 39
        minlon = -106
    if region_name == 'SanFrancisco':
        maxlat = 39 # dlat = 2
        maxlon = -121 # dlon = 2
        minlat = 37
        minlon = -123
    if region_name == 'LosAngeles':
        maxlat = 35 # dlat = 2
        maxlon = -116 # dlon = 4
        minlat = 33
        minlon = -120
    if region_name == 'LasVegas':
        maxlat = 37 # dlat = 2
        maxlon = -114 # dlon = 2
        minlat = 35
        minlon = -116
    return maxlat,maxlon,minlat,minlon

def Distance_Haversine(lat_1,lon_1,lat_2,lon_2):
    R = 6371 # radius of the earth in km
    lat_1 = np.radians(np.array(lat_1))
    lat_2 = np.radians(np.array(lat_2))
    lon_1 = np.radians(np.array(lon_1))
    lon_2 = np.radians(np.array(lon_2))
    delta_lat = lat_2 - lat_1
    delta_lon = lon_2 - lon_1
    a = (np.sin(delta_lat/2))*(np.sin(delta_lat/2)) + (np.cos(lat_1))*(np.cos(lat_2))*(np.sin(delta_lon/2))*(np.sin(delta_lon/2))
    # c = np.arctan2(np.sqrt(a),np.sqrt(1-a))
    c = np.arcsin(np.sqrt(a))
    d = 2*R*c
    return d

def Leave_One_Out_By_Location(series_IN):
    l_coords = [latlon for latlon,singleseries in list(series_IN.groupby(['lat','lon']))]
    m_coords = np.asarray(l_coords)
    l_series_single = [singleseries for latlon,singleseries in list(series_IN.groupby(['lat','lon']))]
    l_series_else = []
    for index in range(len(l_series_single)):
        l_series_single_copy = l_series_single.copy()
        l_series_single_copy.pop(index)
        serieselse = pd.concat(l_series_single_copy)
        l_series_else.append(serieselse)
    return l_series_single,l_series_else,m_coords

def Get_Locations_with_Data(a_data):
    m_data = (~np.isnan(a_data)).any(dim='time')
    m_coords = np.empty([0,2])
    for i_lat in m_data.lat:
        for i_lon in m_data.lon:
            if m_data.sel(lat=i_lat,lon=i_lon):
                m_coords = np.r_[m_coords,np.array([i_lat,i_lon],ndmin=2)]          
    return m_coords

def Array_to_Matrix_at_Coords(a_data,m_coords):
    n_coords = m_coords.shape[0]
    if hasattr(a_data,'time'):
        n_times = len(a_data.time)
    else:
        n_times = 1
    m_data = np.empty([n_times,n_coords])
    m_data[:] = np.nan
    for i_coord in range(n_coords):
        m_data[:,i_coord] = np.array(a_data.sel(lat=m_coords[i_coord,0],lon=m_coords[i_coord,1]))
    return m_data

def Calibrate_Gaussian(a_data,m_coords):
    n_coords = m_coords.shape[0]
    
    # restrict data to m_coords
    m_data = Array_to_Matrix_at_Coords(a_data,m_coords)
    
    # Compute means and covariance
    v_mean = np.nanmean(m_data,axis=0)
    m_covariance = np.zeros([n_coords,n_coords])
    
    for i_1 in range(n_coords):
        v_1 = m_data[:,i_1]
        for i_2 in range(i_1,n_coords):
            v_2 = m_data[:,i_2]
            id_num = (~np.isnan(np.c_[v_1,v_2])).all(axis=1)
            if (sum(id_num) > 1):
                m_covariance[i_1,i_2] = np.cov(np.c_[v_1[id_num],v_2[id_num]],rowvar=False)[0,1]
                m_covariance[i_2,i_1] = m_covariance[i_1,i_2]
          
    return v_mean,m_covariance

def Rescale_A2_to_A1(a_data_1,a_data_2):
    n_slope = np.nan
    n_intercept = np.nan
    a_data_2_reshape = a_data_2.interp(time=a_data_2.time,lat=a_data_1.lat,lon=a_data_1.lon,method='nearest')
    a_data_1_reshape = a_data_1.interp(time=a_data_2.time,lat=a_data_1.lat,lon=a_data_1.lon,method='nearest')
    STARTED = 0
    for i_time in a_data_2.time:
        v_1 = np.array(a_data_1_reshape.sel(time=i_time)).reshape(-1)
        v_2 = np.array(a_data_2_reshape.sel(time=i_time)).reshape(-1)
        id_num = (~np.isnan(np.c_[v_1,v_2])).all(axis=1)
        if (sum(id_num) > 1):
            regr_temp = linear_model.LinearRegression()
            regr_temp.fit(v_2[id_num].reshape(-1,1),v_1[id_num].reshape(-1,1)) 
            n_slope = regr_temp.coef_[0][0]
            n_intercept = regr_temp.intercept_[0]
        a_TEMP = ((n_slope*a_data_2.sel(time=i_time)) + n_intercept).expand_dims('time')
        if (STARTED == 0):
            a_data_2_rescale = a_TEMP
            STARTED = 1
        else:
            a_data_2_rescale = xr.concat([a_data_2_rescale,a_TEMP],dim='time')
    return a_data_2_rescale

def Matrix_Subset_by_index(matrix,rows,columns):
    row_index = np.array(rows)
    col_index = np.array(columns)
    return matrix[row_index[:,None],col_index]

def Matrix_Subset_by_Boolean(matrix,rows,columns):
    rows_index = np.where(rows)[0]
    columns_index = np.where(columns)[0]
    matrix_subset = Matrix_Subset_by_index(matrix,rows_index,columns_index)
    return matrix_subset

def Gaussian_Update(v_mean,m_covariance,m_coords,m_coords_obs,v_values_obs,m_noise_obs,m_coords_target):
    # handle nan observations
    id_obs_num = ~np.isnan(v_values_obs)
    v_values_obs = v_values_obs[id_obs_num]
    m_coords_obs = m_coords_obs[id_obs_num,:]
    m_noise_obs = Matrix_Subset_by_Boolean(m_noise_obs,id_obs_num,id_obs_num)
    
    n_obs = m_coords_obs.shape[0]
    id_obs = [np.where((m_coords == m_coords_obs[i_obs,:]).all(axis=1))[0][0] for i_obs in range(n_obs)]
    
    n_target = m_coords_target.shape[0]
    id_target = [np.where((m_coords == m_coords_target[i_target,:]).all(axis=1))[0][0] for i_target in range(n_target)]
    
    m_covariance_target = Matrix_Subset_by_index(m_covariance,id_target,id_target)
    m_covariance_obs = Matrix_Subset_by_index(m_covariance,id_obs,id_obs)
    m_covariance_target_obs = Matrix_Subset_by_index(m_covariance,id_target,id_obs)
    m_covaraince_obs_target = Matrix_Subset_by_index(m_covariance,id_obs,id_target)
    
    m_update = np.matmul(m_covariance_target_obs,np.linalg.inv(m_covariance_obs + m_noise_obs))
    
    m_covariance_target = m_covariance_target - np.matmul(m_update,m_covaraince_obs_target)
    v_mean_target = v_mean[id_target] + np.matmul(m_update,(v_values_obs - v_mean[id_obs]))
    
    return v_mean_target,m_covariance_target
    
#%% Load Data
# Initial Definitions 
n_max_lat,n_max_lon,n_min_lat,n_min_lon = Regional_LatLon_Bounds(SUBSET_REGION)

v_CITY_REGIONS = ['NewYorkCity','Boston','NewOrleans','Houston','SaltLakeCity','Denver','SanFrancisco','LosAngeles','LasVegas']

# Import EPA Dataset
INPUT_FOLDER_EPA = MAIN_PATH + '\\EPA DATA'
INPUT_FILE_EPA = 'EPA_NO2_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '.csv'
TYPE_EPA = 'NO2'
UNITS_EPA = 'ppb'
LABEL_EPA = 'EPA NO2'

f_data_EPA = pd.read_csv(os.path.join(INPUT_FOLDER_EPA,INPUT_FILE_EPA))
v_year_EPA = np.array(f_data_EPA['Year'])
v_month_EPA = np.array(f_data_EPA['Month'])
v_day_EPA = np.array(f_data_EPA['Day'])
v_hour_EPA = np.array(f_data_EPA['Hour'])
v_minute_EPA = np.array(f_data_EPA['Minute'])
v_time_EPA = [dt.datetime(v_year_EPA[line],v_month_EPA[line],v_day_EPA[line],v_hour_EPA[line],v_minute_EPA[line]) for line in range(len(v_year_EPA))]
v_lat_EPA = np.array(f_data_EPA['Lat'])
v_lon_EPA = np.array(f_data_EPA['Lon']) 
m_coord_EPA = np.c_[v_lat_EPA,v_lon_EPA,v_time_EPA]
v_data_EPA = np.array(f_data_EPA[TYPE_EPA]) 
mindex_EPA = pd.MultiIndex.from_tuples(tuple(m_coord_EPA),names=['lat','lon','time'])
series_data_EPA = pd.Series(v_data_EPA,index = mindex_EPA)
# Correct non-unique indeces
series_data_EPA = series_data_EPA.groupby(['lat','lon','time']).mean()
a_data_EPA = series_data_EPA.to_xarray()
a_data_EPA.lat.attrs['units'] = 'deg N'
a_data_EPA.lon.attrs['units'] = 'deg E'
a_data_EPA.attrs['long_name'] = TYPE_EPA
a_data_EPA.attrs['units'] = UNITS_EPA

# Dataset 1 Test Plots
mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
mapobj.coastlines();
mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
# gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines.top_labels = False
gridlines.right_labels = False
my_cmap = plt.cm.get_cmap('Blues')
m_valid_locations = ~(np.isnan(a_data_EPA).all(dim='time'))
v_lat_plot = np.empty(len(m_valid_locations.lat))
v_lon_plot = np.empty(len(m_valid_locations.lat))
v_data_temp = np.empty(len(m_valid_locations.lat))
v_data_temp[:] = np.nan
v_lat_plot[:] = np.nan
v_lon_plot[:] = np.nan
index = 0
for latval in m_valid_locations.lat:
    lonval = m_valid_locations.lon[m_valid_locations.loc[dict(lat=latval)]]
    v_lat_plot[index] = latval
    v_lon_plot[index] = lonval
    # v_data_temp[index] = a_data_EPA.loc[dict(lat=latval,lon=lonval,time=slice(dt.datetime(2019,9,1),dt.datetime(2019,10,1)))].mean(dim='time')
    v_data_temp[index] = a_data_EPA.loc[dict(lat=latval,lon=lonval)].mean(dim='time')
    index += 1
plt.scatter(x=v_lon_plot,y=v_lat_plot,c=v_data_temp,vmin=0,vmax=20)
plt.colorbar().set_label('Monthly Average NO2 [ppb]')
# plt.xlabel('lon [deg E]')
# plt.ylabel('lat [deg N]')
plt.title(SUBSET_REGION + ' EPA Sites')
figurename = SUBSET_REGION + ' Overview' + '.png'
plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
plt.show()



# EPA Data Coverage
mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
mapobj.coastlines();
mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
# gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines.top_labels = False
gridlines.right_labels = False
my_cmap = plt.cm.get_cmap('Blues')
m_valid_locations = ~(np.isnan(a_data_EPA).all(dim='time'))
v_lat_plot = np.empty(len(m_valid_locations.lat))
v_lon_plot = np.empty(len(m_valid_locations.lat))
v_data_temp = np.empty(len(m_valid_locations.lat))
v_data_temp[:] = np.nan
v_lat_plot[:] = np.nan
v_lon_plot[:] = np.nan
index = 0
for latval in m_valid_locations.lat:
    lonval = m_valid_locations.lon[m_valid_locations.loc[dict(lat=latval)]]
    v_lat_plot[index] = latval
    v_lon_plot[index] = lonval
    # v_data_temp[index] = a_data_EPA.loc[dict(lat=latval,lon=lonval,time=slice(dt.datetime(2019,9,1),dt.datetime(2019,10,1)))].mean(dim='time')
    v_data_temp[index] = (~np.isnan(a_data_EPA)).loc[dict(lat=latval,lon=lonval)].mean(dim='time')
    index += 1
plt.scatter(x=v_lon_plot,y=v_lat_plot,c=v_data_temp,vmin=0,vmax=1)
plt.colorbar().set_label('Monthly Data Coverage')
# plt.xlabel('lon [deg E]')
# plt.ylabel('lat [deg N]')
plt.title(SUBSET_REGION + ' EPA Sites')
figurename = SUBSET_REGION + ' Coverage' + '.png'
plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
plt.show()

# Import GEOS-CF Dataset
INPUT_FOLDER_GEOS = MAIN_PATH + '\\GEOS-CF DATA'
INPUT_FILE_GEOS = 'GEOS-CF_NO2_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '.csv'
TYPE_GEOS = 'NO2'
UNITS_GEOS = 'mol mol-1'
LABEL_GEOS = 'GEOS-CF NO2'

f_data_GEOS = pd.read_csv(os.path.join(INPUT_FOLDER_GEOS,INPUT_FILE_GEOS))
v_year_GEOS = np.array(f_data_GEOS['Year'])
v_month_GEOS = np.array(f_data_GEOS['Month'])
v_day_GEOS = np.array(f_data_GEOS['Day'])
v_hour_GEOS = np.array(f_data_GEOS['Hour'])
v_minute_GEOS = np.array(f_data_GEOS['Minute'])
v_time_GEOS = [dt.datetime(v_year_GEOS[line],v_month_GEOS[line],v_day_GEOS[line],v_hour_GEOS[line],v_minute_GEOS[line]) for line in range(len(v_year_GEOS))]
v_lat_GEOS = np.array(f_data_GEOS['Lat'])
v_lon_GEOS = np.array(f_data_GEOS['Lon']) 
m_coord_GEOS = np.c_[v_lat_GEOS,v_lon_GEOS,v_time_GEOS]
v_data_GEOS = np.array(f_data_GEOS[TYPE_GEOS]) 
mindex_GEOS = pd.MultiIndex.from_tuples(tuple(m_coord_GEOS),names=['lat','lon','time'])
series_data_GEOS = pd.Series(v_data_GEOS,index = mindex_GEOS)
a_data_GEOS = series_data_GEOS.to_xarray()
a_data_GEOS.lat.attrs['units'] = 'deg N'
a_data_GEOS.lon.attrs['units'] = 'deg E'
a_data_GEOS.attrs['long_name'] = TYPE_GEOS
a_data_GEOS.attrs['units'] = UNITS_GEOS

# Dataset 1 Test Plots
mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
mapobj.coastlines();
mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
# gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines.top_labels = False
gridlines.right_labels = False
my_cmap = plt.cm.get_cmap('Blues')
# a_data_GEOS.loc[dict(time=slice(dt.datetime(2019,9,5),dt.datetime(2019,9,10)))].mean(dim='time').plot(cmap=my_cmap)
# a_data_GEOS.sel(time=dt.datetime(2019,9,1,18,30)).plot()
a_data_GEOS.mean(dim='time').plot(cmap=my_cmap)
plt.show()

# Import GEOS-MET Dataset
INPUT_FOLDER_MET = MAIN_PATH + '\\GEOS-CF MET DATA'
INPUT_FILE_MET = 'GEOS-CF_MET_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '.csv'
v_TYPES_MET = ['T2M','Q2M','V2M','U2M','ZPBL']
v_UNITS_MET = ['K','kg kg-1','m s-1','m s-1','m']
v_LABELS_MET = ['Air Temperature @ 2m','Specific Humidity @ 2m','North Wind Velocity @ 2m','East Wind Velocity @ 2m','Boundary Layer Height']

f_data_MET = pd.read_csv(os.path.join(INPUT_FOLDER_MET,INPUT_FILE_MET))
v_year_MET = np.array(f_data_MET['Year'])
v_month_MET = np.array(f_data_MET['Month'])
v_day_MET = np.array(f_data_MET['Day'])
v_hour_MET = np.array(f_data_MET['Hour'])
v_minute_MET = np.array(f_data_MET['Minute'])
v_time_MET = [dt.datetime(v_year_MET[line],v_month_MET[line],v_day_MET[line],v_hour_MET[line],v_minute_MET[line]) for line in range(len(v_year_MET))]
v_lat_MET = np.array(f_data_MET['Lat'])
v_lon_MET = np.array(f_data_MET['Lon']) 
m_coord_MET = np.c_[v_lat_MET,v_lon_MET,v_time_MET]
mindex_MET = pd.MultiIndex.from_tuples(tuple(m_coord_MET),names=['lat','lon','time'])
for index in range(len(v_TYPES_MET)):
    v_data_MET = np.array(f_data_MET[v_TYPES_MET[index]])
    series_data_MET = pd.Series(v_data_MET,index = mindex_MET)
    a_data_MET = series_data_MET.to_xarray()
    a_data_MET.lat.attrs['units'] = 'deg N'
    a_data_MET.lon.attrs['units'] = 'deg E'
    a_data_MET.attrs['long_name'] = v_LABELS_MET[index]
    a_data_MET.attrs['units'] = v_UNITS_MET[index]
    if (index == 0):
        dataset_MET = xr.Dataset({v_TYPES_MET[index]:a_data_MET})
    else:
        a_data_MET = a_data_MET.interp_like(dataset_MET,method='nearest')
        dataset_MET = xr.merge([dataset_MET,xr.Dataset({v_TYPES_MET[index]:a_data_MET})])

# Dataset Test Plots
for datasetname in list(dataset_MET):
    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
    mapobj.coastlines();
    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
    gridlines.top_labels = False
    gridlines.right_labels = False
    my_cmap = plt.cm.get_cmap('Purples')
    # dataset_MET[datasetname].loc[dict(time=slice(dt.datetime(2019,9,5),dt.datetime(2019,9,10)))].mean(dim='time').plot(cmap=my_cmap)
    # dataset_MET[datasetname].sel(time=dt.datetime(2019,9,1,18,30)).plot(cmap=my_cmap)
    dataset_MET[datasetname].mean(dim='time').plot(cmap=my_cmap)
    # figurename = SUBSET_REGION + ' MET ' + datasetname + '.png'
    # plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show()

# Import TROPOMI Dataset
INPUT_FOLDER_TROPOMI = MAIN_PATH + '\\TROPOMI DATA'
if (SUBSET_REGION in v_CITY_REGIONS):
    INPUT_FILE_TROPOMI = 'TROPOMI_NO2_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '_Register_005_hourly.csv'
else:
    INPUT_FILE_TROPOMI = 'TROPOMI_NO2_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '_Register_010_hourly.csv'

TYPE_TROPOMI = 'NO2'
UNITS_TROPOMI = 'mol m-2'
LABEL_TROPOMI = 'TROPOMI NO2'

f_data_TROPOMI = pd.read_csv(os.path.join(INPUT_FOLDER_TROPOMI,INPUT_FILE_TROPOMI))
v_year_TROPOMI = np.array(f_data_TROPOMI['Year'])
v_month_TROPOMI = np.array(f_data_TROPOMI['Month'])
v_day_TROPOMI = np.array(f_data_TROPOMI['Day'])
v_hour_TROPOMI = np.array(f_data_TROPOMI['Hour'])
v_minute_TROPOMI = np.array(f_data_TROPOMI['Minute'])
v_time_TROPOMI = [dt.datetime(v_year_TROPOMI[line],v_month_TROPOMI[line],v_day_TROPOMI[line],v_hour_TROPOMI[line],v_minute_TROPOMI[line]) for line in range(len(v_year_TROPOMI))]
v_lat_TROPOMI = np.array(f_data_TROPOMI['Lat'])
v_lon_TROPOMI = np.array(f_data_TROPOMI['Lon']) 
m_coord_TROPOMI = np.c_[v_lat_TROPOMI,v_lon_TROPOMI,v_time_TROPOMI]
v_data_TROPOMI = np.array(f_data_TROPOMI[TYPE_TROPOMI]) 
mindex_TROPOMI = pd.MultiIndex.from_tuples(tuple(m_coord_TROPOMI),names=['lat','lon','time'])
series_data_TROPOMI = pd.Series(v_data_TROPOMI,index = mindex_TROPOMI)
a_data_TROPOMI = series_data_TROPOMI.to_xarray()
a_data_TROPOMI.lat.attrs['units'] = 'deg N'
a_data_TROPOMI.lon.attrs['units'] = 'deg E'
a_data_TROPOMI.attrs['long_name'] = TYPE_TROPOMI
a_data_TROPOMI.attrs['units'] = UNITS_TROPOMI

# Dataset 1 Test Plots
mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
mapobj.coastlines();
mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
# gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines.top_labels = False
gridlines.right_labels = False
my_cmap = plt.cm.get_cmap('Reds')
# a_data_TROPOMI.loc[dict(time=slice(dt.datetime(2019,9,5),dt.datetime(2019,9,10)))].mean(dim='time').plot(cmap=my_cmap)
# a_data_TROPOMI.sel(time=dt.datetime(2019,9,1,18,30)).plot(cmap=my_cmap)
a_data_TROPOMI.mean(dim='time').plot(cmap=my_cmap)
plt.show()

# Import Night Lights Dataset
INPUT_FOLDER_VNP46A1 = MAIN_PATH + '\\NIGHT LIGHT DATA'
if (SUBSET_REGION in v_CITY_REGIONS):
    INPUT_FILE_VNP46A1 = 'VNP46A1_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '_Register_0005_Hourly.csv'
    # INPUT_FILE_VNP46A1 = 'VNP46A1_Sept_2019_' + SUBSET_REGION + '_Register_001_Hourly.csv'
else:
    INPUT_FILE_VNP46A1 = 'VNP46A1_' + S_TIMEFRAME_1 + '_' + SUBSET_REGION + '_Register_005_Hourly.csv'
    # INPUT_FILE_VNP46A1 = 'VNP46A1_Sept_2019_' + SUBSET_REGION + '_Register_010_Hourly.csv'  
TYPE_VNP46A1 = 'DNB'
UNITS_VNP46A1 = 'nW/(cm2 sr)'
LABEL_VNP46A1 = 'DNB at Sensor Radiance'

f_data_VNP46A1 = pd.read_csv(os.path.join(INPUT_FOLDER_VNP46A1,INPUT_FILE_VNP46A1))
v_year_VNP46A1 = np.array(f_data_VNP46A1['Year'])
v_month_VNP46A1 = np.array(f_data_VNP46A1['Month'])
v_day_VNP46A1 = np.array(f_data_VNP46A1['Day'])
v_hour_VNP46A1 = np.array(f_data_VNP46A1['Hour'])
v_hour_VNP46A1[v_hour_VNP46A1 > 23] = 0
v_minute_VNP46A1 = np.array(f_data_VNP46A1['Minute'])
v_time_VNP46A1 = [dt.datetime(v_year_VNP46A1[line],v_month_VNP46A1[line],v_day_VNP46A1[line],v_hour_VNP46A1[line],v_minute_VNP46A1[line]) for line in range(len(v_year_VNP46A1))]
v_lat_VNP46A1 = np.array(f_data_VNP46A1['Lat'])
v_lon_VNP46A1 = np.array(f_data_VNP46A1['Lon']) 
m_coord_VNP46A1 = np.c_[v_lat_VNP46A1,v_lon_VNP46A1,v_time_VNP46A1]
v_data_VNP46A1 = np.array(f_data_VNP46A1[TYPE_VNP46A1]) 
mindex_VNP46A1 = pd.MultiIndex.from_tuples(tuple(m_coord_VNP46A1),names=['lat','lon','time'])
# series_data_VNP46A1 = pd.Series(v_data_VNP46A1,index = mindex_VNP46A1)
# # mindex_VNP46A1_unique = mindex_VNP46A1.unique()
# # unique_indexes = [mindex_VNP46A1.index(x) for x in mindex_VNP46A1_unique]
series_data_VNP46A1_orig = pd.Series(v_data_VNP46A1,index = mindex_VNP46A1)
# # series_data_VNP46A1 = series_data_VNP46A1_orig.reindex(mindex_VNP46A1_unique)
series_data_VNP46A1 = series_data_VNP46A1_orig.groupby(level=[0,1,2]).mean()

a_data_VNP46A1 = series_data_VNP46A1.to_xarray()
a_data_VNP46A1.lat.attrs['units'] = 'deg N'
a_data_VNP46A1.lon.attrs['units'] = 'deg E'
a_data_VNP46A1.attrs['long_name'] = TYPE_VNP46A1
a_data_VNP46A1.attrs['units'] = UNITS_VNP46A1

# Dataset 1 Test Plots
mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
mapobj.coastlines();
mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
# gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
gridlines.top_labels = False
gridlines.right_labels = False
my_cmap = plt.cm.get_cmap('Blues')
# a_data_VNP46A1.loc[dict(time=slice(dt.datetime(2019,9,5),dt.datetime(2019,9,10)))].mean(dim='time').plot(cmap=my_cmap)
# a_data_VNP46A1.sel(time=dt.datetime(2019,9,1,18,30)).plot(cmap=my_cmap)
a_data_VNP46A1.mean(dim='time').plot(cmap=my_cmap)
plt.show()

# Import Land Use Dataset

if (SUBSET_REGION == 'CONUS'):
    INPUT_FOLDER_LUR_list = [MAIN_PATH + '\\ROADWAY DATA',
                      MAIN_PATH + '\\POPULATION DATA']
    INPUT_FILE_LUR_list = ['Roads_' + SUBSET_REGION + '_Grid_025.csv',
                          'Population_Density_' + SUBSET_REGION + '_Regrid_025.csv']
    TYPE_LUR_list = [['Class_1','Class_2','Class_3'],
                    ['Population_Density_025']]
    UNITS_LUR_list = [['km km-2','km km-2','km km-2'],
                    ['km-2']]
    LABEL_LUR_list = [['Highway Density','Primary Road Density','Secondary Road Density'],
                    ['Population Density']]
else:
    INPUT_FOLDER_LUR_list = [MAIN_PATH + '\\ROADWAY DATA',
                      MAIN_PATH + '\\POPULATION DATA']
    INPUT_FILE_LUR_list = ['Roads_' + SUBSET_REGION + '_Grid_001.csv',
                          'Population_Density_' + SUBSET_REGION + '.csv']
    TYPE_LUR_list = [['Class_1','Class_2','Class_3'],
                    ['Population_Density']]
    UNITS_LUR_list = [['km km-2','km km-2','km km-2'],
                    ['km-2']]
    LABEL_LUR_list = [['Highway Density','Primary Road Density','Secondary Road Density'],
                    ['Population Density']]


for index in range(len(INPUT_FOLDER_LUR_list)):
    INPUT_FOLDER_LUR = INPUT_FOLDER_LUR_list[index]
    INPUT_FILE_LUR = INPUT_FILE_LUR_list[index]
    TYPE_LUR = TYPE_LUR_list[index]
    UNITS_LUR = UNITS_LUR_list[index]
    LABEL_LUR = LABEL_LUR_list[index]

    # Inport Data
    f_data_LUR = pd.read_csv(os.path.join(INPUT_FOLDER_LUR,INPUT_FILE_LUR))
    v_lat_LUR = np.array(f_data_LUR['Lat'])
    v_lon_LUR = np.array(f_data_LUR['Lon']) 
    m_coord_LUR = np.c_[v_lat_LUR,v_lon_LUR]
    mindex_LUR = pd.MultiIndex.from_tuples(tuple(m_coord_LUR),names=['lat','lon'])
    
    for index_2 in range(len(TYPE_LUR)):
        v_data_LUR = np.array(f_data_LUR[TYPE_LUR[index_2]])
        series_data_LUR = pd.Series(v_data_LUR,index = mindex_LUR)
        a_data_LUR = series_data_LUR.to_xarray()
        a_data_LUR.lat.attrs['units'] = 'deg N'
        a_data_LUR.lon.attrs['units'] = 'deg E'
        a_data_LUR.attrs['long_name'] = LABEL_LUR[index_2]
        a_data_LUR.attrs['units'] = UNITS_LUR[index_2]
        if (index == 0 and index_2 == 0):
            dataset_LUR = xr.Dataset({TYPE_LUR[index_2]:a_data_LUR})
        else:
            a_data_LUR = a_data_LUR.interp_like(dataset_LUR,method='nearest')
            dataset_LUR = xr.merge([dataset_LUR,xr.Dataset({TYPE_LUR[index_2]:a_data_LUR})])
        
# Manual deleting of land use information over water (based on population density)
# dataset_LUR['Population_Density'].to_series()[dataset_LUR['Population_Density'].to_series() == 0] = np.nan

# dataset_LUR['Class_1'].to_series()[np.isnan(dataset_LUR['Population_Density'].to_series())] = np.nan
# dataset_LUR['Class_2'].to_series()[np.isnan(dataset_LUR['Population_Density'].to_series())] = np.nan
# dataset_LUR['Class_3'].to_series()[np.isnan(dataset_LUR['Population_Density'].to_series())] = np.nan

# Dataset 2 Test Plots
for datasetname in list(dataset_LUR):
    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
    mapobj.coastlines();
    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
    gridlines.top_labels = False
    gridlines.right_labels = False
    my_cmap = plt.cm.get_cmap('Greens')
    dataset_LUR[datasetname].plot(cmap=my_cmap)
    # figurename = SUBSET_REGION + ' Land Use ' + datasetname + '.png'
    # plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show()
    
# dataset_LUR.to_dataframe()

#%% Baseline: Persistance and Interpolation for Multiple Timesteps

# Setup:
a_data_ground_sorce = a_data_EPA

n_time_start = np.min(a_data_ground_sorce.time).values

n_time_start_2 = n_time_start + np.timedelta64(n_days_calibration,'D')
n_time_end = np.max(a_data_ground_sorce.time).values

a_data_true = a_data_EPA.loc[dict(time=slice(n_time_start_2,n_time_end))]

v_timescale_target = a_data_true.time.values

a_data_pred = a_data_true.copy()
a_data_pred[:] = np.nan
a_data_pred = a_data_pred.expand_dims({'leadtime':v_prediction_lead_times_hrs})

l_series_single,l_series_else,m_coords_plot = Leave_One_Out_By_Location(series_data_EPA)

for site_index in range(len(l_series_single)):

    a_data_ground_sorce = l_series_else[site_index].to_xarray()
    a_data_ground_target = l_series_single[site_index].to_xarray()
    
    for timestep in v_timescale_target:
        timestep_start = timestep - np.timedelta64(n_days_calibration,'D')
        timestep_end = timestep 
        
        a_data_ground_calib_timestep = a_data_ground_sorce.loc[dict(time=slice(timestep_start,timestep_end))]
        
        m_intermediate_1 = a_data_ground_calib_timestep.ffill('time').interp({'time':timestep_end},method='nearest').drop('time')
        
        m_persistence = xr.DataArray(coords=[('lat',a_data_ground_target.lat),('lon',a_data_ground_target.lon)])
        
        m_intermediate_2 = m_intermediate_1.to_series().dropna()
        if (len(m_intermediate_2) > 0):
            m_coords_temp_2 = m_intermediate_2.index
            m_coords_temp = [[valat,vallon] for (valat,vallon) in m_coords_temp_2]
            
            v_data_temp = m_intermediate_2.values
            m_coords_target_temp_2 = m_persistence.to_series().index
            m_coords_target_temp = [[valat,vallon] for (valat,vallon) in m_coords_target_temp_2]
            
            v_data_target_temp = interpolate.griddata(m_coords_temp,v_data_temp,m_coords_target_temp,method='nearest')
            
            m_persistence = pd.Series(v_data_target_temp,index = m_coords_target_temp_2).to_xarray()
            
            # m_persistence = a_data_ground_calib_timestep.where(~np.isnan(a_data_ground_calib_timestep)).interp({'lat':a_data_ground_target.lat,'lon':a_data_ground_target.lon,'time':timestep_end},method='nearest').drop('time')
            
            # m_persistence = a_data_ground_calib_timestep.interp({'lat':a_data_pred.lat,'lon':a_data_pred.lon,'time':timestep},method='nearest')
            
            for n_lead_time_hrs in v_prediction_lead_times_hrs:
                timestep_2 = timestep + np.timedelta64(n_lead_time_hrs,'h')
                if timestep_2 <= n_time_end:
                    m_temp = m_persistence.expand_dims({'leadtime':[n_lead_time_hrs],'time':[timestep_2]})
                    a_temp = m_temp.combine_first(a_data_pred)
                    a_data_pred = a_temp
    
        Progressbar((site_index*len(v_timescale_target))+np.where(v_timescale_target == timestep)[0][0],len(l_series_single)*len(v_timescale_target),50)

# fix size of a_data_pred
a_data_pred = a_data_pred.interp({'time':a_data_true.time},method='nearest')

# Assess Performance  
m_site_coords = Get_Locations_with_Data(a_data_true)
n_sites = m_site_coords.shape[0]
m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)

v_results_r = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_MAE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_RMSE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_bias = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_MAPE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_RMSPE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_Pbias = np.empty([n_sites,len(v_prediction_lead_times_hrs)])

v_results_r[:] = np.nan
v_results_MAE[:] = np.nan
v_results_MAPE[:] = np.nan
v_results_RMSE[:] = np.nan
v_results_RMSPE[:] = np.nan
v_results_bias[:] = np.nan
v_results_Pbias[:] = np.nan



for lead_time in v_prediction_lead_times_hrs:
    id_lead_time = np.where(np.array(v_prediction_lead_times_hrs) == lead_time)[0][0]
    m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
    for site_index in range(n_sites):
        v_y = m_data_true[:,site_index]
        v_y_hat = m_data_pred[:,site_index]
        
        if SHOW_PLOT:
            plt.plot(v_y,color='red')
            plt.plot(v_y_hat,color='blue')
            plt.show()
       
        v_error = v_y - v_y_hat
        v_Perror = np.abs(v_error/v_y)
                
        id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
    
        if (sum(id_num) > 24):
            v_results_r[site_index,id_lead_time] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
            v_results_MAE[site_index,id_lead_time] = np.nanmean(np.abs(v_error))
            v_results_RMSE[site_index,id_lead_time] = np.sqrt(np.nanmean(v_error**2))
            v_results_bias[site_index,id_lead_time] = np.nanmean(v_error)
            v_results_MAPE[site_index,id_lead_time] = np.nanmean(np.abs(v_Perror))
            v_results_RMSPE[site_index,id_lead_time] = np.sqrt(np.nanmean(v_Perror**2))
            v_results_Pbias[site_index,id_lead_time] = np.nanmean(v_error)/np.nanmean(v_y)
           
    if (SUBSET_REGION == 'CONUS'):
        print(lead_time)
        print(np.nanmedian(v_results_r[:,id_lead_time]),np.nanmedian(v_results_MAPE[:,id_lead_time]),np.nanmedian(v_results_RMSPE[:,id_lead_time]),np.nanmedian(np.abs(v_results_Pbias[:,id_lead_time])))
    else:
        print(lead_time)
        print(np.nanmedian(v_results_r[:,id_lead_time]),np.nanmedian(v_results_MAE[:,id_lead_time]),np.nanmedian(v_results_RMSE[:,id_lead_time]),np.nanmedian(np.abs(v_results_bias[:,id_lead_time])))

        
        
        
if UN_NORMED:      
    # Plot Results - Numerical
    fig,axs = plt.subplots(1,4,gridspec_kw={'wspace': 1})
    fig.suptitle(SUBSET_REGION+' Persistance Model')
    ax=fig.add_subplot(111,frameon=False)
    ax.spines['top'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.spines['left'].set_color('none')
    ax.spines['right'].set_color('none')
    ax.tick_params(labelcolor='none',top=False,bottom=False,left=False,right=False)
    ax.set_xlabel('Lead Time [hrs]')
    axs[0].boxplot(v_results_r[~np.isnan(v_results_r).any(axis=1)])
    axs[0].set(ylabel='r',ylim=[-1,1])
    axs[0].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[1].boxplot(v_results_MAE[~np.isnan(v_results_MAE).any(axis=1)])
    axs[1].set(ylabel = 'MAE [' + UNITS_EPA + ']',ylim = [0,10])
    axs[1].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[2].boxplot(v_results_RMSE[~np.isnan(v_results_MAE).any(axis=1)])
    axs[2].set(ylabel = 'RMSE [' + UNITS_EPA + ']',ylim = [0,20])
    axs[2].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[3].boxplot(np.abs(v_results_bias[~np.isnan(v_results_bias).any(axis=1)]))
    axs[3].set(ylabel = 'absolute bias [' + UNITS_EPA + ']',ylim = [0,10])
    axs[3].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    # plt.xlabel('Lead Time [hrs]')
    plt.rc('xtick',labelsize=8)
    plt.rc('ytick',labelsize=8)
    figurename = SUBSET_REGION + ' Persistance Model' + '.png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show() 
    

if NORMALIZE:
    # Plot Results - Numerical
    fig,axs = plt.subplots(1,4,gridspec_kw={'wspace': 1})
    fig.suptitle(SUBSET_REGION+' Persistance Model')
    ax=fig.add_subplot(111,frameon=False)
    ax.spines['top'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.spines['left'].set_color('none')
    ax.spines['right'].set_color('none')
    ax.tick_params(labelcolor='none',top=False,bottom=False,left=False,right=False)
    ax.set_xlabel('Lead Time [hrs]')
    axs[0].boxplot(v_results_r[~np.isnan(v_results_r).any(axis=1)])
    axs[0].set(ylabel='r',ylim=[-1,1])
    axs[0].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[1].boxplot(v_results_MAPE[~np.isnan(v_results_MAPE).any(axis=1)])
    axs[1].set(ylabel = 'CvMAE',ylim = [0,2])
    axs[1].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[2].boxplot(v_results_RMSPE[~np.isnan(v_results_MAE).any(axis=1)])
    axs[2].set(ylabel = 'nRMSE',ylim = [0,5])
    axs[2].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[3].boxplot(np.abs(v_results_bias[~np.isnan(v_results_bias).any(axis=1)]))
    axs[3].set(ylabel = 'absolute normalized bias',ylim = [0,5])
    axs[3].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    # plt.xlabel('Lead Time [hrs]')
    plt.rc('xtick',labelsize=8)
    plt.rc('ytick',labelsize=8)
    figurename = SUBSET_REGION + ' Persistance Model' + ' (norm).png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show() 

if RESIDUALS:  
    # Analysis of Residuals   
    i_lead = 0
            
    a_data_residual = a_data_pred.sel(leadtime=v_prediction_lead_times_hrs[i_lead]) - a_data_true
    m_data_residual = Array_to_Matrix_at_Coords(a_data_residual,m_site_coords)
    
    v_distances = np.empty([n_sites*n_sites])
    v_spatial_correlations = np.empty([n_sites*n_sites])
    
    v_distances[:] = np.nan
    v_spatial_correlations[:] = np.nan
    
    # Spatial Correlations
    i_line = 0
    for i_site_1 in range(n_sites):
        v_1 = m_data_residual[:,i_site_1]
        for i_site_2 in range(n_sites):
            v_2 = m_data_residual[:,i_site_2]
            
            id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
            
            if (sum(id_num) > 24):
                v_distances[i_line] = Distance_Haversine(m_site_coords[i_site_1,0],m_site_coords[i_site_1,1],m_site_coords[i_site_2,0],m_site_coords[i_site_2,1])
                v_spatial_correlations[i_line] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                
            i_line += 1
            
    # Plot Results - Spatial Correlations
    plt.scatter(x=v_distances,y=v_spatial_correlations)
    plt.ylim(-1,1)
    plt.xlim(0,200)
    plt.xlabel('Distance [km]')
    plt.ylabel('Residual Correlation')
    figurename = SUBSET_REGION + ' Persistance Model' + ' (corr-space).png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show()
    
    # Temporal Correlations
    n_differences = 24*2
    v_differences = np.empty([n_sites,n_differences])
    v_temporal_correlations = np.empty([n_sites,n_differences])
    
    v_differences[:] = np.nan
    v_temporal_correlations[:] = np.nan
    
    for i_site in range(n_sites):
        for i_difference in range(n_differences):
            if i_difference == 0:
                v_1 = m_data_residual[:,i_site]
                v_2 = m_data_residual[:,i_site]
            else:
                v_1 = m_data_residual[:-i_difference,i_site]
                v_2 = m_data_residual[i_difference:,i_site]
            
            id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
            if (sum(id_num) > 24):
                v_differences[i_site,i_difference] = i_difference
                v_temporal_correlations[i_site,i_difference] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
            
    # Plot Results - Temporal Correlations
    plt.boxplot(v_temporal_correlations[~np.isnan(v_temporal_correlations).any(axis=1)])
    plt.xticks(range(0,n_differences+1),labels=range(0,n_differences+1),rotation=90)
    plt.ylim(-1,1)
    plt.xlim(0,n_differences)
    plt.xlabel('Time Difference [hrs]')
    plt.ylabel('Residual Correlation')
    figurename = SUBSET_REGION + ' Persistance Model' + ' (corr-time).png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show()
    
    # Land Use Values and Correlations
    for Land_Use_Type in list(dataset_LUR):
        a_LUV_at_sites = dataset_LUR[Land_Use_Type].interp_like(a_data_residual)
        v_LUV = Array_to_Matrix_at_Coords(a_LUV_at_sites,m_site_coords)

        v_differences = np.empty([n_sites*n_sites])
        v_luv_correlations = np.empty([n_sites*n_sites])
        
        v_differences[:] = np.nan
        v_luv_correlations[:] = np.nan
        
        # Spatial Correlations
        i_line = 0
        for i_site_1 in range(n_sites):
            v_1 = m_data_residual[:,i_site_1]
            for i_site_2 in range(n_sites):
                v_2 = m_data_residual[:,i_site_2]
                
                id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
                
                if (sum(id_num) > 24):
                    v_differences[i_line] = np.abs(v_LUV[:,i_site_1] - v_LUV[:,i_site_2])
                    v_luv_correlations[i_line] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                    
                i_line += 1
                
        # Plot Results - Land Use Value Correlations
        plt.scatter(x=v_differences,y=v_luv_correlations)
        plt.ylim(-1,1)
        plt.xlim(0,np.max(v_differences))
        plt.xlabel('Difference in ' + Land_Use_Type)
        plt.ylabel('Residual Correlation')
        figurename = SUBSET_REGION + ' Persistance Model' + ' (corr-luv-' + Land_Use_Type + ').png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
        # Plot Results - Land Use Value v Residuals
        plt.boxplot(m_data_residual[~np.isnan(m_data_residual).any(axis=1)],positions=v_LUV[0],widths=0.05*np.max(v_LUV[0]))
        plt.xlabel(Land_Use_Type)
        plt.ylabel('Residuals')
        plt.xticks(range(0,int(np.max(v_LUV[0])),int(0.1*np.max(v_LUV[0]))),labels=range(0,int(np.max(v_LUV[0])),int(0.1*np.max(v_LUV[0]))),rotation=90)
        figurename = SUBSET_REGION + ' Persistance Model' + ' (residual-v-' + Land_Use_Type + ').png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
# Save Data
filename = SUBSET_REGION + ' Persistance Model (truth)' + '.nc'
a_data_true.to_netcdf(os.path.join(OUTPUT_DATA_FOLDER,filename))
filename = SUBSET_REGION + ' Persistance Model (prediction)' + '.nc'
a_data_pred.to_netcdf(os.path.join(OUTPUT_DATA_FOLDER,filename))

#%% Baseline: Diurnal Pattern and Interpolation for Multiple Timesteps

# Setup:
a_data_ground_sorce = a_data_EPA


n_time_start = np.min(a_data_ground_sorce.time).values

n_time_start_2 = n_time_start + np.timedelta64(n_days_calibration,'D')
n_time_end = np.max(a_data_ground_sorce.time).values

a_data_true = a_data_EPA.loc[dict(time=slice(n_time_start_2,n_time_end))]

v_timescale_target = a_data_true.time.values

a_data_pred = a_data_true.copy()
a_data_pred[:] = np.nan

a_data_pred = a_data_pred.expand_dims({'leadtime':v_prediction_lead_times_hrs})

l_series_single,l_series_else,m_coords_plot = Leave_One_Out_By_Location(series_data_EPA)

for site_index in range(len(l_series_single)):
    a_data_ground_sorce = l_series_else[site_index].to_xarray()
    a_data_ground_target = l_series_single[site_index].to_xarray()
    for timestep in v_timescale_target:
        timestep_start = timestep - np.timedelta64(n_days_calibration,'D')
        timestep_end = timestep 
        
        a_data_ground_calib_timestep = a_data_ground_sorce.loc[dict(time=slice(timestep_start,timestep_end))]
        
        a_diurnal_temp = a_data_ground_calib_timestep.groupby('time.hour').mean(dim='time')

        a_diurnal = xr.DataArray(coords=[('lat',a_data_ground_target.lat),('lon',a_data_ground_target.lon),('hour',a_diurnal_temp.hour)])
        
        m_intermediate_2 = a_diurnal_temp.to_series().dropna()
        m_coords_temp_2 = m_intermediate_2.index
        m_coords_temp = [[valat,vallon,valhour] for (valat,vallon,valhour) in m_coords_temp_2]
        
        v_data_temp = m_intermediate_2.values
        m_coords_target_temp_2 = a_diurnal.to_series().index
        m_coords_target_temp = [[valat,vallon,valhour] for (valat,vallon,valhour) in m_coords_target_temp_2]
        
        v_data_target_temp = interpolate.griddata(m_coords_temp,v_data_temp,m_coords_target_temp,method='nearest')
        
        a_diurnal = pd.Series(v_data_target_temp,index = m_coords_target_temp_2).to_xarray()
        
       
        for n_lead_time_hrs in v_prediction_lead_times_hrs:
            timestep_2 = timestep + np.timedelta64(n_lead_time_hrs,'h')
            if timestep_2 <= n_time_end:
                n_hour = pd.Timestamp(timestep_2).hour
                if n_hour in a_diurnal.hour:
                    m_temp = a_diurnal.sel(hour=n_hour).drop('hour').expand_dims({'leadtime':[n_lead_time_hrs],'time':[timestep_2]})
                    a_temp = m_temp.combine_first(a_data_pred)
                    a_data_pred = a_temp
    
        Progressbar((site_index*len(v_timescale_target))+np.where(v_timescale_target == timestep)[0][0],len(l_series_single)*len(v_timescale_target),50)

# fix size of a_data_pred
a_data_pred = a_data_pred.interp({'time':a_data_true.time},method='nearest')

# Assess Performance  
m_site_coords = Get_Locations_with_Data(a_data_true)
n_sites = m_site_coords.shape[0]
m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)

v_results_r = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_MAE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_RMSE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_bias = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_MAPE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_RMSPE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
v_results_Pbias = np.empty([n_sites,len(v_prediction_lead_times_hrs)])

v_results_r[:] = np.nan
v_results_MAE[:] = np.nan
v_results_MAPE[:] = np.nan
v_results_RMSE[:] = np.nan
v_results_RMSPE[:] = np.nan
v_results_bias[:] = np.nan
v_results_Pbias[:] = np.nan

for lead_time in v_prediction_lead_times_hrs:
    id_lead_time = np.where(np.array(v_prediction_lead_times_hrs) == lead_time)[0][0]
    m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
    for site_index in range(n_sites):
        v_y = m_data_true[:,site_index]
        v_y_hat = m_data_pred[:,site_index]
        
        if SHOW_PLOT:
            plt.plot(v_y,color='red')
            plt.plot(v_y_hat,color='blue')
            plt.show()
       
        v_error = v_y - v_y_hat
        v_Perror = np.abs(v_error/v_y)
                
        id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
    
        if (sum(id_num) > 24):
            v_results_r[site_index,id_lead_time] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
            v_results_MAE[site_index,id_lead_time] = np.nanmean(np.abs(v_error))
            v_results_RMSE[site_index,id_lead_time] = np.sqrt(np.nanmean(v_error**2))
            v_results_bias[site_index,id_lead_time] = np.nanmean(v_error)
            v_results_MAPE[site_index,id_lead_time] = np.nanmean(np.abs(v_Perror))
            v_results_RMSPE[site_index,id_lead_time] = np.sqrt(np.nanmean(v_Perror**2))
            v_results_Pbias[site_index,id_lead_time] = np.nanmean(v_error)/np.nanmean(v_y)
           
    if (SUBSET_REGION == 'CONUS'):
        print(lead_time)
        print(np.nanmedian(v_results_r[:,id_lead_time]),np.nanmedian(v_results_MAPE[:,id_lead_time]),np.nanmedian(v_results_RMSPE[:,id_lead_time]),np.nanmedian(np.abs(v_results_Pbias[:,id_lead_time])))
    else:
        print(lead_time)
        print(np.nanmedian(v_results_r[:,id_lead_time]),np.nanmedian(v_results_MAE[:,id_lead_time]),np.nanmedian(v_results_RMSE[:,id_lead_time]),np.nanmedian(np.abs(v_results_bias[:,id_lead_time])))

        
if UN_NORMED:      
    # Plot Results - Numerical
    fig,axs = plt.subplots(1,4,gridspec_kw={'wspace': 1})
    fig.suptitle(SUBSET_REGION+' Diurnal Model')
    ax=fig.add_subplot(111,frameon=False)
    ax.spines['top'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.spines['left'].set_color('none')
    ax.spines['right'].set_color('none')
    ax.tick_params(labelcolor='none',top=False,bottom=False,left=False,right=False)
    ax.set_xlabel('Lead Time [hrs]')
    axs[0].boxplot(v_results_r[~np.isnan(v_results_r).any(axis=1)])
    axs[0].set(ylabel='r',ylim=[-1,1])
    axs[0].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[1].boxplot(v_results_MAE[~np.isnan(v_results_MAE).any(axis=1)])
    axs[1].set(ylabel = 'MAE [' + UNITS_EPA + ']',ylim = [0,10])
    axs[1].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[2].boxplot(v_results_RMSE[~np.isnan(v_results_MAE).any(axis=1)])
    axs[2].set(ylabel = 'RMSE [' + UNITS_EPA + ']',ylim = [0,20])
    axs[2].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[3].boxplot(np.abs(v_results_bias[~np.isnan(v_results_bias).any(axis=1)]))
    axs[3].set(ylabel = 'absolute bias [' + UNITS_EPA + ']',ylim = [0,10])
    axs[3].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    # plt.xlabel('Lead Time [hrs]')
    plt.rc('xtick',labelsize=8)
    plt.rc('ytick',labelsize=8) 
    figurename = SUBSET_REGION + ' Diurnal Model' + '.png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show() 

if NORMALIZE:
    # Plot Results - Numerical
    fig,axs = plt.subplots(1,4,gridspec_kw={'wspace': 1})
    fig.suptitle(SUBSET_REGION+' Diurnal Model')
    ax=fig.add_subplot(111,frameon=False)
    ax.spines['top'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.spines['left'].set_color('none')
    ax.spines['right'].set_color('none')
    ax.tick_params(labelcolor='none',top=False,bottom=False,left=False,right=False)
    ax.set_xlabel('Lead Time [hrs]')
    axs[0].boxplot(v_results_r[~np.isnan(v_results_r).any(axis=1)])
    axs[0].set(ylabel='r',ylim=[-1,1])
    axs[0].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[1].boxplot(v_results_MAPE[~np.isnan(v_results_MAPE).any(axis=1)])
    axs[1].set(ylabel = 'CvMAE',ylim = [0,2])
    axs[1].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[2].boxplot(v_results_RMSPE[~np.isnan(v_results_MAE).any(axis=1)])
    axs[2].set(ylabel = 'nRMSE',ylim = [0,5])
    axs[2].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    axs[3].boxplot(np.abs(v_results_bias[~np.isnan(v_results_bias).any(axis=1)]))
    axs[3].set(ylabel = 'absolute normalized bias',ylim = [0,5])
    axs[3].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
    # plt.xlabel('Lead Time [hrs]')
    plt.rc('xtick',labelsize=8)
    plt.rc('ytick',labelsize=8) 
    figurename = SUBSET_REGION + ' Diurnal Model' + ' (norm).png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show() 

if RESIDUALS:  
    # Analysis of Residuals   
    i_lead = 0
            
    a_data_residual = a_data_pred.sel(leadtime=v_prediction_lead_times_hrs[i_lead]) - a_data_true
    m_data_residual = Array_to_Matrix_at_Coords(a_data_residual,m_site_coords)
    
    v_distances = np.empty([n_sites*n_sites])
    v_spatial_correlations = np.empty([n_sites*n_sites])
    
    v_distances[:] = np.nan
    v_spatial_correlations[:] = np.nan
    
    # Spatial Correlations
    i_line = 0
    for i_site_1 in range(n_sites):
        v_1 = m_data_residual[:,i_site_1]
        for i_site_2 in range(n_sites):
            v_2 = m_data_residual[:,i_site_2]
            
            id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
            
            if (sum(id_num) > 24):
                v_distances[i_line] = Distance_Haversine(m_site_coords[i_site_1,0],m_site_coords[i_site_1,1],m_site_coords[i_site_2,0],m_site_coords[i_site_2,1])
                v_spatial_correlations[i_line] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                
            i_line += 1
            
    # Plot Results - Spatial Correlations
    plt.scatter(x=v_distances,y=v_spatial_correlations)
    plt.ylim(-1,1)
    plt.xlim(0,200)
    plt.xlabel('Distance [km]')
    plt.ylabel('Residual Correlation')
    figurename = SUBSET_REGION + ' Diurnal Model' + ' (corr-space).png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show()
    
    # Temporal Correlations
    n_differences = 24*2
    v_differences = np.empty([n_sites,n_differences])
    v_temporal_correlations = np.empty([n_sites,n_differences])
    
    v_differences[:] = np.nan
    v_temporal_correlations[:] = np.nan
    
    for i_site in range(n_sites):
        for i_difference in range(n_differences):
            if i_difference == 0:
                v_1 = m_data_residual[:,i_site]
                v_2 = m_data_residual[:,i_site]
            else:
                v_1 = m_data_residual[:-i_difference,i_site]
                v_2 = m_data_residual[i_difference:,i_site]
            
            id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
            if (sum(id_num) > 24):
                v_differences[i_site,i_difference] = i_difference
                v_temporal_correlations[i_site,i_difference] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
            
    # Plot Results - Temporal Correlations
    plt.boxplot(v_temporal_correlations[~np.isnan(v_temporal_correlations).any(axis=1)])
    plt.xticks(range(0,n_differences+1),labels=range(0,n_differences+1),rotation=90)
    plt.ylim(-1,1)
    plt.xlim(0,n_differences)
    plt.xlabel('Time Difference [hrs]')
    plt.ylabel('Residual Correlation')
    figurename = SUBSET_REGION + ' Diurnal Model' + ' (corr-time).png'
    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
    plt.show()
    
    # Land Use Values and Correlations
    for Land_Use_Type in list(dataset_LUR):
        a_LUV_at_sites = dataset_LUR[Land_Use_Type].interp_like(a_data_residual)
        v_LUV = Array_to_Matrix_at_Coords(a_LUV_at_sites,m_site_coords)

        v_differences = np.empty([n_sites*n_sites])
        v_luv_correlations = np.empty([n_sites*n_sites])
        
        v_differences[:] = np.nan
        v_luv_correlations[:] = np.nan
        
        # Spatial Correlations
        i_line = 0
        for i_site_1 in range(n_sites):
            v_1 = m_data_residual[:,i_site_1]
            for i_site_2 in range(n_sites):
                v_2 = m_data_residual[:,i_site_2]
                
                id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
                
                if (sum(id_num) > 24):
                    v_differences[i_line] = np.abs(v_LUV[:,i_site_1] - v_LUV[:,i_site_2])
                    v_luv_correlations[i_line] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                    
                i_line += 1
                
        # Plot Results - Land Use Value Correlations
        plt.scatter(x=v_differences,y=v_luv_correlations)
        plt.ylim(-1,1)
        plt.xlim(0,np.max(v_differences))
        plt.xlabel('Difference in ' + Land_Use_Type)
        plt.ylabel('Residual Correlation')
        figurename = SUBSET_REGION + ' Diurnal Model' + ' (corr-luv-' + Land_Use_Type + ').png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
        # Plot Results - Land Use Value v Residuals
        plt.boxplot(m_data_residual[~np.isnan(m_data_residual).any(axis=1)],positions=v_LUV[0],widths=0.05*np.max(v_LUV[0]))
        plt.xlabel(Land_Use_Type)
        plt.ylabel('Residuals')
        plt.xticks(range(0,int(np.max(v_LUV[0])),int(0.1*np.max(v_LUV[0]))),labels=range(0,int(np.max(v_LUV[0])),int(0.1*np.max(v_LUV[0]))),rotation=90)
        figurename = SUBSET_REGION + ' Diurnal Model' + ' (residual-v-' + Land_Use_Type + ').png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
# Save Data
filename = SUBSET_REGION + ' Diurnal Model (truth)' + '.nc'
a_data_true.to_netcdf(os.path.join(OUTPUT_DATA_FOLDER,filename))
filename = SUBSET_REGION + ' Diurnal Model (prediction)' + '.nc'
a_data_pred.to_netcdf(os.path.join(OUTPUT_DATA_FOLDER,filename))

#%% Model 1: Multiple ahead-of-time predictions
     
#           USE_AUX USE_FINE USE_COARSE PATTERN_METHOD PATTERN_USAGE REGRESSION_APPROACH      UPRES_METHOD
# SETTINGS = [\
#             [False, False,   False,     'alltime',     'addon',      'alltime',               'linear'],\
#             [False, False,   False,     'alltime',     'addon',      'time-of-day_2',         'linear'],\
#             [False, False,   False,     'alltime',     'addon',      'weighted-diurnal',      'linear'],\
#             [False, False,   False,     'alltime',     'addon',      'weighted-time',         'linear'],\
#             [False, False,   False,     'alltime',     'addon',      'weighted-time+diurnal', 'linear'],\
#             [False, False,   False,     'alltime',     'addon',      'alltime',               'nearest'],\
#             [False, False,   False,     'alltime',     'addon',      'time-of-day_2',         'nearest'],\
#             [False, False,   False,     'alltime',     'addon',      'weighted-diurnal',      'nearest'],\
#             [False, False,   False,     'alltime',     'addon',      'weighted-time',         'nearest'],\
#             [False, False,   False,     'alltime',     'addon',      'weighted-time+diurnal', 'nearest'],\
#             ]
   

# SETTINGS = [\
            # [False, False, False, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
            # [False, False, True, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
            # [False, False, False, 'sametime',  'addon', 'time-of-day_2', 'linear'],\
            # [False, False, True, 'sametime',  'addon', 'time-of-day_2', 'linear'],\
#             [False, False, False, 'sametime',  'regress', 'weighted-time+diurnal', 'linear'],\
#             [False, False, True, 'sametime',  'regress', 'weighted-time+diurnal', 'linear'],\
#             [False, True, True, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
#             [False, True, False, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
#             [True, False, True, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
#            ]
    
    
# SETTINGS = [\
#             [False, False,   True,      'sametime',    'regress',    'alltime',               'nearest'],\
#             [False, False,   True,      'sametime',    'regress',    'time-of-day_2',         'nearest'],\
#             [False, False,   True,      'sametime',    'regress',    'weighted-diurnal',      'nearest'],\
#             [False, False,   True,      'sametime',    'regress',    'weighted-time',         'nearest'],\
#             [False, False,   True,      'sametime',    'regress',    'weighted-time+diurnal', 'nearest'],\
#             [False, False,   True,      'sametime',    'addon',      'alltime',               'nearest'],\
#             [False, False,   True,      'sametime',    'addon',      'time-of-day_2',         'nearest'],\
#             [False, False,   True,      'sametime',    'addon',      'weighted-diurnal',      'nearest'],\
#             [False, False,   True,      'sametime',    'addon',      'weighted-time',         'nearest'],\
#             [False, False,   True,      'sametime',    'addon',      'weighted-time+diurnal', 'nearest'],\
#             [False, False,   True,      'alltime',     'regress',    'alltime',               'nearest'],\
#             [False, False,   True,      'alltime',     'regress',    'time-of-day_2',         'nearest'],\
#             [False, False,   True,      'alltime',     'regress',    'weighted-diurnal',      'nearest'],\
#             [False, False,   True,      'alltime',     'regress',    'weighted-time',         'nearest'],\
#             [False, False,   True,      'alltime',     'regress',    'weighted-time+diurnal', 'nearest'],\
#             [False, False,   True,      'alltime',     'addon',      'alltime',               'nearest'],\
#             [False, False,   True,      'alltime',     'addon',      'time-of-day_2',         'nearest'],\
#             [False, False,   True,      'alltime',     'addon',      'weighted-diurnal',      'nearest'],\
#             [False, False,   True,      'alltime',     'addon',      'weighted-time',         'nearest'],\
#             [False, False,   True,      'alltime',     'addon',      'weighted-time+diurnal', 'nearest'],\
#             [False, False,   True,      'sametime',    'regress',    'alltime',               'linear'],\
#             [False, False,   True,      'sametime',    'regress',    'time-of-day_2',         'linear'],\
#             [False, False,   True,      'sametime',    'regress',    'weighted-diurnal',      'linear'],\
#             [False, False,   True,      'sametime',    'regress',    'weighted-time',         'linear'],\
#             [False, False,   True,      'sametime',    'regress',    'weighted-time+diurnal', 'linear'],\
#             [False, False,   True,      'sametime',    'addon',      'alltime',               'linear'],\
#             [False, False,   True,      'sametime',    'addon',      'time-of-day_2',         'linear'],\
#             [False, False,   True,      'sametime',    'addon',      'weighted-diurnal',      'linear'],\
#             [False, False,   True,      'sametime',    'addon',      'weighted-time',         'linear'],\
#             [False, False,   True,      'sametime',    'addon',      'weighted-time+diurnal', 'linear'],\
#             [False, False,   True,      'alltime',     'regress',    'alltime',               'linear'],\
#             [False, False,   True,      'alltime',     'regress',    'time-of-day_2',         'linear'],\
#             [False, False,   True,      'alltime',     'regress',    'weighted-diurnal',      'linear'],\
#             [False, False,   True,      'alltime',     'regress',    'weighted-time',         'linear'],\
#             [False, False,   True,      'alltime',     'regress',    'weighted-time+diurnal', 'linear'],\
#             [False, False,   True,      'alltime',     'addon',      'alltime',               'linear'],\
#             [False, False,   True,      'alltime',     'addon',      'time-of-day_2',         'linear'],\
#             [False, False,   True,      'alltime',     'addon',      'weighted-diurnal',      'linear'],\
#             [False, False,   True,      'alltime',     'addon',      'weighted-time',         'linear'],\
#             [False, False,   True,      'alltime',     'addon',      'weighted-time+diurnal', 'linear'],\
#             ]

# SETTINGS = [\
#             [False, False, False, 'alltime',  'addon', 'alltime', 'linear'],\
#             [False, False, False, 'alltime',  'addon', 'weighted-time+diurnal', 'linear'],\
#             [False, False, True,  'alltime',  'addon', 'weighted-time+diurnal', 'linear'],\
#             [False, False, True,  'alltime',  'addon', 'weighted-time+diurnal', 'nearest'],\
#             [False, False, True,  'sametime', 'addon', 'weighted-time+diurnal', 'linear'],\
#             [False, False, True,  'alltime',  'regress', 'weighted-time+diurnal', 'linear'],\
#             [False, False, True,  'sametime',  'regress', 'weighted-time+diurnal', 'linear'],\
#             [False, False, True,  'alltime',  'addon', 'alltime', 'linear'],\
#             [False, False, True,  'alltime',  'addon', 'time-of-day_2', 'linear'],\
#             [False, False, True,  'alltime',  'addon', 'weighted-diurnal', 'linear'],\
#             [False, False, True,  'alltime',  'addon', 'weighted-time', 'linear'],\
#             [True, False, True, 'alltime', 'addon', 'weighted-time+diurnal', 'linear'],\
#             [False, True, True, 'alltime', 'addon', 'weighted-time+diurnal', 'linear'],\
#             ]

if SUBSET_REGION == 'LasVegas':
    SETTINGS = [\
                [False, False, False, 'alltime',  'addon', 'time-of-day_2', 'linear'],\
                [False, False, True, 'alltime',  'addon', 'time-of-day_2', 'linear'],\
                ]
elif (SUBSET_REGION == 'SaltLakeCity'):
    SETTINGS = [\
                [False, False, False, 'sametime',  'addon', 'time-of-day_2', 'linear'],\
                [False, False, True, 'sametime',  'addon', 'time-of-day_2', 'linear'],\
                ]
else:
    SETTINGS = [\
                [False, False, False, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
                [False, False, True, 'sametime',  'addon', 'weighted-time+diurnal', 'linear'],\
                ]

# Settings:

USE_AUX = False
USE_FINE = True
USE_COARSE = True


PATTERN_METHOD = 'alltime'
# Options:
# 1. 'alltime'
# 2. 'sametime'
# 3. 'sametime+mask'

PATTERN_USAGE = 'addon'
# Options:
# A. 'addon'
# R. 'regress'

REGRESSION_APPROACH = 'alltime' # options: 
# Options:
# 1. 'alltime'
# 2. 'last-time'
# 3. 'time-of-day'
# 4. 'time-of-day_2'
# 5. 'weighted-diurnal'
# 6. 'weighted-time'
# 7. 'weighted-time+diurnal'

UPRES_METHOD = 'nearest'
# Options:
# N. 'nearest'
# L. 'linear'


for iset in range(len(SETTINGS)):
    USE_AUX = SETTINGS[iset][0]
    USE_FINE = SETTINGS[iset][1]
    USE_COARSE = SETTINGS[iset][2]
    PATTERN_METHOD = SETTINGS[iset][3]
    PATTERN_USAGE = SETTINGS[iset][4]
    REGRESSION_APPROACH = SETTINGS[iset][5]
    UPRES_METHOD = SETTINGS[iset][6]
    
    PATTERN_METHOD_DICT = {'alltime':'1','sametime':'2','sametime+mask':'3'}
    PATTERN_USAGE_DICT = {'addon':'A','regress':'R'}
    REGRESSION_APPROACH_DICT = {'alltime':'1','last-time':'2','time-of-day':'3','time-of-day_2':'4','weighted-diurnal':'5','weighted-time':'6','weighted-time+diurnal':'7'}
    UPRES_METHOD_DICT = {'nearest':'N','linear':'L'}
    
    # Setup:
    a_data_ground_sorce = a_data_EPA
    a_data_pred_source = a_data_GEOS*1e9
    a_data_aux_source = a_data_MET
    a_data_pattern_source_coarse = a_data_TROPOMI
    a_data_pattern_source_fine = a_data_VNP46A1
    
    n_time_start = np.min(a_data_EPA.time).values
    
    n_time_start_2 = n_time_start + np.timedelta64(n_days_calibration,'D')
    n_time_end = np.max(a_data_EPA.time).values
    
    a_data_true = a_data_EPA.loc[dict(time=slice(n_time_start_2,n_time_end))]
    
    v_timescale_target = a_data_true.time.values
    
    a_data_pred = a_data_true.copy()
    a_data_pred[:] = np.nan
    
    a_data_pred = a_data_pred.expand_dims({'leadtime':v_prediction_lead_times_hrs})
    
    l_series_single,l_series_else,m_coords_plot = Leave_One_Out_By_Location(series_data_EPA)
    
    
    for timestep in v_timescale_target:
        timestep_start = timestep - np.timedelta64(n_days_calibration,'D')
        timestep_end = timestep
        
        a_data_pred_calib_timestep = a_data_pred_source.loc[dict(time=slice(timestep_start,timestep_end))]
        a_data_aux_calib_timestep = a_data_aux_source.loc[dict(time=slice(timestep_start,timestep_end))]
        a_data_pattern_coarse_calib_timestep = a_data_pattern_source_coarse.loc[dict(time=slice(timestep_start,timestep_end))]
        a_data_pattern_fine_calib_timestep = a_data_pattern_source_fine.loc[dict(time=slice(timestep_start,timestep_end))]
        
        v_timegrid_pred_calib = a_data_pred_calib_timestep.time
        
        
        if USE_COARSE:
            # Extract coarse Pattern
            # # Get average pattern for calibration period, fill NANs
            m_data_pattern_course_temp = a_data_pattern_coarse_calib_timestep.mean(dim='time').interpolate_na('lat').interpolate_na('lon')
            
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('Reds')
                m_data_pattern_course_temp.plot(cmap=my_cmap,vmin=0,vmax=25)
                plt.show()
            
            # # Rescale to prediction pattern
            if (PATTERN_METHOD == 'alltime'):
                # over entire time period:
                m_data_pattern_pred_temp = a_data_pred_calib_timestep.mean(dim='time').interp_like(m_data_pattern_course_temp,method=UPRES_METHOD)
            elif (PATTERN_METHOD == 'sametime'):
                # only times when coarse pattern has data (satellite passes):
                m_data_pattern_pred_temp = a_data_pred_calib_timestep.interp_like(a_data_pattern_coarse_calib_timestep,method=UPRES_METHOD).mean(dim='time')
            elif (PATTERN_METHOD == 'sametime+mask'):
                # only locations where coarse pattern has data
                a_data_pattern_pred_temp = a_data_pred_calib_timestep.interp_like(a_data_pattern_coarse_calib_timestep,method=UPRES_METHOD).copy()
                m_data_pattern_pred_temp = a_data_pattern_pred_temp.where(~np.isnan(a_data_pattern_coarse_calib_timestep)).mean(dim='time')
                
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('Reds')
                m_data_pattern_pred_temp.plot(cmap=my_cmap,vmin=0,vmax=25)
                plt.show()
            
            v_1 = m_data_pattern_pred_temp.values.reshape(-1)
            v_2 = m_data_pattern_course_temp.values.reshape(-1)
            id_num = (~np.isnan(np.c_[v_1,v_2])).all(axis=1)
            regr_temp = linear_model.LinearRegression()
            regr_temp.fit(v_2[id_num].reshape(-1,1),v_1[id_num].reshape(-1,1)) 
            n_slope = regr_temp.coef_[0][0]
            n_intercept = regr_temp.intercept_[0]
            m_data_pattern_course_temp_rescale = (n_slope*m_data_pattern_course_temp) + n_intercept
        
            if (SHOW_PLOT):
                x_temp = m_data_pattern_pred_temp.values.reshape(-1)
                y_temp = m_data_pattern_course_temp.values.reshape(-1)
                id_num = ~np.isnan(np.c_[x_temp,y_temp]).any(axis=1)
                plt.scatter(x=x_temp[id_num],y=y_temp[id_num])
                y1 = np.min(y_temp[id_num])
                y2 = np.max(y_temp[id_num])
                x1 = (n_slope*y1) + n_intercept
                x2 = (n_slope*y2) + n_intercept
                plt.plot([x1,x2],[y1,y2],'r-')
                plt.ylim(np.min(y_temp[id_num]),np.max(y_temp[id_num]))
                plt.xlim(np.min(x_temp[id_num]),np.max(x_temp[id_num]))
                plt.xlabel('Model Values')
                plt.ylabel('Coarse Pattern')
                plt.title(SUBSET_REGION + ' Coarse Pattern Rescaling\nx = '+ str(n_slope) + ' y + ' + str(n_intercept) + '\nr = ' + str(np.corrcoef(y_temp[id_num],x_temp[id_num])[0,1]))
                figurename = SUBSET_REGION + ' Coarse Pattern Rescaling (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                plt.show()
                
            # # Reconstruct pattern
            m_difference_coarse = m_data_pattern_course_temp_rescale - m_data_pattern_pred_temp
                
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('bwr')
                m_difference_coarse.plot(cmap=my_cmap,vmin=-15,vmax=15)
                plt.xlabel('')
                plt.ylabel('')
                plt.title(SUBSET_REGION + ' Coarse Pattern')
                figurename = SUBSET_REGION + ' Coarse Pattern (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                plt.show()
            
            a_data_pred_calib_timestep_upres = a_data_pred_calib_timestep.interp({'lat':m_difference_coarse.lat,'lon':m_difference_coarse.lon,'time':v_timegrid_pred_calib},method=UPRES_METHOD)
            
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('Blues')
                a_data_pred_calib_timestep_upres.mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                plt.xlabel('')
                plt.ylabel('')
                plt.title(SUBSET_REGION + ' Average Model Field')
                figurename = SUBSET_REGION + ' Avg Model (' + UPRES_METHOD + ').png'
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                plt.show()
            
            
            if (PATTERN_USAGE == 'addon'):
                a_pred_coarse_calib = a_data_pred_calib_timestep_upres + m_difference_coarse.expand_dims({'time':v_timegrid_pred_calib})
                    
                if (SHOW_PLOT):
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('Reds')
                    a_pred_coarse_calib.mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                    plt.xlabel('')
                    plt.ylabel('')
                    plt.title(SUBSET_REGION + ' Average Coarse Prediction Field')
                    figurename = SUBSET_REGION + ' Avg Coarse Pred (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                    plt.show()
                    
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('bwr')
                    (a_pred_coarse_calib - a_data_pred_calib_timestep_upres).mean(dim='time').plot(cmap=my_cmap,vmin=-15,vmax=15)
                    plt.title('Change to Prediction for Calibration Period')
                    plt.show()
        
                    
            elif (PATTERN_USAGE == 'regress'):
                a_pred_coarse_calib = xr.Dataset({'Pred':a_data_pred_calib_timestep_upres,'Coarse':m_difference_coarse.expand_dims({'time':v_timegrid_pred_calib})})
                
                if (SHOW_PLOT):
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('Reds')
                    a_pred_coarse_calib['Pred'].mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                    plt.title('Average Coarse Prediction for Calibration Period')
                    plt.show()
                    
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('bwr')
                    a_pred_coarse_calib['Coarse'].mean(dim='time').plot(cmap=my_cmap,vmin=-15,vmax=15)
                    plt.title('Change to Prediction for Calibration Period')
                    plt.show()
        else:
            if (PATTERN_USAGE == 'addon'):
                a_pred_coarse_calib = a_data_pred_calib_timestep
            elif (PATTERN_USAGE == 'regress'):
                a_pred_coarse_calib = xr.Dataset({'Pred':a_data_pred_calib_timestep})
            
            m_difference_coarse = a_data_pred_calib_timestep.mean(dim='time').copy()
            m_difference_coarse = m_difference_coarse - m_difference_coarse
    
    
        if (USE_FINE):
            # Extract fine Pattern
            # # Get average pattern for calibration period, fill NANs
            m_data_pattern_fine_temp = a_data_pattern_fine_calib_timestep.mean(dim='time').interpolate_na('lat').interpolate_na('lon')
        
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('Reds')
                m_data_pattern_fine_temp.plot(cmap=my_cmap,vmin=0,vmax=25)
                plt.show()
            
            # # Rescale to prediction pattern
            if (PATTERN_METHOD == 'alltime'):
                # over entire time period:
                if (type(a_pred_coarse_calib) == xr.core.dataset.Dataset):
                    m_data_pattern_pred_temp = a_pred_coarse_calib['Pred'].mean(dim='time').interp_like(m_data_pattern_fine_temp,method=UPRES_METHOD)
                else:
                    m_data_pattern_pred_temp = a_pred_coarse_calib.mean(dim='time').interp_like(m_data_pattern_fine_temp,method=UPRES_METHOD)
        
            elif (PATTERN_METHOD == 'sametime'):
                # only times when coarse pattern has data (satellite passes):
                if (type(a_pred_coarse_calib) == xr.core.dataset.Dataset):
                    m_data_pattern_pred_temp = a_pred_coarse_calib['Pred'].interp_like(a_data_pattern_fine_calib_timestep,method=UPRES_METHOD).mean(dim='time')
                else:
                    m_data_pattern_pred_temp = a_pred_coarse_calib.interp_like(a_data_pattern_fine_calib_timestep,method=UPRES_METHOD).mean(dim='time')
          
            elif (PATTERN_METHOD == 'sametime+mask'):
                # only locations where coarse pattern has data
                if (type(a_pred_coarse_calib) == xr.core.dataset.Dataset):
                    a_data_pattern_pred_temp = a_pred_coarse_calib['Pred'].interp_like(a_data_pattern_fine_calib_timestep,method=UPRES_METHOD).copy()
                    m_data_pattern_pred_temp = a_data_pattern_pred_temp.where(~np.isnan(a_data_pattern_fine_calib_timestep)).mean(dim='time')
                else:
                    a_data_pattern_pred_temp = a_pred_coarse_calib.interp_like(a_data_pattern_fine_calib_timestep,method=UPRES_METHOD).copy()
                    m_data_pattern_pred_temp = a_data_pattern_pred_temp.where(~np.isnan(a_data_pattern_fine_calib_timestep)).mean(dim='time')
                
                
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('Reds')
                m_data_pattern_pred_temp.plot(cmap=my_cmap,vmin=0,vmax=25)
                plt.show()
            
            v_1 = m_data_pattern_pred_temp.values.reshape(-1)
            v_2 = m_data_pattern_fine_temp.values.reshape(-1)
            id_num = (~np.isnan(np.c_[v_1,v_2])).all(axis=1)
            regr_temp = linear_model.LinearRegression()
            regr_temp.fit(v_2[id_num].reshape(-1,1),v_1[id_num].reshape(-1,1)) 
            n_slope = regr_temp.coef_[0][0]
            n_intercept = regr_temp.intercept_[0]
            m_data_pattern_fine_temp_rescale = (n_slope*m_data_pattern_fine_temp) + n_intercept
        
            if (SHOW_PLOT):
                x_temp = m_data_pattern_pred_temp.values.reshape(-1)
                y_temp = m_data_pattern_fine_temp_rescale.values.reshape(-1)
                id_num = ~np.isnan(np.c_[x_temp,y_temp]).any(axis=1)
                plt.scatter(x=x_temp[id_num],y=y_temp[id_num])
                y1 = np.min(y_temp[id_num])
                y2 = np.max(y_temp[id_num])
                x1 = (n_slope*y1) + n_intercept
                x2 = (n_slope*y2) + n_intercept
                plt.plot([x1,x2],[y1,y2],'r-')
                plt.ylim(np.min(y_temp[id_num]),np.max(y_temp[id_num]))
                plt.xlim(np.min(x_temp[id_num]),np.max(x_temp[id_num]))
                plt.xlabel('Model Values')
                plt.ylabel('Fine Pattern')
                plt.title(SUBSET_REGION + ' Fine Pattern Rescaling\nx = '+ str(n_slope) + ' y + ' + str(n_intercept) + '\nr = ' + str(np.corrcoef(y_temp[id_num],x_temp[id_num])[0,1]))
                figurename = SUBSET_REGION + ' Fine Pattern Rescaling (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                plt.show()
                
            # # Reconstruct pattern
            m_difference_fine = m_data_pattern_fine_temp_rescale - m_data_pattern_pred_temp
                
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('bwr')
                m_difference_fine.plot(cmap=my_cmap,vmin=-15,vmax=15)
                plt.xlabel('')
                plt.ylabel('')
                plt.title(SUBSET_REGION + ' Fine Pattern')
                figurename = SUBSET_REGION + ' Fine Pattern (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                plt.show()
            
            a_data_coarse_calib_timestep_upres = a_pred_coarse_calib.interp({'lat':m_difference_fine.lat,'lon':m_difference_fine.lon,'time':v_timegrid_pred_calib},method=UPRES_METHOD)
            
            if (SHOW_PLOT):
                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                mapobj.coastlines();
                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                gridlines.top_labels = False
                gridlines.right_labels = False
                my_cmap = plt.cm.get_cmap('Blues')
                if (PATTERN_USAGE == 'addon'):
                    a_data_coarse_calib_timestep_upres['Pred'].mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                elif (PATTERN_USAGE == 'regress'):
                    a_data_coarse_calib_timestep_upres['Pred'].mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                plt.xlabel('')
                plt.ylabel('')
                plt.clabel('')
                plt.title(SUBSET_REGION + ' Average Model Field')
                figurename = SUBSET_REGION + ' Avg Model (' + UPRES_METHOD + ').png'
                plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                plt.show()
                
            if (PATTERN_USAGE == 'addon'):
                a_pred_fine_calib = a_data_coarse_calib_timestep_upres + m_difference_fine.expand_dims({'time':v_timegrid_pred_calib})
                    
                if (SHOW_PLOT):
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('Reds')
                    a_pred_fine_calib.mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                    plt.xlabel('')
                    plt.ylabel('')
                    plt.title(SUBSET_REGION + ' Average Fine Prediction Field')
                    figurename = SUBSET_REGION + ' Avg Fine Pred (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                    plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                    plt.show()
                    
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('bwr')
                    (a_pred_fine_calib - a_data_coarse_calib_timestep_upres).mean(dim='time').plot(cmap=my_cmap,vmin=-15,vmax=15)
                    plt.title('Change to Prediction for Calibration Period')
                    plt.show()
    
                    
            elif (PATTERN_USAGE == 'regress'):
                if USE_COARSE:
                    a_pred_fine_calib = xr.Dataset({'Pred':a_data_coarse_calib_timestep_upres['Pred'],'Coarse':a_data_coarse_calib_timestep_upres['Coarse'],'Fine':m_difference_fine.expand_dims({'time':v_timegrid_pred_calib})})
                else:
                    a_pred_fine_calib = xr.Dataset({'Pred':a_data_coarse_calib_timestep_upres['Pred'],'Fine':m_difference_fine.expand_dims({'time':v_timegrid_pred_calib})})
                
                if (SHOW_PLOT):
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('Reds')
                    a_pred_fine_calib['Pred'].mean(dim='time').plot(cmap=my_cmap,vmin=0,vmax=25)
                    plt.title('Average Fine Prediction for Calibration Period')
                    plt.show()
                    
                    mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                    mapobj.coastlines();
                    mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                    mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                    # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                    gridlines.top_labels = False
                    gridlines.right_labels = False
                    my_cmap = plt.cm.get_cmap('bwr')
                    a_pred_fine_calib['Fine'].mean(dim='time').plot(cmap=my_cmap,vmin=-15,vmax=15)
                    plt.title('Change to Prediction for Calibration Period')
                    plt.show()
             
        # Choose prediction pattern to use
        if (USE_FINE):
            a_data_pred_calib = a_pred_fine_calib
        else:
           a_data_pred_calib = a_pred_coarse_calib
           
        if not(type(a_data_pred_calib) == xr.core.dataset.Dataset):
             a_data_pred_calib = xr.Dataset({'Pred':a_data_pred_calib})
           
        # Add Aux. Variables if necessary
        if (USE_AUX):
            a_aux_interp_calib = a_data_aux_calib_timestep.interp_like(a_data_pred_calib,method=UPRES_METHOD)
                    
            if (type(a_data_pred_calib) == xr.core.dataset.Dataset):
                if (type(a_aux_interp_calib) == xr.core.dataset.Dataset):
                    a_data_pred_calib = xr.merge([a_data_pred_calib,a_aux_interp_calib])
                else:
                    a_data_pred_calib = xr.merge([a_data_pred_calib,xr.Dataset({'AUX':a_aux_interp_calib})])
            else:
                if (type(a_aux_interp_calib) == xr.core.dataset.Dataset):
                    a_data_pred_calib = xr.merge([xr.Dataset({'Pred':a_data_pred_calib}),a_aux_interp_calib])
                else:
                    a_data_pred_calib = xr.merge([xr.Dataset({'Pred':a_data_pred_calib}),xr.Dataset({'AUX':a_aux_interp_calib})])
            
        # Cycle Through Ground Sites
        for site_index in range(len(l_series_single)):
            a_data_ground_sorce = l_series_else[site_index].to_xarray()
            a_data_ground_target = l_series_single[site_index].to_xarray()
            
            a_data_ground_calib_timestep = a_data_ground_sorce.loc[dict(time=slice(timestep_start,timestep_end))]
    
            # Cycle through prediction lead times    
            for n_lead_time_hrs in v_prediction_lead_times_hrs:
                timestep_2 = timestep + np.timedelta64(n_lead_time_hrs,'h')
                
                if timestep_2 <= n_time_end:
                    # m_data_ground_target_timestep = a_data_ground_sorce.interp(time=timestep_2)
                    m_data_ground_target_timestep = a_data_ground_target.interp(time=timestep_2)
                    m_data_pred_target_timestep = a_data_pred_source.interp(time=timestep_2)
                    m_data_aux_target_timestep = a_data_aux_source.interp(time=timestep_2)
                
                    m_data_pred_target_timestep_upres = m_data_pred_target_timestep.interp({'lat':m_difference_coarse.lat,'lon':m_difference_coarse.lon},method=UPRES_METHOD)
                
                    if (PATTERN_USAGE == 'addon'):
                        m_pred_coarse_target = m_data_pred_target_timestep_upres + m_difference_coarse
                            
                        if (SHOW_PLOT):
                            mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                            mapobj.coastlines();
                            mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                            mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                            # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines.top_labels = False
                            gridlines.right_labels = False
                            my_cmap = plt.cm.get_cmap('Reds')
                            m_pred_coarse_target.plot(cmap=my_cmap,vmin=0,vmax=25)
                            plt.title('Coarse Prediction at Target Time')
                            plt.show()
                            
                            mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                            mapobj.coastlines();
                            mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                            mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                            # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines.top_labels = False
                            gridlines.right_labels = False
                            my_cmap = plt.cm.get_cmap('bwr')
                            (m_pred_coarse_target - m_data_pred_target_timestep_upres).plot(cmap=my_cmap,vmin=-15,vmax=15)
                            plt.title('Change to Prediction at Target Time')
                            plt.show()
                            
                    elif (PATTERN_USAGE == 'regress'):
                        if USE_COARSE:
                            m_pred_coarse_target = xr.Dataset({'Pred':m_data_pred_target_timestep_upres,'Coarse':m_difference_coarse})
                        else:
                            m_pred_coarse_target = xr.Dataset({'Pred':m_data_pred_target_timestep_upres})
                        
                        if (SHOW_PLOT):
                            mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                            mapobj.coastlines();
                            mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                            mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                            # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines.top_labels = False
                            gridlines.right_labels = False
                            my_cmap = plt.cm.get_cmap('Reds')
                            m_pred_coarse_target['Pred'].plot(cmap=my_cmap,vmin=0,vmax=25)
                            plt.title('Coarse Prediction at Target Time')
                            plt.show()
                            
                            mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                            mapobj.coastlines();
                            mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                            mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                            # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                            gridlines.top_labels = False
                            gridlines.right_labels = False
                            my_cmap = plt.cm.get_cmap('bwr')
                            m_pred_coarse_target['Coarse'].plot(cmap=my_cmap,vmin=-15,vmax=15)
                            plt.title('Change to Prediction at Target Time')
                            plt.show()
                    
                    if (USE_FINE):
                        m_data_coarse_target_timestep_upres = m_pred_coarse_target.interp({'lat':m_difference_fine.lat,'lon':m_difference_fine.lon},method=UPRES_METHOD)
                        
                        if (PATTERN_USAGE == 'addon'):
                            m_pred_fine_target = m_data_coarse_target_timestep_upres + m_difference_fine
                                
                            if (SHOW_PLOT):
                                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                                mapobj.coastlines();
                                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines.top_labels = False
                                gridlines.right_labels = False
                                my_cmap = plt.cm.get_cmap('Reds')
                                m_pred_fine_target.plot(cmap=my_cmap,vmin=0,vmax=25)
                                plt.title('Fine Prediction at Target Time')
                                plt.show()
                                
                                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                                mapobj.coastlines();
                                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines.top_labels = False
                                gridlines.right_labels = False
                                my_cmap = plt.cm.get_cmap('bwr')
                                (m_pred_fine_target - m_data_coarse_target_timestep_upres).plot(cmap=my_cmap,vmin=-15,vmax=15)
                                plt.title('Change to Prediction at Target Time')
                                plt.show()
                                
                        elif (PATTERN_USAGE == 'regress'):
                            if USE_COARSE:
                                m_pred_fine_target = xr.Dataset({'Pred':m_data_coarse_target_timestep_upres['Pred'],'Coarse':m_data_coarse_target_timestep_upres['Coarse'],'Fine':m_difference_fine})
                            else:
                                m_pred_fine_target = xr.Dataset({'Pred':m_data_coarse_target_timestep_upres['Pred'],'Fine':m_difference_fine})
                            
                            if (SHOW_PLOT):
                                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                                mapobj.coastlines();
                                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines.top_labels = False
                                gridlines.right_labels = False
                                my_cmap = plt.cm.get_cmap('Reds')
                                m_pred_fine_target['Pred'].plot(cmap=my_cmap,vmin=0,vmax=25)
                                plt.title('Fine Prediction at Target Time')
                                plt.show()
                                
                                mapobj = plt.figure().add_subplot(1,1,1,projection=cartopy.crs.PlateCarree());
                                mapobj.coastlines();
                                mapobj.add_feature(cartopy.feature.LAKES, edgecolor='black', facecolor='none', linestyle='-');
                                mapobj.set_extent([np.max([n_min_lon,-180])-0.001, np.min([n_max_lon,180])+0.001, np.max([n_min_lat,-90])-0.001, np.min([n_max_lat,90])+0.001])
                                # gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=[-113,-112,-111], ylocs=[40,41,42], draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines = mapobj.gridlines(crs=cartopy.crs.PlateCarree(), xlocs=range(np.int_(np.floor(np.max([n_min_lon,-180]))),np.int_(np.ceil(np.min([n_max_lon,180])))), ylocs=range(np.int_(np.floor(np.max([n_min_lat,-90]))),np.int_(np.ceil(np.min([n_max_lat,90])))), draw_labels=True, linewidth=1, color='black', alpha=0.5, linestyle='--')
                                gridlines.top_labels = False
                                gridlines.right_labels = False
                                my_cmap = plt.cm.get_cmap('bwr')
                                m_pred_fine_target['Fine'].plot(cmap=my_cmap,vmin=-15,vmax=15)
                                plt.title('Change to Prediction at Target Time')
                                plt.show()
                    
                    # Choose prediction pattern to use
                    if (USE_FINE):
                        m_data_pred_target = m_pred_fine_target
                    else:
                       m_data_pred_target = m_pred_coarse_target
                       
                    if not(type(m_data_pred_target) == xr.core.dataset.Dataset):
                         m_data_pred_target = xr.Dataset({'Pred':m_data_pred_target})
                       
                    # Add Aux. Variables if necessary
                    if (USE_AUX):
                        m_aux_interp_target = m_data_aux_target_timestep.interp_like(m_data_pred_target,method=UPRES_METHOD)
                                
                        if (type(a_data_pred_calib) == xr.core.dataset.Dataset):
                            if (type(a_aux_interp_calib) == xr.core.dataset.Dataset):
                                m_data_pred_target = xr.merge([m_data_pred_target,m_aux_interp_target])
                            else:
                                m_data_pred_target = xr.merge([m_data_pred_target,xr.Dataset({'AUX':m_aux_interp_target})])
                        else:
                            if (type(a_aux_interp_calib) == xr.core.dataset.Dataset):
                                m_data_pred_target = xr.merge([xr.Dataset({'Pred':m_data_pred_target}),m_aux_interp_target])
                            else:
                                m_data_pred_target = xr.merge([xr.Dataset({'Pred':m_data_pred_target}),xr.Dataset({'AUX':m_aux_interp_target})])
                     
                   # Fit Patterns to Ground Data
                    a_data_pred_calib_at_ground = a_data_pred_calib.interp_like(a_data_ground_calib_timestep,method='nearest')
                    m_data_pred_calib_at_ground = a_data_pred_calib_at_ground.to_dataframe()
                    m_data_ground_calib = xr.Dataset({'Ground':a_data_ground_calib_timestep}).to_dataframe()
                    
                    m_X = np.array(m_data_pred_calib_at_ground)
                    v_y = np.array(m_data_ground_calib)
                    v_deltaT = np.array(m_data_ground_calib.index.get_level_values('time')) - timestep_2
                    
                    if SHOW_PLOT:
                        x_temp = m_data_ground_calib['Ground']
                        y_temp = m_data_pred_calib_at_ground['Pred']
                        # c_temp = m_data_ground_calib.index.get_level_values('time')
                        # c_temp = np.cos(np.pi*v_deltaT/np.timedelta64(1,'D'))**2
                        c_temp = np.exp(-(np.abs(v_deltaT/np.timedelta64(n_time_resolution_hrs,'h'))/24))
                        id_num = ~np.isnan(np.c_[x_temp,y_temp]).any(axis=1)
                        plt.scatter(x=x_temp[id_num],y=y_temp[id_num],c=c_temp[id_num])
                        plt.ylim(np.min(y_temp[id_num]),np.max(y_temp[id_num]))
                        plt.xlim(np.min(x_temp[id_num]),np.max(x_temp[id_num]))
                        plt.colorbar()
                        plt.xlabel('')
                        plt.ylabel('')
                        plt.title('r = ' + str(np.corrcoef(y_temp[id_num],x_temp[id_num])[0,1]))
                        plt.show()
                        
    
                    id_num = ~np.isnan(np.c_[m_X,v_y]).any(axis=1)
                    
                    regr_timestep = linear_model.LinearRegression()
                
                    
                    if (REGRESSION_APPROACH == 'alltime'):
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1)) 
                    elif (REGRESSION_APPROACH == 'last-time'):
                        v_weights = np.ones(len(v_deltaT))
                        v_weights[(np.abs(v_deltaT) > np.timedelta64(n_time_resolution_hrs,'h'))] = 0
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1),sample_weight=v_weights[id_num]) 
                    elif (REGRESSION_APPROACH == 'time-of-day'):
                        v_weights = np.ones(len(v_deltaT))
                        v_weights[np.abs(v_deltaT/np.timedelta64(1,'D')) - np.floor(np.abs(v_deltaT/np.timedelta64(1,'D'))) >= (np.timedelta64(n_time_resolution_hrs,'h')/np.timedelta64(1,'D'))] = 0
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1),sample_weight=v_weights[id_num])  
                    elif (REGRESSION_APPROACH == 'time-of-day_2'):
                        v_weights = np.ones(len(v_deltaT))
                        v_weights[np.abs(v_deltaT/np.timedelta64(1,'D')) - np.floor(np.abs(v_deltaT/np.timedelta64(1,'D'))) > (np.timedelta64(n_time_resolution_hrs,'h')/np.timedelta64(1,'D'))] = 0
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1),sample_weight=v_weights[id_num])  
                    elif (REGRESSION_APPROACH == 'weighted-diurnal'):
                        v_weights = np.cos(np.pi*v_deltaT/np.timedelta64(1,'D'))**2
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1),sample_weight=v_weights[id_num])  
                    elif (REGRESSION_APPROACH == 'weighted-time'):
                        n_timescale_hours = 24
                        v_weights = np.exp(-(np.abs(v_deltaT/np.timedelta64(n_time_resolution_hrs,'h'))/n_timescale_hours))
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1),sample_weight=v_weights[id_num]) 
                    elif (REGRESSION_APPROACH == 'weighted-time+diurnal'):
                        n_weight_cycle = 0.5
                        n_timescale_hours = 24
                        v_weights_time = np.exp(-(np.abs(v_deltaT/np.timedelta64(n_time_resolution_hrs,'h'))/n_timescale_hours))
                        v_weights_cycle = np.cos(np.pi*v_deltaT/np.timedelta64(1,'D'))**2
                        v_weights = (n_weight_cycle*v_weights_cycle) + ((1-n_weight_cycle)*v_weights_time)
                        # plt.plot(v_deltaT,v_weights)
                        regr_timestep.fit(m_X[id_num,:],v_y[id_num].reshape(-1,1),sample_weight=v_weights[id_num])  
                    
                    # Apply fit at current timestamp - Ground Sites
                    m_data_pred_target_at_ground = m_data_pred_target.interp_like(m_data_ground_target_timestep,method='nearest')
                    mindex_data_pred_target_at_ground = m_data_pred_target_at_ground.to_dataframe().drop(columns='time',errors='ignore').index
                    m_X = np.array(m_data_pred_target_at_ground.to_dataframe().drop(columns='time',errors='ignore'))
                  
                    id_num = ~np.isnan(m_X).any(axis=1)
                    
                    if (np.sum(id_num) > 0):
                        
                        v_y_pred = regr_timestep.predict(m_X[id_num,:])   
                    
                        m_pred_target_at_ground = pd.Series(v_y_pred.reshape(-1),index = mindex_data_pred_target_at_ground[id_num]).to_xarray()
                    
                        if SHOW_PLOT:
                            x_temp = m_data_ground_target_timestep.values.reshape(-1)
                            y_temp = v_y_pred.reshape(-1)
                            id_num = ~np.isnan(np.c_[x_temp,y_temp]).any(axis=1)
                            plt.scatter(x=x_temp[id_num],y=y_temp[id_num])
                            plt.ylim(np.min(y_temp[id_num]),np.max(y_temp[id_num]))
                            plt.xlim(np.min(x_temp[id_num]),np.max(x_temp[id_num]))
                            plt.xlabel('Measrement')
                            plt.ylabel('Prediction')
                            plt.title('r = ' + str(np.corrcoef(y_temp[id_num],x_temp[id_num])[0,1]))
                            plt.show()
                            
                        if (SHOW_PLOT):
                            m_data_pred_target_at_ground_PLOT = a_data_pred_calib_at_ground
                            mindex_data_pred_target_at_ground_PLOT = m_data_pred_target_at_ground_PLOT.to_dataframe().index
                            m_X_PLOT = np.array(m_data_pred_target_at_ground_PLOT.to_dataframe())
                          
                            id_num_PLOT = ~np.isnan(m_X_PLOT).any(axis=1)
                            v_y_pred_PLOT = regr_timestep.predict(m_X_PLOT[id_num_PLOT,:])   
                        
                            m_pred_target_at_ground_PLOT = pd.Series(v_y_pred_PLOT.reshape(-1),index = mindex_data_pred_target_at_ground_PLOT[id_num_PLOT])
                         
                            x_temp = m_data_ground_calib['Ground']
                            y_temp = m_pred_target_at_ground_PLOT
                            id_num = ~np.isnan(np.c_[x_temp,y_temp]).any(axis=1)
                            plt.scatter(x=x_temp[id_num],y=y_temp[id_num])
                            y1 = np.min(y_temp[id_num])
                            y2 = np.max(y_temp[id_num])
                            x1 = (n_slope*y1) + n_intercept
                            x2 = (n_slope*y2) + n_intercept
                            plt.plot([x1,x2],[y1,y2],'r-')
                            plt.ylim(np.min(y_temp[id_num]),np.max(y_temp[id_num]))
                            plt.xlim(np.min(x_temp[id_num]),np.max(x_temp[id_num]))
                            plt.xlabel('Ground Data (Calibration Period)')
                            plt.ylabel('Model Prediction (Calibration Period)')
                            plt.title(SUBSET_REGION + ' Regression to Ground Data\nr = ' + str(np.corrcoef(y_temp[id_num],x_temp[id_num])[0,1]))
                            figurename = SUBSET_REGION + ' Ground Data Regression (' + UPRES_METHOD + ', ' + PATTERN_METHOD + ').png'
                            plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
                            plt.show()
                        
                        m_temp = m_pred_target_at_ground.expand_dims({'time':[timestep_2],'leadtime':[n_lead_time_hrs]})
                        a_temp = m_temp.combine_first(a_data_pred)
                        a_data_pred = a_temp
                        
                        
                        # a_data_pred.loc[dict(time=timestep_2,leadtime=n_lead_time_hrs)] = m_pred_target_at_ground
                        # a_data_pred.sel(time=timestep_2,leadtime=n_lead_time_hrs) = m_pred_target_at_ground
                    
            Progressbar((site_index)+(len(l_series_single)*np.where(v_timescale_target == timestep)[0][0]),len(l_series_single)*len(v_timescale_target),50)
    
    # fix size of a_data_pred
    a_data_pred = a_data_pred.interp({'time':a_data_true.time},method='nearest')
    
    # Assess Performance  
    m_site_coords = Get_Locations_with_Data(a_data_true)
    n_sites = m_site_coords.shape[0]
    m_data_true = Array_to_Matrix_at_Coords(a_data_true,m_site_coords)
    
    v_results_r = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    v_results_MAE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    v_results_RMSE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    v_results_bias = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    v_results_MAPE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    v_results_RMSPE = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    v_results_Pbias = np.empty([n_sites,len(v_prediction_lead_times_hrs)])
    
    v_results_r[:] = np.nan
    v_results_MAE[:] = np.nan
    v_results_MAPE[:] = np.nan
    v_results_RMSE[:] = np.nan
    v_results_RMSPE[:] = np.nan
    v_results_bias[:] = np.nan
    v_results_Pbias[:] = np.nan
    
    for lead_time in v_prediction_lead_times_hrs:
        id_lead_time = np.where(np.array(v_prediction_lead_times_hrs) == lead_time)[0][0]
        m_data_pred = Array_to_Matrix_at_Coords(a_data_pred.sel(leadtime=lead_time),m_site_coords)
        for site_index in range(n_sites):
            v_y = m_data_true[:,site_index]
            v_y_hat = m_data_pred[:,site_index]
            
            if SHOW_PLOT:
                plt.plot(v_y,color='red')
                plt.plot(v_y_hat,color='blue')
                plt.show()
           
            v_error = v_y - v_y_hat
            v_Perror = np.abs(v_error/v_y)
                    
            id_num = ~np.isnan(np.c_[v_y,v_y_hat]).any(axis=1)
        
            if (sum(id_num) > 24):
                v_results_r[site_index,id_lead_time] = np.corrcoef(v_y[id_num],v_y_hat[id_num])[0,1]
                v_results_MAE[site_index,id_lead_time] = np.nanmean(np.abs(v_error))
                v_results_RMSE[site_index,id_lead_time] = np.sqrt(np.nanmean(v_error**2))
                v_results_bias[site_index,id_lead_time] = np.nanmean(v_error)
                v_results_MAPE[site_index,id_lead_time] = np.nanmean(np.abs(v_Perror))
                v_results_RMSPE[site_index,id_lead_time] = np.sqrt(np.nanmean(v_Perror**2))
                v_results_Pbias[site_index,id_lead_time] = np.nanmean(v_error)/np.nanmean(v_y)
               
        if (SUBSET_REGION == 'CONUS'):
            print(lead_time)
            print(np.nanmedian(v_results_r[:,id_lead_time]),np.nanmedian(v_results_MAPE[:,id_lead_time]),np.nanmedian(v_results_RMSPE[:,id_lead_time]),np.nanmedian(np.abs(v_results_Pbias[:,id_lead_time])))
        else:
            print(lead_time)
            print(np.nanmedian(v_results_r[:,id_lead_time]),np.nanmedian(v_results_MAE[:,id_lead_time]),np.nanmedian(v_results_RMSE[:,id_lead_time]),np.nanmedian(np.abs(v_results_bias[:,id_lead_time])))
    
    # Save Data
    s_figtitle_a = SUBSET_REGION+' Prediction Model'
    if USE_AUX & USE_FINE & USE_COARSE:
        s_figtitle_b = '(GEOS-CF + TROPOMI + VPN46A1 + MET)'
    elif USE_AUX & USE_FINE:
        s_figtitle_b = '(GEOS-CF + VPN46A1 + MET)'
    elif USE_AUX & USE_COARSE:
        s_figtitle_b = '(GEOS-CF + TROPOMI + MET)'
    elif USE_FINE & USE_COARSE:
        s_figtitle_b = '(GEOS-CF + TROPOMI + VPN46A1)'
    elif USE_AUX:
        s_figtitle_b = '(GEOS-CF + MET)'
    elif USE_FINE:
        s_figtitle_b = '(GEOS-CF + VPN46A1)'
    elif USE_COARSE:
        s_figtitle_b = '(GEOS-CF + TROPOMI)'
    else:
        s_figtitle_b = '(GEOS-CF)'
    s_namebase = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD]
          
    filename = s_namebase + ' (truth)' + '.nc'
    a_data_true.to_netcdf(os.path.join(OUTPUT_DATA_FOLDER,filename))
    filename = s_namebase + ' (prediction)' + '.nc'
    a_data_pred.to_netcdf(os.path.join(OUTPUT_DATA_FOLDER,filename))        
            
            
    if NORMALIZE:
        # Plot Results - Numerical
        fig,axs = plt.subplots(1,4,gridspec_kw={'wspace': 1})
        s_figtitle_a = SUBSET_REGION+' Prediction Model'
        if USE_AUX & USE_FINE & USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI + VPN46A1 + MET)'
        elif USE_AUX & USE_FINE:
            s_figtitle_b = '(GEOS-CF + VPN46A1 + MET)'
        elif USE_AUX & USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI + MET)'
        elif USE_FINE & USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI + VPN46A1)'
        elif USE_AUX:
            s_figtitle_b = '(GEOS-CF + MET)'
        elif USE_FINE:
            s_figtitle_b = '(GEOS-CF + VPN46A1)'
        elif USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI)'
        else:
            s_figtitle_b = '(GEOS-CF)'
        fig.suptitle(s_figtitle_a + '\n' + s_figtitle_b)
        ax=fig.add_subplot(111,frameon=False)
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_color('none')
        ax.spines['left'].set_color('none')
        ax.spines['right'].set_color('none')
        ax.tick_params(labelcolor='none',top=False,bottom=False,left=False,right=False)
        ax.set_xlabel('Lead Time [hrs]')
        axs[0].boxplot(v_results_r[~np.isnan(v_results_r).any(axis=1)])
        axs[0].set(ylabel='r',ylim=[-1,1])
        axs[0].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        axs[1].boxplot(v_results_MAPE[~np.isnan(v_results_MAPE).any(axis=1)])
        axs[1].set(ylabel = 'CvMAE',ylim = [0,2])
        axs[1].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        axs[2].boxplot(v_results_RMSPE[~np.isnan(v_results_RMSPE).any(axis=1)])
        axs[2].set(ylabel = 'nRMSE',ylim = [0,5])
        axs[2].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        axs[3].boxplot(np.abs(v_results_Pbias[~np.isnan(v_results_Pbias).any(axis=1)]))
        axs[3].set(ylabel = 'absolute normalized bias',ylim = [0,5])
        axs[3].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        # plt.xlabel('Lead Time [hrs]')
        plt.rc('xtick',labelsize=8)
        plt.rc('ytick',labelsize=8) 
        figurename = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD] + ' (norm).png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
    if UN_NORMED:
        # Plot Results - Numerical
        fig,axs = plt.subplots(1,4,gridspec_kw={'wspace': 1})
        s_figtitle_a = SUBSET_REGION+' Prediction Model'
        if USE_AUX & USE_FINE & USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI + VPN46A1 + MET)'
        elif USE_AUX & USE_FINE:
            s_figtitle_b = '(GEOS-CF + VPN46A1 + MET)'
        elif USE_AUX & USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI + MET)'
        elif USE_FINE & USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI + VPN46A1)'
        elif USE_AUX:
            s_figtitle_b = '(GEOS-CF + MET)'
        elif USE_FINE:
            s_figtitle_b = '(GEOS-CF + VPN46A1)'
        elif USE_COARSE:
            s_figtitle_b = '(GEOS-CF + TROPOMI)'
        else:
            s_figtitle_b = '(GEOS-CF)'
        fig.suptitle(s_figtitle_a + '\n' + s_figtitle_b)
        ax=fig.add_subplot(111,frameon=False)
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_color('none')
        ax.spines['left'].set_color('none')
        ax.spines['right'].set_color('none')
        ax.tick_params(labelcolor='none',top=False,bottom=False,left=False,right=False)
        ax.set_xlabel('Lead Time [hrs]')
        axs[0].boxplot(v_results_r[~np.isnan(v_results_r).any(axis=1)])
        axs[0].set(ylabel='r',ylim=[-1,1])
        axs[0].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        axs[1].boxplot(v_results_MAE[~np.isnan(v_results_MAE).any(axis=1)])
        axs[1].set(ylabel = 'MAE [' + UNITS_EPA + ']',ylim = [0,10])
        axs[1].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        axs[2].boxplot(v_results_RMSE[~np.isnan(v_results_MAE).any(axis=1)])
        axs[2].set(ylabel = 'RMSE [' + UNITS_EPA + ']',ylim = [0,20])
        axs[2].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        axs[3].boxplot(np.abs(v_results_bias[~np.isnan(v_results_bias).any(axis=1)]))
        axs[3].set(ylabel = 'absolute bias [' + UNITS_EPA + ']',ylim = [0,10])
        axs[3].set_xticklabels(v_prediction_lead_times_hrs,rotation=90)
        # plt.xlabel('Lead Time [hrs]')
        plt.rc('xtick',labelsize=8)
        plt.rc('ytick',labelsize=8) 
        figurename = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD] + '.png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
    if RESIDUALS:  
        # Analysis of Residuals   
        i_lead = 0
                
        a_data_residual = a_data_pred.sel(leadtime=v_prediction_lead_times_hrs[i_lead]) - a_data_true
        m_data_residual = Array_to_Matrix_at_Coords(a_data_residual,m_site_coords)
        
        v_distances = np.empty([n_sites*n_sites])
        v_spatial_correlations = np.empty([n_sites*n_sites])
        
        v_distances[:] = np.nan
        v_spatial_correlations[:] = np.nan
        
        # Spatial Correlations
        i_line = 0
        for i_site_1 in range(n_sites):
            v_1 = m_data_residual[:,i_site_1]
            for i_site_2 in range(n_sites):
                v_2 = m_data_residual[:,i_site_2]
                
                id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
                
                if (sum(id_num) > 24):
                    v_distances[i_line] = Distance_Haversine(m_site_coords[i_site_1,0],m_site_coords[i_site_1,1],m_site_coords[i_site_2,0],m_site_coords[i_site_2,1])
                    v_spatial_correlations[i_line] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                    
                i_line += 1
                
        # Plot Results - Spatial Correlations
        plt.scatter(x=v_distances,y=v_spatial_correlations)
        plt.ylim(-1,1)
        plt.xlim(0,200)
        plt.xlabel('Distance [km]')
        plt.ylabel('Residual Correlation')
        figurename = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD] + ' (corr-space).png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
        # Temporal Correlations
        n_differences = 24*2
        v_differences = np.empty([n_sites,n_differences])
        v_temporal_correlations = np.empty([n_sites,n_differences])
        
        v_differences[:] = np.nan
        v_temporal_correlations[:] = np.nan
        
        for i_site in range(n_sites):
            for i_difference in range(n_differences):
                if i_difference == 0:
                    v_1 = m_data_residual[:,i_site]
                    v_2 = m_data_residual[:,i_site]
                else:
                    v_1 = m_data_residual[:-i_difference,i_site]
                    v_2 = m_data_residual[i_difference:,i_site]
                
                id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
                if (sum(id_num) > 24):
                    v_differences[i_site,i_difference] = i_difference
                    v_temporal_correlations[i_site,i_difference] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                
        # Plot Results - Temporal Correlations
        plt.boxplot(v_temporal_correlations[~np.isnan(v_temporal_correlations).any(axis=1)])
        plt.xticks(range(0,n_differences+1),labels=range(0,n_differences+1),rotation=90)
        plt.ylim(-1,1)
        plt.xlim(0,n_differences)
        plt.xlabel('Time Difference [hrs]')
        plt.ylabel('Residual Correlation')
        figurename = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD] + ' (corr-time).png'
        plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
        plt.show()
        
        # Land Use Values and Correlations
        for Land_Use_Type in list(dataset_LUR):
            a_LUV_at_sites = dataset_LUR[Land_Use_Type].interp_like(a_data_residual)
            v_LUV = Array_to_Matrix_at_Coords(a_LUV_at_sites,m_site_coords)
    
            v_differences = np.empty([n_sites*n_sites])
            v_luv_correlations = np.empty([n_sites*n_sites])
            
            v_differences[:] = np.nan
            v_luv_correlations[:] = np.nan
            
            # Spatial Correlations
            i_line = 0
            for i_site_1 in range(n_sites):
                v_1 = m_data_residual[:,i_site_1]
                for i_site_2 in range(n_sites):
                    v_2 = m_data_residual[:,i_site_2]
                    
                    id_num = ~np.isnan(np.c_[v_1,v_2]).any(axis=1)
                    
                    if (sum(id_num) > 24):
                        v_differences[i_line] = np.abs(v_LUV[:,i_site_1] - v_LUV[:,i_site_2])
                        v_luv_correlations[i_line] = np.corrcoef(v_1[id_num],v_2[id_num])[0,1]
                        
                    i_line += 1
                    
            # Plot Results - Land Use Value Correlations
            plt.scatter(x=v_differences,y=v_luv_correlations)
            plt.ylim(-1,1)
            plt.xlim(0,np.nanmax(v_differences))
            plt.xlabel('Difference in ' + Land_Use_Type)
            plt.ylabel('Residual Correlation')
            figurename = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD] + ' (corr-luv-' + Land_Use_Type + ').png'
            plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
            plt.show()
            
            # Plot Results - Land Use Value v Residuals
            plt.boxplot(m_data_residual[~np.isnan(m_data_residual).any(axis=1)],positions=v_LUV[0],widths=0.05*np.nanmax(v_LUV[0]))
            plt.xlabel(Land_Use_Type)
            plt.ylabel('Residuals')
            plt.xticks(range(0,int(np.nanmax(v_LUV[0])),int(0.1*np.nanmax(v_LUV[0]))),labels=range(0,int(np.nanmax(v_LUV[0])),int(0.1*np.nanmax(v_LUV[0]))),rotation=90)
            figurename = s_figtitle_a + ' ' + s_figtitle_b + ' ' + PATTERN_METHOD_DICT[PATTERN_METHOD] + PATTERN_USAGE_DICT[PATTERN_USAGE] + REGRESSION_APPROACH_DICT[REGRESSION_APPROACH] + UPRES_METHOD_DICT[UPRES_METHOD] + ' (residual-v-' + Land_Use_Type + ').png'
            plt.savefig(os.path.join(OUTPUT_FIGURE_FOLDER,figurename),dpi=300)
            plt.show()